{-# LANGUAGE NoMonomorphismRestriction, TypeFamilies, FlexibleInstances, MultiParamTypeClasses, RecordWildCards, NamedFieldPuns, ScopedTypeVariables, RankNTypes, StandaloneDeriving, DeriveDataTypeable, DeriveGeneric, TemplateHaskell, OverloadedStrings, FlexibleContexts #-}

{- |

Module: Colony

Description : Data structure representing a generic colony of cells, information about those cells, etc.

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

A 'Colony' is a network of objects with communications channels
linking them.

The input to a colony at any given time is a list of integer
messages that are translated into modifications of the activation
levels of particular cells.

The output from a colony at any given time is a list of integer
messages that indicate that the activation level in a particular
cell meets a threshold.

-}

module Colony
       (
         -- * Specifications for a colony
         ColonySpec(..),
         EmitterSpec(..),
         ReceptorSpec(..),
         SynapseSpec(..),
         BadColonySpecException(..),
         SpikeTrain,
         Colony(..),
         ColonySnapshot,
         ColonyInput,
         ColonyInputStream,
         ColonyOutput,
         -- * Extracting information
         perCellOutputPatterns,
         perCellPresynapticPatterns,
         -- * Running input through a colony
         ColonyExecutionException(..),
         runColony,
         traceColony,
         -- * Results of tracing a run
         TraceResult(..),
         TracePoint(..),
         TraceStep(..),
         -- * Utilities
         mergeVectorMinMax
       ) where

import GHC.Generics

import Control.DeepSeq
import Control.Exception
import Control.Monad
import Control.Monad.ST

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Char
import Data.Data
import Data.DeriveTH
import qualified Data.Foldable as F
import qualified Data.Graph.Inductive as G
-- from fgl rather than containers:
import Data.Graph.Inductive (Node, LNode, Edge, LEdge)
import Data.Graph.Inductive.PatriciaTree
import Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import Data.IntSet (IntSet)
import qualified Data.IntSet as IS
import Data.List (foldl1', intercalate)
import Data.Maybe
import qualified Data.Map as M
import Data.Map (Map)
import qualified Data.Sequence as Seq
import Data.Sequence (
  Seq, (<|), (|>), ViewL(..), ViewR(..), viewl, viewr)
import Data.STRef
import Data.Typeable
import qualified Data.Vector.Unboxed as V
import Data.Vector.Unboxed (Vector)
import Text.Printf

import CircularBuffer
import Compiler
import MVM
import ShortShow
import Snapshot

colony_vcid :: String
colony_vcid = "$Id: Colony.hs,v 002890d5829b 2017-01-19 18:09:45Z garrett $"

-- | Integer message keys for inputs to send to a 'Colony'
type ColonyInput = [Int]

-- | Stream of inputs to a 'Colony'
type ColonyInputStream = [ColonyInput]

-- | A list of emitted integer message keys.
type ColonyOutput = [Int]


data EmitterSpec =
  EmitterSpec {
    sourcePattern :: !Int,
    -- ^ pattern to read in presynaptic or output cell
    sourceThreshold :: !Int,
    -- ^ if @A[sourcePattern] ≥ sourceThreshold@ then the
    -- channel is active
    activationDelta :: !Int
    -- ^ activation to add to @A[sourcePattern]@ in the
    -- presynaptic cell when this emitter is triggered, /usually
    -- negative!/
    } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)
-- Use generics:
$(deriveJSON defaultOptions ''EmitterSpec)
$(derive makeNFData ''EmitterSpec)
$(derive makeBinary ''EmitterSpec)

data ReceptorSpec =
  ReceptorSpec {
    targetPattern :: !Int,
    -- ^ pattern to emit
    targetDelta :: !Int
    -- ^ activation to added to @A[targetPattern]@ in the
    -- postsynaptic cell
    } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)
$(deriveJSON defaultOptions ''ReceptorSpec)
$(derive makeNFData ''ReceptorSpec)
$(derive makeBinary ''ReceptorSpec)

data SynapseSpec =
  SynapseSpec {
    emitterSpec :: EmitterSpec,
    delay :: Int,
    -- ^ how many time steps after activation until the signal
    -- reaches the target.  If 0, the signal arrives immediately.
    receptorSpec :: ReceptorSpec
    } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

$(deriveJSON defaultOptions ''SynapseSpec)
$(derive makeNFData ''SynapseSpec)
$(derive makeBinary ''SynapseSpec)

type SpikeTrain = Seq Bool


data ColonySpec =
  ColonySpec {
    numCells :: Int,
    -- ^ number of cells, indexed @0@ to @numCells - 1@
    cellNames :: IntMap String,
    -- ^ map from cell index to the name of that cell
    synapseSpecs :: [LEdge SynapseSpec],
    -- ^ initial synapse specifications
    inputSpecs :: IntMap (String, Int, ReceptorSpec),
    -- ^ input source specifications, interpreted as a map from
    -- the integer key for a source of input to a tuple of its
    -- name, the index of the cell to adjust, and the receptor
    -- for how the cell is triggered
    emitterSpecs :: [(String, Int, EmitterSpec, Int)]
    -- ^ output emitter specifications, interpreted as a tuple of
    -- the output's name, the index of the cell to read, emitter
    -- to use to read from that cell, integer message
    } deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

$(deriveJSON defaultOptions ''ColonySpec)
$(derive makeNFData ''ColonySpec)
$(derive makeBinary ''ColonySpec)

data Colony cell synapseState =
  Colony {
    colonySpec :: ColonySpec,
    cellGraph :: Gr cell (SynapseSpec, synapseState)
    -- ^ graph of cells
    } deriving (Typeable, Generic)

$(derive makeNFData ''Colony)
$(deriveJSON defaultOptions ''Colony)

-- These are a problem because the graph data type doesn't define
-- certain important instances:

-- $(derive makeNFData ''Colony)
-- $(derive makeBinary ''Colony)


-- | Shorthand for a mutable colony
type MColony s = Colony (MVM s) (CircularBuffer s)

-- | This is the type of the memento produced by calling
-- 'snapshot' on an 'MColony'
type ColonySnapshot = Colony (Vector Int) SpikeTrain

-- | Return an 'IntMap' from cell index to an 'IntSet' of all
-- patterns read from that cell to generate output.

perCellOutputPatterns :: ColonySpec -> IntMap IntSet
perCellOutputPatterns ColonySpec{..} =
  IM.fromListWith IS.union
  [(cellIx, IS.singleton sourcePattern)
  | (name, cellIx, EmitterSpec{..}, outMsg) <- emitterSpecs]

-- | Return an 'IntMap' from cell index to an 'IntSet' of all
-- patterns read from that cell by a synapse.

perCellPresynapticPatterns :: ColonySpec -> IntMap IntSet
perCellPresynapticPatterns ColonySpec{..} =
  IM.fromListWith IS.union
  [(preIx, IS.singleton sourcePattern)
  | (preIx, postIx, SynapseSpec{..}) <- synapseSpecs,
    let EmitterSpec{..} = emitterSpec]


data BadColonySpecException =
  BadColonySpec {
    badCellNames :: [(Int, String)],
    badInputs :: [(Int, (String, Int, ReceptorSpec))],
    badEmitters :: [(String, Int, EmitterSpec, Int)]
    } deriving (Read, Show, Eq, Typeable, Data, Generic)

instance Exception BadColonySpecException


checkColonySpec ColonySpec{..} =
  if null badCellNames
     && null badSynapses
     && null badInputs
     && null badEmitters
  then
    return ()
  else
    throw BadColonySpec{..}
  where
    cellIxBad j =
      j < 0 || j >= numCells
    cellNameBad (j, _) =
      cellIxBad j
    badCellNames =
      filter cellNameBad (IM.toList cellNames)
    synapseBad (j, k, _) =
      cellIxBad j || cellIxBad k
    badSynapses =
      filter synapseBad synapseSpecs
    inputBad (_, (_, j, _)) =
      cellIxBad j
    badInputs =
      filter inputBad (IM.toList inputSpecs)
    emitterBad (_, j, _, _) =
      cellIxBad j
    badEmitters =
      filter emitterBad emitterSpecs


-- | Return the highest pattern mentioned by any instruction from
-- the program that was compiled, or any of the synapses, inputs,
-- or emitters.
maxPattern :: ColonySpec -> ProgramInformation -> Int
maxPattern ColonySpec{..} ProgramInformation{..} =
  overallMax
  where
    maxRefReactions = patternsReferenced

    maxRefSyn = maximum (0:[
      max targetPattern sourcePattern
      | (u, v, synSp)  <- synapseSpecs,
        let SynapseSpec{..} = synSp,
        let ReceptorSpec{..} = receptorSpec,
        let EmitterSpec{..} = emitterSpec])

    maxRefIn = maximum (0:[
      targetPattern
      | (_, _, ReceptorSpec{..}) <- IM.elems inputSpecs])

    maxRefEm = maximum (0:[
      sourcePattern
      | (_, _, EmitterSpec{..}, _) <- emitterSpecs])

    overallMax = maximum [
      IS.findMax maxRefReactions,
      maxRefSyn,
      maxRefIn,
      maxRefEm
      ]


makeMutableColony ::
  ColonySpec
  -> (ProgramInformation, ReactionProgram) -- ^ genetic program for the cells
  -> ST s (MColony s)

makeMutableColony  cs@ColonySpec{..} (progInfo, reactions) = do
  -- Throw an exception if something is off
  checkColonySpec cs
  -- construct vitrual machines
  annotatedNodes <- forM [0 .. numCells - 1] $ \ j -> do
    mvm <- makeZeroMVM (1 + maxPattern cs progInfo) reactions
    return (j, mvm)
  annotatedEdges <- forM synapseSpecs $ \(pre, post, spec@SynapseSpec{..}) -> do
    synapse <- synapseWire delay
    return (pre, post, (spec, synapse))
  let cellGraph = G.mkGraph annotatedNodes annotatedEdges
  let colonySpec = ColonySpec{..}
  return Colony{..}

-- | Make a wire with @n@ zeros in it, and enough room for at
-- least @n+1@ items.
synapseWire :: Int -> ST s (CircularBuffer s)
synapseWire numSteps =
  let
    -- This is kind of overkill, but there's no point in skimping
    -- here.  There has to be room for at least two more bytes,
    -- so we can push a new byte in one end and pop one out the
    -- other, and at all times have room in the buffer so that
    -- the two pointers don't hit each other: numSteps + 1 here
    -- would theoretically be enough, except that it allows the
    -- two pointers to hit each other after the push and before
    -- the pop, and the resulting memento would indicate that the
    -- buffer is empty.
    q = quot (numSteps + 2) 16
    bSize = 16 * (q + 1)
  in do
     b <- makeCircularBuffer bSize
     replicateM numSteps (pushFront b 0)
     return b

-- | (Private) Get a cell by its index, raising an explanatory
-- error if it isn't present.

mcCheckedGetCell ::
  String
  -> MColony s
  -> Node -> ST s (MVM s)

mcCheckedGetCell hint c cellIx =
  case G.lab (cellGraph c) cellIx of
    Just cell ->
      return cell
    Nothing -> do
      s <- snapshot c
      let (minNode, maxNode) = G.nodeRange (cellGraph c)
      (error
       (printf
        "%s: cell index %d not found\nGraph includes %d nodes (%d, %d)\nlabeled nodes: %s\n"
        hint
        cellIx
        (G.noNodes (cellGraph c))
        minNode maxNode
        (show (G.labNodes (cellGraph s)))))


data ColonyExecutionException =
  UnknownInputMessage Int [(Int, (String, Int, ReceptorSpec))]
  deriving (Read, Show, Eq, Typeable, Data, Generic)

instance Exception ColonyExecutionException


-- | (Private) Send an input message to a specific cell
mcSendInputToCell ::
  MColony s
  -> Int -- ^ cell index
  -> ReceptorSpec -- ^ receptor to trigger
  -> ST s ()

mcSendInputToCell c cellIx ReceptorSpec{..} = do
  cell <- mcCheckedGetCell "Colony.mcSendInputToCell" c cellIx
  mvmApplyDelta cell targetPattern targetDelta

-- | (Private) Send simultaneous input messages

mcSendInputs ::
  MColony s
  -> ColonyInput
  -> ST s ()

mcSendInputs colony@Colony{..} colonyInput = do
  forM_ colonyInput $ \inMsg ->
    case IM.lookup inMsg (inputSpecs colonySpec) of
      Just (name, cellIx, rcpt) ->
        mcSendInputToCell colony cellIx rcpt
      Nothing ->
        throw (UnknownInputMessage inMsg (IM.toList (inputSpecs colonySpec)))


-- | (Private) Read an emitter for a cell

readEmitter ::
  MVM s
  -> EmitterSpec
  -> ST s Bool

readEmitter cell EmitterSpec{..}  = do
    a <- mvmReadActivation cell sourcePattern
    return (a >= sourceThreshold)



-- | (Private) Check the emiter part of a synapse, read the
-- interesting activation level in the presynaptic cell

mcRunSynapseEmitter ::
  MColony s
  -> LEdge (SynapseSpec, CircularBuffer s)
  -> ST s ()

mcRunSynapseEmitter colony@Colony{..} (emIx, rcIx, (synapseSpec, spikeTrainB)) =
  let
    ColonySpec{..} = colonySpec
    SynapseSpec{..} = synapseSpec
    EmitterSpec{..} = emitterSpec
    Just emittingCell = G.lab cellGraph emIx

  in do
    justEmitted <- readEmitter emittingCell emitterSpec
    when justEmitted $ do
      mvmApplyDelta emittingCell sourcePattern activationDelta
    pushBackBool spikeTrainB justEmitted


mcRunSynapseEmitters ::
  MColony s
  -> ST s ()

mcRunSynapseEmitters colony@Colony{..} =
  forM_ (G.labEdges cellGraph) (mcRunSynapseEmitter colony)



mcRunSynapseReceptor ::
  MColony s
  -> LEdge (SynapseSpec, CircularBuffer s)
  -> ST s ()

mcRunSynapseReceptor colony@Colony{..} (emIx, rcIx, (synapseSpec, spikeTrainB)) =
  let
    ColonySpec{..} = colonySpec
    SynapseSpec{..} = synapseSpec
    ReceptorSpec{..} = receptorSpec
    Just receivingCell = G.lab cellGraph rcIx

  in do
    justArrived <- popFrontBool spikeTrainB
    when justArrived $ do
      mvmApplyDelta receivingCell targetPattern targetDelta


mcRunSynapseReceptors ::
  MColony s
  -> ST s ()

mcRunSynapseReceptors colony@Colony{..} =
  forM_ (G.labEdges cellGraph) (mcRunSynapseReceptor colony)



runOutputEmitter ::
  MVM s
  -> EmitterSpec
  -> ST s Bool

runOutputEmitter emittingCell emitterSpec =
  let
    EmitterSpec{..} = emitterSpec
  in do
    justEmitted <- readEmitter emittingCell emitterSpec
    when justEmitted $ do
      mvmApplyDelta emittingCell sourcePattern activationDelta
    return justEmitted


mcRunOutput ::
  MColony s
  -> (String, Int, EmitterSpec, Int)
  -> ST s [Int]

mcRunOutput colony@Colony{..} (name, cellIx, emSpec, outMsg) =
  let
    Just emittingCell = G.lab cellGraph cellIx
  in do
    out <- runOutputEmitter emittingCell emSpec
    return (if out then [outMsg] else [])


mcRunOutputEmitters ::
  MColony s
  -> ST s ColonyOutput

mcRunOutputEmitters colony@Colony{colonySpec=ColonySpec{..}, ..} = do
  outMsgss <- mapM (mcRunOutput colony) emitterSpecs
  return (concat outMsgss)

-- NOTE: In earlier versions, I sent background input AFTER
-- handling synaptic messages.  I think it makes more sense to
-- send all inputs first, then deal with synaptic messages.  This
-- is potentially a disruptive change.  If background input ever
-- happens to get wired directly to an output pattern, and that
-- alone can trigger emission of output, then switching the order
-- of these actions might change the final result.
--
-- I don't think it affects
-- any of my sample runs because I never found one where the
-- background input got wired directly to an output pattern.

-- The order here is tricky: You have to be careful to read every
-- emitter, enqueue those values on the left end of their spike
-- trains.  After all the emitting, dequeue a value from the
-- right end of all the spike trains, then send those to the
-- appropriate receiving cell.  That way, if the spike train is
-- configured to impose zero delay, you send the signal straight
-- through from emitter to receiver.  Also the order in which
-- cells are processed should never change the result, so you
-- have to do all the reading before doing any writing.
-- Otherwise, it's possible that a signal from one cell pushes
-- the next to emit a signal this time step, which shouldn't be
-- possible.


-- | A 'TracePoint' is a time during a time step when only part
-- of the computation of a 'Colony' has been completed.
data TracePoint =
  AfterInputs
  | AfterSynapses
  | AfterReactions
  | AfterEmitters
    deriving (Read, Show, Eq, Ord, Enum, Data, Typeable, Generic)

instance NFData TracePoint
$(deriveJSON defaultOptions ''TracePoint)
$(derive makeBinary ''TracePoint)

class TraceRecorder recorder where

  type TracePointResult recorder :: *

  traceRecord :: recorder s -> TracePoint -> MColony s -> ST s ()

  readTracePointResult :: recorder s -> ST s (TracePointResult recorder)

  reset :: recorder s -> ST s ()


data NullTraceRecorder s = NullTraceRecorder

instance TraceRecorder NullTraceRecorder where
  type TracePointResult NullTraceRecorder = ()
  traceRecord NullTraceRecorder _ _ = return ()
  readTracePointResult NullTraceRecorder = return ()
  reset NullTraceRecorder = return ()

data SnapshotTraceRecorder s =
  SnapshotTraceRecorder {
    r :: STRef s (Map TracePoint ColonySnapshot)
    }

instance TraceRecorder SnapshotTraceRecorder where

  type TracePointResult SnapshotTraceRecorder = Map TracePoint ColonySnapshot

  traceRecord SnapshotTraceRecorder{..} tp c = do
    s <- snapshot c
    deepseq s $ modifySTRef' r (M.insert tp s)

  readTracePointResult SnapshotTraceRecorder{..} = readSTRef r

  reset SnapshotTraceRecorder{..} = writeSTRef r M.empty

makeSnapshotTraceRecorder :: ST s (SnapshotTraceRecorder s)

makeSnapshotTraceRecorder = do
  r <- newSTRef M.empty
  return SnapshotTraceRecorder{..}


-- | (Private) The rationale for all that class TraceRecorder
-- stuff is so that I can write this exactly once and have it
-- work with tracing on or off.

mcStepTr :: forall s tr result.
  (TraceRecorder tr,
   result ~ TracePointResult tr)
  =>
  tr s
  -> MColony s
  -> ColonyInput
  -> ST s (Int, Int, ColonyOutput)

mcStepTr tr colony@Colony{..} inputs = do
  -- phase
  mcSendInputs colony inputs
  traceRecord tr AfterInputs colony
  -- phase
  mcRunSynapseEmitters colony
  -- phase
  mcRunSynapseReceptors colony
  traceRecord tr AfterSynapses colony
  -- phase
  perCellReactionTotals <- forM (G.labNodes cellGraph) (\(node, mvm) -> mvmRunReactions mvm)
  traceRecord tr AfterReactions colony
  -- phase
  output <- mcRunOutputEmitters colony
  traceRecord tr AfterEmitters colony
  let totalReactionCount = sum (map fst perCellReactionTotals)
      totalReactionDeltas = sum (map snd perCellReactionTotals)
  return (totalReactionCount, totalReactionDeltas, output)

{-# SPECIALIZE mcStepTr :: NullTraceRecorder s -> MColony s -> ColonyInput -> ST s (Int, Int, ColonyOutput) #-}

{-# SPECIALIZE mcStepTr :: SnapshotTraceRecorder s -> MColony s -> ColonyInput -> ST s (Int, Int, ColonyOutput) #-}

data TraceStep = TraceStep {
  tsInput :: ColonyInput,
  tsColonySnapshots :: Map TracePoint ColonySnapshot,
  tsOutput :: ColonyOutput
  } deriving (Typeable, Generic)

$(derive makeNFData ''TraceStep)

-- | Details of a run of a 'Colony'
data TraceResult = TraceResult {
  trInputStream :: ColonyInputStream,
  trFinalOutput :: ColonyOutput,

  trPerCellOutputPatterns :: IntMap IntSet,
  -- ^ indexed by cell index, a set of all patterns ever read
  -- from that cell for colony output

  trPerCellPresynapticPatterns :: IntMap IntSet,
  -- ^ indexed by cell index, a set of all patterns ever read
  -- from that cell by a synapse

  trPerCellActivationRanges :: IntMap (Vector Int, Vector Int),
  -- ^ indexed by cell index, each pair consists of arrays of the
  -- (minimum, maximum) activation levels of each pattern in that
  -- cell over the entire trace
  trOverallActivationRanges :: (Vector Int, Vector Int)
  -- ^ arrays of the (minimum, maximum) activation levels of each
  -- pattern in all cells over the entire trace
  } deriving (Typeable, Generic)

$(derive makeNFData ''TraceResult)

-- | Map from cell index to the set of patterns that could be
-- read for 'Colony' output during a 'TraceStep'.
traceStepPerCellOutputPatterns ::
  TraceStep
  -> IntMap IntSet

traceStepPerCellOutputPatterns TraceStep{..} =
  IM.unionsWith IS.union
  [perCellOutputPatterns colonySpec
  | Colony{..} <- M.elems tsColonySnapshots]


-- | Map from cell index to the set of patterns that could be
-- read by the input portion of a synapse during a 'TraceStep'.
traceStepPerCellPresynapticPatterns ::
  TraceStep
  -> IntMap IntSet

traceStepPerCellPresynapticPatterns TraceStep{..} =
  IM.unionsWith IS.union
  [perCellPresynapticPatterns colonySpec
  | Colony{..} <- M.elems tsColonySnapshots]



-- | Map from cell index to the set of patterns that are read for
-- 'Colony' output at any time in a 'TraceResult'.

traceResultPerCellOutputPatterns ::
    Seq TraceStep
    -> IntMap IntSet

traceResultPerCellOutputPatterns trSteps =
  (IM.unionsWith IS.union
   (map traceStepPerCellOutputPatterns (F.toList trSteps)))


-- | Map from cell index to the set of patterns that could be
-- read by the input portion of a synapse at any time in a
-- 'TraceResult'.

traceResultPerCellPresynapticPatterns ::
    Seq TraceStep
    -> IntMap IntSet

traceResultPerCellPresynapticPatterns trSteps =
  (IM.unionsWith IS.union
   (map traceStepPerCellPresynapticPatterns (F.toList trSteps)))


mcTraceOn ::
  SnapshotTraceRecorder s
  -> (ColonyOutput -> Bool)
  -> MColony s
  -> [ColonyInput]
  -> ST s (Int, Int, Seq TraceStep)

mcTraceOn tr stopCheck colony (tsInput:moreInputs) = do
  (totalReactionCount, totalReactionDeltas, tsOutput) <- mcStepTr tr colony tsInput
  tsColonySnapshots <- readTracePointResult tr
  reset tr
  let ts = TraceStep{..}
  if stopCheck tsOutput
    then return (totalReactionCount, totalReactionDeltas, Seq.singleton ts)
    else do
    (totalReactionCount', totalReactionDeltas', moreTs) <- mcTraceOn tr stopCheck colony moreInputs
    let tSeq = ts <| moreTs
    deepseq tSeq $ return (totalReactionCount + totalReactionCount',
                           totalReactionDeltas + totalReactionDeltas',
                           tSeq)


mcTraceOn tr stopCheck colony [] =
  return (0, 0, Seq.empty)


mcRunOn ::
  (ColonyOutput -> Bool)
  -> MColony s
  -> [ColonyInput]
  -> ST s (Int, Int, Seq ColonyOutput)

mcRunOn stopCheck colony (input:moreInputs) = do
  (totalReactionCount, totalReactionDeltas, output) <- mcStepTr NullTraceRecorder colony input
  if stopCheck output
    then return (totalReactionCount, totalReactionDeltas, Seq.singleton output)
    else do
    (totalReactionCount', totalReactionDeltas', moreOutputs) <- mcRunOn stopCheck colony moreInputs
    return (totalReactionCount + totalReactionCount',
            totalReactionDeltas + totalReactionDeltas',
            output <| moreOutputs)

mcRunOn stopCheck colony [] =
  return (0, 0, Seq.empty)



-- | Run a colony on a given input stream, returning the stream
-- of outputs
runColony ::
  ColonySpec
  -> (ColonyOutput -> Bool) -- ^ function that checks when to stop
  -> (ProgramInformation, ReactionProgram) -- ^ reaction program
  -> ColonyInputStream -- ^ input to feed to the colony
  -> (Int, Int, Seq ColonyOutput)
  -- ^ (total reaction count, total reaction deltas, colony outputs)

runColony colonySpec stopCheck progAndInfo inputStream =
  runST $ do
    mc <- makeMutableColony colonySpec progAndInfo
    mcRunOn stopCheck mc inputStream

-- | Run a colony on a given input stream, returning a detailed
-- trace report.
traceColony ::
  ColonySpec
  -> (ColonyOutput -> Bool) -- ^ function that checks when to stop
  -> (ProgramInformation, ReactionProgram) -- ^ reaction program
  -> ColonyInputStream -- ^ input to feed to the colony
  -> (TraceResult, (Int, Int, Seq TraceStep))
  -- ^ (trace result, (total reaction count, total reaction deltas, colony outputs))

traceColony colonySpec stopCheck progAndInfo trInputStream =
  runST $ do
    tr <- makeSnapshotTraceRecorder
    mc <- makeMutableColony colonySpec progAndInfo
    runInfo@(totalReactionCount, totalReactionDeltas, trSteps) <-
      mcTraceOn tr stopCheck mc trInputStream
    let trPerCellActivationRanges =
          manyTraceStepsActivationRanges trSteps
        trOverallActivationRanges =
          foldr1 mergeVectorMinMax (IM.elems trPerCellActivationRanges)
        _ :> lastStep = viewr trSteps
        trFinalOutput = tsOutput lastStep
        trPerCellOutputPatterns =
          traceResultPerCellOutputPatterns trSteps
        trPerCellPresynapticPatterns =
          traceResultPerCellPresynapticPatterns trSteps
    return (TraceResult{..}, runInfo)


--- ACTIVATION RANGE ANALYSIS ---

-- | Map from cell index (vector of minimum activation of each
-- pattern, vector of maximum activation of each pattern) during
-- one 'TraceStep'.

traceStepActivationRanges ::
  TraceStep
  -> IntMap (Vector Int, Vector Int)

traceStepActivationRanges TraceStep{..} =
  IM.fromListWith mergeVectorMinMax
  [(cellIx, (cellMemento, cellMemento))
  | Colony{..} <- M.elems tsColonySnapshots,
    (cellIx, cellMemento) <- G.labNodes cellGraph]


manyTraceStepsActivationRanges ::
  Seq TraceStep
  -> IntMap (Vector Int, Vector Int)

manyTraceStepsActivationRanges trStepsSeq =
  IM.unionsWith mergeVectorMinMax (map traceStepActivationRanges (F.toList trStepsSeq))

-- | Given (minV1, maxV1) and (minV2, maxV2) of equal length
-- vectors, produce vectors (minV3, maxV3), where each element of
-- minV3 is the minimum of the corresponding elements of minV1
-- and minV2, and each element of maxV3 is the maximum of the
-- corresponding elements of maxV1 and maxV2.

mergeVectorMinMax ::
  (Vector Int, Vector Int)
  -> (Vector Int, Vector Int)
  -> (Vector Int, Vector Int)

mergeVectorMinMax (min1, max1) (min2, max2) =
  (V.zipWith min min1 min2,
   V.zipWith max max1 max2)


--- OTHER INSTANCES ---

instance
  (Snapshot m cell,
   Snapshot m synapseState) =>
  Snapshot m (Colony cell synapseState) where

    type Memento (Colony cell synapseState) =
      Colony (Memento cell) (Memento synapseState)

    snapshot Colony{..} = do
      -- mementos for all cells (nodes)
      cellMs <- forM (G.labNodes cellGraph) $ \(n, cell) -> do
        cellM <- snapshot cell
        return (n, cellM)

      -- mementos for all synapses (edges)
      synMs <- forM (G.labEdges cellGraph) $ \(pre, post, (synSpec, syn)) -> do
        synM <- snapshot syn
        return (pre, post, (synSpec, synM))

      let newCellGraph = G.mkGraph cellMs synMs
      return (Colony{ cellGraph = newCellGraph, ..})




-- Has to follow ColonySnapshot
$(deriveJSON defaultOptions{fieldLabelModifier = drop 2, constructorTagModifier = map toLower} ''TraceStep)
$(deriveJSON defaultOptions{fieldLabelModifier = drop 2, constructorTagModifier = map toLower} ''TraceResult)

-- These require corresponding instances for the graph implementation:
-- $(derive makeNFData ''TraceStep)
-- $(derive makeNFData ''TraceResult)
-- $(derive makeBinary ''TraceStep)
-- $(derive makeBinary ''TraceResult)

-- Orphan instances that I need to get all of this to work

instance
  (FromJSON nodeLabel,
   FromJSON edgeLabel) =>
  FromJSON (Gr nodeLabel edgeLabel) where

    parseJSON (Object v) = do
      nodes <- v .: "nodes"
      edges <- v .: "edges"
      return (G.mkGraph nodes edges)

    parseJSON _ = mzero

instance
  (ToJSON nodeLabel,
   ToJSON edgeLabel) =>
  ToJSON (Gr nodeLabel edgeLabel) where

    toJSON gr =
      object [
        "nodes" .= G.labNodes gr,
        "edges" .= G.labEdges gr
        ]

instance
  (FromJSON key,
   FromJSON value,
   Ord key) =>
  FromJSON (Map key value) where

    parseJSON obj = do
      aList <- parseJSON obj
      return (M.fromList aList)

instance
  (ToJSON key,
   ToJSON value,
   Ord key) =>
  ToJSON (Map key value) where

    toJSON = toJSON . M.toList


