{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables, DeriveDataTypeable, DeriveGeneric, NoMonomorphismRestriction, ParallelListComp, TemplateHaskell #-}

{-|

Module: EfficaciousAnalysis

Description: Tools for analyzing UM programs

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : POSIX

-}


module EfficaciousAnalysis (
  EfficaciousAnalysis(..),
  efficaciousAnalysis,
  ) where

import GHC.Generics

import Control.DeepSeq
import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.DeriveTH
import Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import Data.IntSet (IntSet)
import qualified Data.IntSet as IS
import Data.Set (Set)
import qualified Data.Set as S
--import Debug.Trace
import Data.Typeable
import qualified Data.Vector.Unboxed as V
import Data.Vector.Unboxed (Vector, (!))

import Colony
import Compiler
import Instruction

efficaciousAnalysis_vcid :: String
efficaciousAnalysis_vcid = "$Id: EfficaciousAnalysis.hs,v 002890d5829b 2017-01-19 18:09:45Z garrett $"



-- | (Private) Starting with the given patterns, repeatedly add
-- reactions from the given list of instructions that activate or
-- inhibit a pattern of interest.  Go until a closed network of
-- co-referential reactions has been built.  Return
-- the reactions of interest, and target patterns found by this
-- process.

efficaciousConnections ::
  IntSet
  -- ^ interesting patterns to start with
  -> Set Reaction
  -- ^ each instruction fragment to consider
  -> (Set Reaction, IntSet)

efficaciousConnections startingPatterns startingRxns =
  let
    -- closureLoop :: Set Reaction -> Set Reacion -> Set Int -> ...
    closureLoop remainingRxns includedRxns includedTPats =
      let
        -- Split remaining reactions into:
        --
        -- moreRxns = those that target a member of includedTPats
        --
        -- and
        --
        -- leftoverRxns = those that don't

        (moreRxns, leftoverRxns) =
          (S.partition
           (\rxn -> tPat rxn `S.member` includedTPats)
           remainingRxns)

        -- The switch patterns of moreFrgs are now of interest as
        -- targets of other instructions
        moreTPats = S.map sPat moreRxns
      in
       if S.null moreRxns
       then -- no more reactions of interest
         (includedRxns, includedTPats)
       else let -- more links to go; prep to loop

         -- We've found more fragments to include in the interesting list
         nextRxns = moreRxns `S.union` includedRxns

         -- And more target patterns of interest
         nextTPats = moreTPats `S.union` includedTPats
         in
          -- trace ("closureLoop:\n includedTPats="
          --       ++ show includedTPats
          --       ++ "\n moreTPats="
          --       ++ show moreTPats
          --       ++ "\n nextTpats="
          --       ++ show nextTPats ++ "\n") $
          closureLoop leftoverRxns nextRxns nextTPats


    -- Run the main loop to get
    --
    -- effRxnsSet = set of efficacious fragments
    --
    -- effTPatsSet = set of patterns targetted by efficacious fragments
    --  (Has to be a Set rather than an IntSet because of an fmap)
    (effRxnsSet, effTPatsSet) =
      closureLoop startingRxns S.empty (S.fromDistinctAscList (IS.toAscList startingPatterns))

  in
   (effRxnsSet, IS.fromDistinctAscList (S.toAscList effTPatsSet))


-- | This data type collects a bunch of pieces of information
-- that are produced together by 'efficaciousAnalysis' for
-- efficiency.

data EfficaciousAnalysis = EfficaciousAnalysis {
  {-
  -- Captured for debugging purposes
  perCellOutputPatterns :: IntMap IntSet,
  perCellPresynapticPatterns :: IntMap IntSet,
  perCellInterestingPatterns :: IntMap IntSet,
  perCellActivationRanges :: IntMap (Vector Int, Vector Int),
  -}


  perCellEfficacious ::
     IntMap (IntSet, Set Reaction, IntSet),
  -- ^ map from cell indices to the set of indices of efficacious
  --  instructions, a set of which reactions within those
  --  instructions are efficacious, and a set of all target
  --  patterns in all efficacious reactions

  allEfficaciousInstructionIxs :: IntSet,
  -- ^ indices of all efficacious instructions across all cells

  allEfficaciousReactions :: Set Reaction
  -- ^ all efficacious reactions

  } deriving (Eq, Ord, Read, Show, Data, Typeable, Generic)

-- $(derive makeBinary ''EfficaciousAnalysis)
$(derive makeNFData ''EfficaciousAnalysis)
$(deriveJSON defaultOptions ''EfficaciousAnalysis)


-- | Find the efficacious instructions and reactions.

efficaciousAnalysis ::
  ProgramInformation
  -> [TraceResult] -- ^ Trace results from all possible inputs
  -> EfficaciousAnalysis

efficaciousAnalysis ProgramInformation{..} traceResults =
  EfficaciousAnalysis {..}
  where
    reactionSet :: Set Reaction
    reactionSet = S.fromList reactions

    -- Interesting patterns come from these sources:
    -- * Output of the colony via emitters
    -- * Presynaptic patterns

    perCellOutputPatterns :: IntMap IntSet
    perCellOutputPatterns =
      (IM.unionsWith IS.union
       (map trPerCellOutputPatterns traceResults))


    perCellPresynapticPatterns :: IntMap IntSet
    perCellPresynapticPatterns =
      (IM.unionsWith IS.union
       (map trPerCellPresynapticPatterns traceResults))

    perCellInterestingPatterns :: IntMap IntSet
    perCellInterestingPatterns =
      IM.unionWith IS.union perCellOutputPatterns perCellPresynapticPatterns

    perCellActivationRanges :: IntMap (Vector Int, Vector Int)
    perCellActivationRanges =
      (IM.unionsWith mergeVectorMinMax
       (map trPerCellActivationRanges traceResults))


    effForCell :: Int -> IntSet -> (IntSet, Set Reaction, IntSet)
    effForCell cellIx interestingPatterns =
      let
        (minActivations, maxActivations) =
          perCellActivationRanges IM.! cellIx

        -- all reactions that are ever active in any of the traced runs
        activeRxns = S.filter isEverActive reactionSet

        isEverActive Reaction{..} =
          maxActivations ! sPat >= thr

        (effRxns, effPats) =
          efficaciousConnections interestingPatterns activeRxns

        effIxs = IS.fromAscList (map index (S.toAscList effRxns))

      in
       (effIxs, effRxns, effPats)

    -- This is the main calculation:
    perCellEfficacious =
      IM.mapWithKey effForCell perCellInterestingPatterns

    allEfficaciousInstructionIxs =
      IS.unions [effIxs
                | (effIxs, effRxns, effPats) <- IM.elems perCellEfficacious]

    allEfficaciousReactions =
      S.unions [effRxns
               | (effIxs, effRxns, effPats) <- IM.elems perCellEfficacious]
