{-# LANGUAGE RecordWildCards, TypeFamilies, RecordWildCards, NamedFieldPuns, DeriveDataTypeable, DeriveGeneric, TemplateHaskell #-}

{-|

Module      : Instruction

Description : Virtual machine instructions

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Instruction
       (Instruction(..),
        Reaction(..),
        printNumberedInstruction,
        printInstruction,
        printReaction,
        splitInstruction
        )
       where

import Control.DeepSeq
import Control.Monad

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Bits
import Data.Data
import Data.Default
import Data.DeriveTH
import GHC.Generics
import Data.List
import Data.String
import Data.Typeable

import Text.Printf

import ShortShow

instruction_vcid :: String
instruction_vcid = "$Id: Instruction.hs,v 002890d5829b 2017-01-19 18:09:45Z garrett $"

-- | One genetically encoded instruction, consisting of a switch
-- pattern and threshold, and a list of reactions.

data Instruction = Instruction {
  switchPattern :: Int,
  switchThreshold :: Int,
  targetsAndDeltas :: [(Int, Int)] -- ^ list of (target pattern, delta)
  } deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

$(derive makeNFData ''Instruction)
$(derive makeBinary ''Instruction)

-- I'm putting in explicit JSON serialization instances that don't
-- include all the key strings, just so things don't get so huge
instance ToJSON Instruction where
  toJSON Instruction{..} =
    toJSON (switchPattern, switchThreshold, targetsAndDeltas)

instance FromJSON Instruction where
  parseJSON obj = do
    (switchPattern, switchThreshold, targetsAndDeltas) <- parseJSON obj
    return Instruction{..}





-- | This uses fixed field widths, so it's appropriate for a
-- table in a fixed-width font.
printNumberedInstruction :: (Int, Instruction) -> String
printNumberedInstruction (j, instruction) =
  (printf "%5d: %s" j (printInstruction instruction))

-- | This uses fixed field widths, so it's appropriate for a
-- table in a fixed-width font.
printInstruction :: Instruction -> String
printInstruction Instruction{..} =
  (printf "A[%2d]≥%2d ⇒ %s"
   switchPattern switchThreshold
   reactions)
  where
    reactions = intercalate " " [
      printf "A[%2d]+=%2d" target delta
      | (target, delta) <- targetsAndDeltas ]

-- | For finer-grained program analysis, a 'Reaction' consists of
-- a switch pattern, a threshold, and a delta.  The
-- interpretation is that if @A[sPat] >= thr@ add @delta@ to
-- @A[tPat]@.  Each genetically encoded instruction can encode
-- several reactions.
data Reaction =
  Reaction {
    index :: Int, -- ^ location in list of instructions generated from a genome
    sPat :: Int, -- ^ switch pattern
    thr :: Int, -- ^ threshold
    tPat :: Int,  -- ^ target pattern
    delta :: Int -- ^ amount to add to @A[tPat]@ when active
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

$(deriveJSON defaultOptions ''Reaction)
$(derive makeBinary ''Reaction)
$(derive makeNFData ''Reaction)

-- | This uses fixed field widths, so it's appropriate for a
-- table in a fixed-width font.
printReaction :: Reaction -> String
printReaction Reaction{..} =
  (printf "%5d: A[%2d]≥%2d ⇒ A[%2d]+=%2d"
   index sPat thr tPat delta)

-- | Split an indexed instruction into reactions

splitInstruction ::
  (Int, Instruction) -> [Reaction]
splitInstruction (j, Instruction {..}) =
  [Reaction j switchPattern switchThreshold target delta
   | (target, delta) <- targetsAndDeltas]

-- I don't know why, but GHC complains that 'printInstruction' is
-- not in scope if this isn't at the bottom... !?!?
instance ShortShow Instruction where
  shortShow = printInstruction
