{-# LANGUAGE MultiParamTypeClasses, TypeFamilies, RecordWildCards, RankNTypes, ScopedTypeVariables  #-}

{-|

Module: MVM

Description: Mutable virtual machine

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module MVM (
  MVM,
  makeZeroMVM,
  mvmReadActivation,
  mvmApplyDelta,
  mvmRunReactions,
  ) where

import Control.Monad
import Control.Monad.ST

import qualified Data.Vector.Unboxed as V

import VMViaPrim
import Snapshot

-- | Mutable virtual machine
data MVM s = MVM {
  mvmProgram :: ReactionProgram,
  mvmState :: MutableVMState s,
  mvmNumPatterns :: Int
  }

instance Snapshot (ST s) (MVM s) where
  -- | Memento is a vector of @A[p]@
  type Memento (MVM s) = V.Vector Int

  snapshot mvm@MVM{..} =
    V.generateM mvmNumPatterns (mvmReadActivation mvm)


-- | Make an 'MVM' in which the initial activations are all 0.
makeZeroMVM ::
  Int -- ^ number of patterns in activation table
  -> ReactionProgram
  -> ST s (MVM s)

makeZeroMVM mvmNumPatterns mvmProgram = do
  mvmState <- newZeroMutableVMState mvmNumPatterns
  return MVM{..}



{-# INLINE mvmReadActivation #-}

-- | Read the current activation of a

mvmReadActivation :: MVM s -> Int -> ST s Int
mvmReadActivation MVM{..} pat =
  readCurrentActivation mvmState pat


{-# INLINE mvmApplyDelta #-}

-- | Adjust the activation of a pattern.  The input is
-- interpreted as a tuple of (pattern to affect, amount to add to
-- its activation)

mvmApplyDelta :: MVM s -> Int -> Int -> ST s ()
mvmApplyDelta  MVM{..} tgt delta =
  addActivation mvmState tgt delta


{-# INLINE mvmRunReactions #-}

-- | Run all reactions.  Return (total number of reactions run,
-- total delta added / subtracted to activation levels)

mvmRunReactions :: MVM s -> ST s (Int, Int)
mvmRunReactions MVM{..} =
  runReactions mvmProgram mvmState
