{-# LANGUAGE MagicHash, UnboxedTuples, RecordWildCards, BangPatterns #-}

{-|

Module: VMViaPrim

Description: Low level virtual machine

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : Non-portable (GHC only)

-}

module VMViaPrim (
  -- * Main interface
  BasicReaction,
  ReactionProgram,
  primCompile,
  primDecompile,
  MutableVMState,
  newZeroMutableVMState,
  cloneMutableVMState,
  readCurrentActivation,
  -- readFutureActivation,
  -- readBothActivations,
  readAllCurrentActivations,
  addActivation,
  runReactions,
  -- * Debugging tools (likely to change)
  dbgListActivations
  ) where

import Control.Monad
import Control.Monad.ST
import GHC.Exts
import GHC.ST -- not Control.Monad.ST by the way...

-- | (switch pattern, threshold, target pattern, activation delta),
-- Meaning if @A[switch pattern] >= threshold@,
-- then @ANext[target pattern] += activation delta@.
-- This is the format used by 'primCompile' and 'primDecompile'.

type BasicReaction = (Int, Int, Int, Int)

-- | A compiled reaction program.  Hardcoded in this
-- implementation: Use low 32 bits of Int's in each 'BasicReaction'.
-- Requires four int32's for each reaction.
data ReactionProgram = ReactionProgram {
  numReactions :: Int#,
                  -- ^ This has to be kept because GHC can
                  -- allocate longer 'ByteArray#'s than the size
                  -- requested.
  reactions :: ByteArray#
  }

instance Show ReactionProgram where
  show rp@ReactionProgram{..} =
    concat [
      "ReactionProgram {numReactions=",
      show (I# numReactions),
      ", reactions=",
      show (primDecompile rp),
      "}" ]


-- | Compile a list of reactions down to a primitive
-- 'ReactionProgram#'.  This does no error checking.  It just
-- packs everything to an internal byte array.

primCompile ::
  [BasicReaction]
  -> ReactionProgram

primCompile rs =
  runST (ST (compile# rs))

compile# ::
  [BasicReaction]
  -> State# s
  -> (# State# s, ReactionProgram #)
compile# rs s0 =
  let
    !(I# numReactions) = length rs
    (# s1, mArray #) = newByteArray# (numReactions *# 16#) s0
  in
   compileLoop# numReactions mArray 0# rs s0

compileLoop# ::
  Int# -- ^ number of reactions total
  -> MutableByteArray# s -- ^ where to compile to
  -> Int# -- ^ index of next reaction to write
  -> [BasicReaction]
  -> State# s
  -> (# State# s, ReactionProgram #)

compileLoop# numReactions mArray j [] s0 =
  let
    (# s1, reactions #) = unsafeFreezeByteArray# mArray s0
  in
   (# s1, ReactionProgram {..} #)

compileLoop# numReactions mArray j ((I# swp, I# thr, I# tgt, I# delta):rs) s0 =
  let
    s1 = writeInt32Array# mArray j swp s0
    s2 = writeInt32Array# mArray (j +# 1#) thr s1
    s3 = writeInt32Array# mArray (j +# 2#) tgt s2
    s4 = writeInt32Array# mArray (j +# 3#) delta s3
  in
   compileLoop# numReactions mArray (j +# 4#) rs s4


-- | Convert a 'ReactionProgram' back to an array of tuples
-- representing the reactions.  Assumes 4 int32's per reaction.
primDecompile ::
  ReactionProgram
  -> [(Int, Int, Int, Int)]

primDecompile (ReactionProgram {..}) =
  decompileLoop# numReactions reactions 0#

decompileLoop# numReactions a j
  | isTrue# (j >=# (numReactions *# 4#)) = []
  | otherwise =
    let
      swp = indexInt32Array# a j
      thr = indexInt32Array# a (j +# 1#)
      tgt = indexInt32Array# a (j +# 2#)
      delta = indexInt32Array# a (j +# 3#)
    in
     (I# swp, I# thr, I# tgt, I# delta) : decompileLoop# numReactions a (j +# 4#)





-- | An array mapping patterns (integers) to activation levels
type ActivationTable# = ByteArray#

-- | Implementation notes.  To avoid doing any allocation or
-- memory copying, a mutable VM state is stored twice.  There's
-- an old (current, read-only) state and a new (future,
-- in-progress, write-only) state.  During a time step,
-- activation levels are read from the old state, but updates are
-- applied to the new state.  Both states are stored in the same
-- array, but the old state starts at element 0, and the new
-- state starts at an offset of 'mvmNumPatterns'.  When a time
-- step ends, the new state is copied to become the next old
-- state.  The new state begins directly after the old state in
-- the internal array.

data MutableVMState s = MutableVMState {
  activationTable :: MutableActivationTable# s,
  numPatterns :: Int#
  }

type MutableActivationTable# s = MutableByteArray# s

{-# INLINE newZeroMutableVMState #-}

-- | Create a new machine state with zero activation everywhere.
newZeroMutableVMState ::
  Int -- ^ number of patterns to keep activation records for
  -> ST s (MutableVMState s)

newZeroMutableVMState numPatterns =
  ST axn
  where
    axn s0 =
      let
        (# sAfterAlloc, mvm #) = newMutableVMState# numPatterns s0
        sAfterZero = zeroOut# mvm sAfterAlloc
      in
       (# sAfterZero, mvm #)


{-# INLINE newMutableVMState# #-}

newMutableVMState# ::
  Int
  -> State# s
  -> (# State# s, MutableVMState s #)

newMutableVMState# (I# numPatterns) s0 =
  let
    (# sAfterAlloc, activationTable #) =
      newByteArray# (8# *# numPatterns) s0
      -- 2 of int32 for each pattern's activation
  in
   (# sAfterAlloc, MutableVMState{..} #)


{-# INLINE zeroOut# #-}

zeroOut# ::
  MutableVMState s
  -> State# s
  -> State# s

zeroOut# MutableVMState{..} =
  setByteArray# activationTable 0# (8# *# numPatterns) 0#


{-# INLINE cloneMutableVMState #-}

-- | Clone a 'MutableVMState' object.

cloneMutableVMState ::
  MutableVMState s
  -> ST s (MutableVMState s)

cloneMutableVMState mvm =
  ST (cloneMutableVMState# mvm)


{-# INLINE cloneMutableVMState# #-}

cloneMutableVMState# ::
  MutableVMState s
  -> State# s
  -> (# State# s, MutableVMState s #)

cloneMutableVMState# MutableVMState{..} s0 =
  let
    (# sAfterAlloc, clonedActivationTable #) =
      newByteArray# (8# *# numPatterns) s0
    sAfterCopy =
      copyMutableByteArray#
      activationTable 0#
      clonedActivationTable 0#
      (8# *# numPatterns)
      sAfterAlloc
  in
   (#
    sAfterCopy,
    MutableVMState {
      activationTable = clonedActivationTable,
      ..} #)

dbgDump ::
  MutableVMState s ->
  ST s [(Int,Int)]

dbgDump MutableVMState{..} =
  mapM fetch [0 .. jLast]
  where
    nBytes = sizeofMutableByteArray# activationTable
    jLast = I# (nBytes -# 1#)
    fetch = ST . fetch#
    fetch# (I# j) s0 =
      let
        (# s1, v #) = readInt8Array# activationTable j s0
      in
       (# s1, (I# j, I# v) #)

dbgListActivations ::
  MutableVMState s ->
  ST s [(Int, (Int, Int))]
dbgListActivations mvms@MutableVMState{..} =
  mapM fetch [0 .. pLast]
  where
    pLast = I# (numPatterns -# 1#)
    fetch j = do
      as <- readBothActivations mvms j
      return (j, as)

{-# INLINE readBothActivations #-}

-- | Read both activations (current, future)
readBothActivations ::
  MutableVMState s
  -> Int -- ^ pattern index to read
  -> ST s (Int, Int)

readBothActivations mvms j = do
  ca <- readCurrentActivation mvms j
  fa <- readFutureActivation mvms j
  return (ca, fa)


-- I really did try to "factor" the common structure out of these
-- functions, but there are "kind" problems dealing with primops
-- and I couldn't figure out how to make it compile..


{-# INLINE readCurrentActivation #-}

-- | Read the current activation level of a pattern.  Current
-- activation levels determine which reactions are active at this
-- time.

readCurrentActivation ::
  MutableVMState s
  -> Int -- ^ pattern index to read
  -> ST s Int

readCurrentActivation mvms (I# j) =
  ST axn
  where
    axn s0 =
      let
        (# s1, a #) = readCurrentActivation# mvms j s0
      in
       (# s1, I# a #)


{-# INLINE readFutureActivation #-}

-- | Read the future activation level of a pattern.  This is the
-- value of the activation pattern in the future state, as in the
-- current value plus any changes applied by reactions.

readFutureActivation ::
  MutableVMState s
  -> Int -- ^ pattern index to read
  -> ST s Int

readFutureActivation mvms (I# j) =
  ST axn
  where
    axn s0 =
      let
        (# s1, a #) = readFutureActivation# mvms j s0
      in
       (# s1, I# a #)



{-# INLINE readCurrentActivation# #-}

-- | Read the activation of the pattern from the old state.

readCurrentActivation# ::
  MutableVMState s
  -> Int#
  -> State# s
  -> (# State# s, Int# #)

readCurrentActivation# MutableVMState{..} j =
  readInt32Array# activationTable j


{-# INLINE readFutureActivation# #-}

-- | Read the activation of the pattern from the new (in progress) state.

readFutureActivation# ::
  MutableVMState s
  -> Int#
  -> State# s
  -> (# State# s, Int# #)

readFutureActivation# MutableVMState{..} j =
  readInt32Array# activationTable (numPatterns +# j)



{-# INLINE readAllCurrentActivations #-}

-- | Produce an a-list of all current activations

readAllCurrentActivations ::
  MutableVMState s ->
  ST s [(Int, Int)]

readAllCurrentActivations mvms@MutableVMState{..} =
  mapM fetch [0 .. pLast]
  where
    pLast = I# (numPatterns -# 1#)
    fetch j = do
      as <- readCurrentActivation mvms j
      return (j, as)


{-# INLINE addActivation #-}

-- | Add activation to a pattern.  Implementation note: Since the
-- mutable state requires a duplicate copy of all activations,
-- this function has to add to both places where @A[tgt]@ is
-- stored.

addActivation ::
  MutableVMState s
  -> Int -- ^ target pattern
  -> Int -- ^ activation delta
  -> ST s ()

addActivation MutableVMState{..} (I# tgt) (I# delta) =
  let
    axn s0 =
      let
        sAfterUpdateOld =
          addActivation# activationTable
          tgt delta s0

        sAfterUpdateNew =
          addActivation# activationTable
          (numPatterns +# tgt) delta sAfterUpdateOld
      in
       (# sAfterUpdateNew, () #)
  in
   ST axn


{-# INLINE addActivation# #-}

-- | This adds activation at a specific index, not for a
-- particular pattern, knows nothing of the duplicate state.
-- Assumes int32's.
addActivation# ::
  MutableActivationTable# s
  -> Int#
  -> Int#
  -> State# s
  -> State# s

addActivation# mActivationTable k delta s0 =
  let
    (# s1, v0 #) = readInt32Array# mActivationTable k s0
  in
   writeInt32Array# mActivationTable k (v0 +# delta) s1


{-# INLINE runReactions #-}

-- | Apply a program to the machine state

runReactions ::
  ReactionProgram
  -> MutableVMState s
  -> ST s (Int, Int)

runReactions ReactionProgram{..} MutableVMState{..} =
  ST axn
  where
  axn s0 =
    let
      (# reactionCount, activationTotal, sNext #) =
        runReactions# reactions numReactions activationTable numPatterns s0
    in 
    (# sNext, (I# reactionCount, I# activationTotal) #)

{-# INLINE runReactions# #-}

runReactions# ::
  ByteArray# -- ^ compiled reactions
  -> Int# -- ^ number of reactions
  -> MutableActivationTable# s -- ^ mutable state buffer
  -> Int# -- ^ newOffset, index of first int32 of new state
  -> State# s
  -> (# Int#, Int#, State# s #)
  -- ^ (reaction count, activation total, new world state)

runReactions# reactions numReactions activationTable newOffset s0 =
  (# reactionCount, activationTotal, sAfterTransfer #)
  where
    (# reactionCount, activationTotal, sAfterReactions #) =
      runReactionsLoop# 0# 0# 0# s0

    -- Copy new state to old state
    sAfterTransfer = -- based on Data.Primitive.ByteArray,
      copyMutableByteArray# -- I think these are correct:
      activationTable (newOffset *# 4#) -- source and offset
      activationTable 0# -- destination and offset
      (newOffset *# 4#) -- number of bytes
      sAfterReactions

    runReactionsLoop# reactionCount activationTotal j s0 =
      if isTrue# (j >=# (numReactions *# 4#))
      then
        (# reactionCount, activationTotal, s0 #)
      else let
        -- fetch reaction j
        swp = indexInt32Array# reactions j
        thr = indexInt32Array# reactions (j +# 1#)
        tgt = indexInt32Array# reactions (j +# 2#)
        delta = indexInt32Array# reactions (j +# 3#)

        -- fetch A[swp]
        (# sAfterReadSwp, a #) = readInt32Array# activationTable swp s0

        (# sAfterWrite, newReactionCount, newActivationTotal #) =
          -- If A[swp] >= thr
          if isTrue# (a >=# thr)
          then
            -- set ANew[tgt] := ANew[tgt] + 1
            (#
             addActivation# activationTable (newOffset +# tgt) delta sAfterReadSwp,
             reactionCount +# 1#,
             activationTotal +# (abs# delta)
             #)
          else
            (#
             sAfterReadSwp,
             reactionCount,
             activationTotal
             #)
        in
         runReactionsLoop# newReactionCount newActivationTotal (j +# 4#) sAfterWrite


{-# INLINE abs# #-}

abs# n# =
  if isTrue# (n# >=# 0#)
  then n#
  else negateInt# n#

testMakeZero = stToIO $ do
  mvm <- newZeroMutableVMState 4
  dbgListActivations mvm

testModify = stToIO $ do
  mvm <- newZeroMutableVMState 4
  addActivation mvm 1 5
  dbgListActivations mvm


-- swp, thr, tgt, delta
progNAND = primCompile [
  (0,1,3,-1)
  ,(1,1,3,-1)
  ,(2,0,3,2)
  ]

testNAND n p0 p1 = stToIO $ do
  mvm <- newZeroMutableVMState 4
  when p0 (addActivation mvm 0 1)
  when p1 (addActivation mvm 1 1)
  replicateM n (runReactions progNAND mvm)
  --dbgDump mvm
  dbgListActivations mvm

testNANDClone m n p0 p1 = stToIO $ do
  mvm <- newZeroMutableVMState 4
  when p0 (addActivation mvm 0 1)
  when p1 (addActivation mvm 1 1)
  replicateM m (runReactions progNAND mvm)
  mvmCopy <- cloneMutableVMState mvm
  replicateM n (runReactions progNAND mvm)
  copyActivations <- dbgListActivations mvmCopy
  vmActivations <- dbgListActivations mvm
  return (copyActivations, vmActivations)

testNANDFreeze m n p0 p1 = stToIO $ do
  mvm <- newZeroMutableVMState 4
  when p0 (addActivation mvm 0 1)
  when p1 (addActivation mvm 1 1)
  replicateM m (runReactions progNAND mvm)
  earlyAct <- readAllCurrentActivations mvm
  replicateM n (runReactions progNAND mvm)
  lateAct <- readAllCurrentActivations mvm
  return (earlyAct, lateAct)
