{-# LANGUAGE RecordWildCards, NamedFieldPuns, DeriveDataTypeable, DeriveGeneric, ScopedTypeVariables, ConstraintKinds, FlexibleContexts, OverloadedStrings, Arrows #-}

{-|
Module      : AllCrossHammingDistances

Description : Extract useful genes for doing phylogenetics experiments

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Main where

import Control.Applicative
import Control.DeepSeq
import Control.Monad

import Data.List
import Data.Map (Map)
import qualified Data.Map as M
import Data.Maybe

import Options.Applicative
import Options.Applicative.Arrows

import System.Directory
import System.FilePath
import System.IO

import Text.Regex.PCRE

import BitString

crossHammingDistances_vcid :: String
crossHammingDistances_vcid = "$Id: AllCrossHammingDistances.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"

data AllCrossHammingJob =
  AllCrossHammingJob {
    aDir :: FilePath,
    bDir :: FilePath
    }
  deriving (Show)

achjobParser :: ParserInfo AllCrossHammingJob
achjobParser =
  (info achjobOpts
   (progDesc "\
\Given two directory paths on the command line, A and B "))


achjobOpts :: Parser AllCrossHammingJob
achjobOpts = runA $ proc () -> do
  aDir <-
    (asA $ argument str
     (metavar "DIR-A"
      <> help "First directory")
    ) -< ()
  bDir <-
    (asA $ argument str
     (metavar "DIR-B"
      <> help "Second directory")
    ) -< ()
  returnA -< AllCrossHammingJob{..}


read01StringFile :: FilePath -> IO [BitString]
read01StringFile filePath = do
  fileData <- readFile filePath
  let fileLines = lines fileData
      genes = map bsFrom01String fileLines
  deepseq genes $ return genes


loadAgentFile :: String -> FilePath -> IO (Maybe (Integer, [BitString]))
loadAgentFile key path =
  let
      rx = "A([0-9]+)-" ++ key
      stem = takeBaseName path
      before :: String
      matching :: String
      after :: String
      subgroups :: [String]
      (before, matching, after, subgroups) =
        stem =~ rx
  in
   case subgroups of
     [nStr] ->
       let
         n :: Integer
         n = read nStr
       in do
         genes <- read01StringFile path
         return (Just (n, genes))
     _  ->
       return Nothing

loadGenomeFiles :: String -> FilePath -> IO (Map Integer [BitString])
loadGenomeFiles key dir = do
  allFiles <- getDirectoryContents dir
  agentDataMs <- mapM (loadAgentFile key) [dir </> p | p <- allFiles]
  let agentDatas = catMaybes agentDataMs
  return (M.fromList agentDatas)

scoreGeneSets :: [BitString] -> [BitString] -> Maybe Int
scoreGeneSets [] _ = Nothing
scoreGeneSets _ [] = Nothing
scoreGeneSets aStrings bStrings =
  let
    hammingDistances = sort [
      minimum hds
      | a <- aStrings,
        let hds = [bsHammingDistance a b | b <- bStrings]]
  in case hammingDistances of
    (n:_) -> Just n
    [] -> Nothing


scoreAgentSets ::
  Map Integer [BitString]
  -> Map Integer [BitString]
  -> Maybe Double

scoreAgentSets leftAMap rightAMap =
  if null scores
  then Nothing
  else Just (floatMean scores)

  where
    scoreMs :: [Maybe Int]
    scoreMs = do
      a <- M.elems leftAMap
      b <- M.elems rightAMap
      return (scoreGeneSets a b)

    scores = catMaybes scoreMs

floatMean :: [Int] -> Double
floatMean xs =
  fromIntegral sumXs / fromIntegral n
  where
    n = length xs
    sumXs = sum xs


main = do
  AllCrossHammingJob{..} <- execParser achjobParser
  xs <- forM ["B", "G83", "G93"] $ \key -> do
    leftAMap <- loadGenomeFiles key aDir
    rightAMap <- loadGenomeFiles key bDir
    let x = scoreAgentSets leftAMap rightAMap
    return x
  putStrLn (intercalate ", " (map show xs))
  return ()
