{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables, FlexibleContexts, ParallelListComp, TypeFamilies, DeriveDataTypeable, DeriveGeneric, OverloadedStrings, BangPatterns, TemplateHaskell, MultiParamTypeClasses, FlexibleInstances #-}

{-|

Module      : CoincidenceDetectorProblem

Description : Detecting two spikes in close succession

Copyright   :  © 2017 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Code for the problem of responding to two action potentials in
close succession

-}

module CoincidenceDetectorProblem where

import GHC.Generics

import Control.DeepSeq

import Data.Aeson()
import Data.Aeson.TH
import Data.Data
import Data.Default
import Data.DeriveTH
import Data.List
import Data.Typeable ()
import Text.Pandoc()
import Text.Pandoc.Builder

import AesonUtil
import GeneralName
import NeuronProblem
import Problem
import Snip


coincidenceDetectorProblem_vcid :: String
coincidenceDetectorProblem_vcid = "$Id: CoincidenceDetectorProblem.hs,v b51b0c75f531 2017-01-22 03:16:45Z garrett $"

data CoincidenceDetectorProblem =
  CoincidenceDetectorProblem {
    neuronProblem :: NeuronProblem,
    -- ^ underlying 'NeuronProblem'
    correctOutputPoints :: Double,
    -- ^ points earned for correct output during the window of an
    -- 'ActionPotentialTasklet'.  For a tasklet that requires a
    -- spike, these points are given if at least once during the
    -- output window the colony sets the output channel.  For a
    -- tasklet that requires silence, these points are given if
    -- the colony never sets the output channel.
    excessOutputPoints :: Double,
    -- ^ points lost for each step during the output window that
    -- the colony sets the output channel in excess of 1; usually
    -- of small magnitude
    falsePositivePoints :: Double,
    -- ^ points lost if an 'ActionPotentialTasklet' specifies no
    -- output should be generated, but the colony sets the output
    -- channel at any time during the output window; usually of
    -- medium magnitude.
    falseNegativePoints :: Double,
    -- ^ points lost if an 'ActionPotentialTasklet' specifies
    -- that a spike should be generated, but the colony never
    -- sets the output channel during the output window; usually
    -- of medium magnitude.
    quietPoints :: Double,
    -- ^ points earned per time step for generating no output
    -- when there is no input
    inputFailurePoints :: Double,
    -- ^ points lost per time step for generating output during
    -- a time when it's supposed to be quietly accumulating input
    trainingTasks :: TrainingSet,
    -- ^ inventory of tasks for training
    reactionCostFactor :: Double,
    -- ^ multiply the number of reactions used by this and
    -- subtract it from the total score
    deltaCostFactor :: Double
    -- ^ multiply the total of all reaction deltas applied less
    -- by this and subtract it from the total score
    } deriving (Read, Show, Eq, Ord, Data, Typeable)


data TrainingSet =
  Medium
  | Long
  | LearnMedium
  | LearnLong
  | Custom [CoincidenceDetectorTask]
  deriving (Read, Show, Eq, Ord, Data, Typeable)

instance GeneralName CoincidenceDetectorProblem where
  generalName _ = "CoincidenceDetector"

-- | score factor and tasklet sequence
data CoincidenceDetectorTask =
  CoincidenceDetectorTask Double [ActionPotentialTasklet]
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)



instance ToBools CoincidenceDetectorTask where
  toBools (CoincidenceDetectorTask _ tasklets) =
    toBools tasklets

-- | list of (pulse width, time steps until next pulse begins)
type ActionPotentialStream = [(Int, Int)]


data ActionPotentialTasklet =
  ActionPotentialTasklet {
    numLeadingOff :: Int,
    -- ^ begin with this many steps of no input
    actionPotentials :: ActionPotentialStream,
    -- ^ then send these spikes
    numTrailingOff :: Int,
    -- ^ end with this many steps of no input
    numOutputSteps :: Int,
    -- ^ neuron should spike during this time window from the end
    -- of the input or not at all
    shouldSpike :: Bool
    -- ^ should a spike during the output window
    } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)



instance ToBools ActionPotentialTasklet where
  toBools ActionPotentialTasklet{..} =
    (concat
     (concat
      [
        [replicate numLeadingOff False],
        [replicate pulseWidth True
         ++ replicate gap False
        | (pulseWidth, gap) <- actionPotentials],
        [replicate numTrailingOff False]]))


taskletDuration :: ActionPotentialTasklet -> Int

taskletDuration t@ActionPotentialTasklet{..} =
  (numLeadingOff
   + taskletAPDuration t
   + numTrailingOff)


taskletAPDuration :: ActionPotentialTasklet -> Int

taskletAPDuration ActionPotentialTasklet{..} =
  sum [width + gap
      | (width, gap) <- actionPotentials]


data TaskletTally =
  TaskletTally {
    quietTimeSteps :: Int,
    preludeFailures :: Int,
    inputFailures :: Int,
    correctOutput :: Int,
    excessOutputs :: Int,
    falsePositives :: Int,
    falseNegatives :: Int,
    postludeFailures :: Int
    } deriving (Read, Show, Eq, Ord, Data, Typeable)



tallyScore :: CoincidenceDetectorProblem -> TaskletTally -> Double

tallyScore CoincidenceDetectorProblem{..} TaskletTally{..} =
    (fromIntegral correctOutput * correctOutputPoints
     - fromIntegral excessOutputs * excessOutputPoints
     - fromIntegral falsePositives * falsePositivePoints
     - fromIntegral falseNegatives * falseNegativePoints
     - fromIntegral inputFailures * inputFailurePoints
     + (fromIntegral
        (quietTimeSteps
         - preludeFailures
         - postludeFailures)) * quietPoints)

taskletTally :: ActionPotentialTasklet -> [Bool] -> TaskletTally
taskletTally t@ActionPotentialTasklet{..} output =
--  trace
--  ("CoincidenceDetectorProblem.taskletTally:\n"
--   ++ show t
--   ++ "\n"
--   ++ boolsTo01s output
--    ++ "\n"
--    ++ show tally)
  tally
  where
    quietTimeSteps = taskletDuration t - numOutputSteps
    (duringPrelude, afterPrelude) =
      splitAt numLeadingOff output

    -- count steps with high output during initial off prelude
    preludeFailures = sum (map fromEnum duringPrelude)

    apDuration = taskletAPDuration t

    (duringAPs, afterAPs) =
      splitAt apDuration afterPrelude

    --  high output during input action potentials, basically a
    --  false positive
    inputFailures = maximum (0 : map fromEnum duringAPs)

    (duringWindow, afterWindow) =
      splitAt numOutputSteps afterAPs

    -- count the number of on steps during output window
    apDuringOutputWindow = sum (map fromEnum duringWindow)

    correctOutput =
      if numOutputSteps == 0
      then 0
      else
        if shouldSpike
        then min 1 apDuringOutputWindow
        else
          if apDuringOutputWindow == 0
          then 1
          else 0

    excessOutputs =
      if shouldSpike && apDuringOutputWindow >= 1
      then apDuringOutputWindow - 1
      else 0

    falsePositives =
      if not shouldSpike && apDuringOutputWindow >= 1
      then 1
      else 0

    falseNegatives =
      if shouldSpike && apDuringOutputWindow < 1
      then 1
      else 0

    -- count steps with high output after output window
    postludeFailures = sum (map fromEnum afterWindow)

    tally = TaskletTally {..}



cdTaskTally :: [ActionPotentialTasklet] -> [Bool] -> [TaskletTally]

cdTaskTally [] _ = []

cdTaskTally (t:ts) output =
  taskletTally t thisOutput : cdTaskTally ts moreOutput
  where
    (thisOutput, moreOutput) =
      splitAt (taskletDuration t) output




--- Training sets ---

-- | Make all lists consisting of 'n' tasklets in all possible
-- orders.  There's probably some slick way to do this, but this
-- will do:

buildCDTasks ::
  Int -- ^ 'n'
  -> [a] -- ^ 'tasklets'
  -> [[a]]

buildCDTasks 1 tasklets =
  [[t]
  | t <- tasklets]

buildCDTasks numTaskletsPerTask tasklets =
  let
    taskTails = buildCDTasks (numTaskletsPerTask - 1) tasklets
  in
   [tasklet : taskTail
    | tasklet <- tasklets,
      taskTail <- taskTails]


-- | Medium length set of tasks
tsMedium :: [CoincidenceDetectorTask]
tsMedium =
  concat [
    [CoincidenceDetectorTask 10000 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 100 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2]
    ]


-- | Long set of tasks
tsLong :: [CoincidenceDetectorTask]
tsLong =
  concat
  [
    [CoincidenceDetectorTask 10000 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 100 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 3 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 3 cdTasklets2]
  ]


-- | Medium-length set of tasks, along with tasks for learning:
-- Some begin with a long period of inactivity, which should
-- weaken the input synapse and cause the neuron to spike less
-- easily.  Others begin with a period of very high activity,
-- which should strengthen the input synapse and cause the neuron
-- to spike more easily.
tsLearn :: [CoincidenceDetectorTask]
tsLearn =
  concat
  [
    [CoincidenceDetectorTask 10000 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 100 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2],

    -- There aren't many of these, so make them worth more each
    [CoincidenceDetectorTask 10 [tasklet]
    | tasklet <- cdTaskletsWeaken],

    [CoincidenceDetectorTask 1
     (highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen]
  ]

-- | Long set of tasks, along with tasks for learning
tsLearnLong :: [CoincidenceDetectorTask]
tsLearnLong =
  concat
  [
    [CoincidenceDetectorTask 10000 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 100 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2],

    -- There aren't many of these, so make them worth more each
    [CoincidenceDetectorTask 10 [tasklet]
    | tasklet <- cdTaskletsWeaken],

    [CoincidenceDetectorTask 1
     (highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen],

    -- Weaken then strengthen
    [CoincidenceDetectorTask 1
     (cdTaskletsWeaken ++ highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen],

    -- Strengthen then weaken
    [CoincidenceDetectorTask 1
     (highActTasklets ++ [afterStrTasklet] ++ cdTaskletsWeaken)
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen]
  ]


--- Tasklets ---

-- | A list of the most basic coincidence detection tasklets:
-- various intervals with one or two spikes, some of which are
-- close enough that the receiving neuron should respond with an
-- action potential  of its own.

cdTasklets1 :: [ActionPotentialTasklet]

cdTasklets1 = [
  -- totally quiet, shouldn't trigger
  ActionPotentialTasklet {
     numLeadingOff = 0,
     actionPotentials = [],
     numTrailingOff = 20,
     numOutputSteps = 0,
     shouldSpike = False },
  -- isloated spike, shouldn't trigger
  ActionPotentialTasklet {
     numLeadingOff = 4,
     actionPotentials = [(1,0)],
     numTrailingOff = 15,
     numOutputSteps = 4,
     shouldSpike = False },
  -- back-to-back spikes, should trigger
  ActionPotentialTasklet {
     numLeadingOff = 4,
     actionPotentials = [(1,0),(1,0)],
     numTrailingOff = 14,
     numOutputSteps = 4,
     shouldSpike = True },
  -- two very close spikes, should trigger
  ActionPotentialTasklet {
     numLeadingOff = 4,
     actionPotentials = [(1,1),(1,0)],
     numTrailingOff = 13,
     numOutputSteps = 4,
     shouldSpike = True },
  -- two close spikes, should trigger
  ActionPotentialTasklet {
    numLeadingOff = 4,
    actionPotentials = [(1,2),(1,0)],
    numTrailingOff = 12,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two separated spikes, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 4,
    actionPotentials = [(1,6),(1,0)],
    numTrailingOff = 8,
    numOutputSteps = 4,
    shouldSpike = False },
  -- two widely separated spikes, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 4,
    actionPotentials = [(1,8),(1,0)],
    numTrailingOff = 6,
    numOutputSteps = 4,
    shouldSpike = False }
  ]


-- | A list of more difficult coincidence detection tasklets of
-- various lengths: various intervals with several spikes, some
-- of which are close enough that the receiving neuron should
-- respond with an action potential of its own.

cdTasklets2 :: [ActionPotentialTasklet]

cdTasklets2 = [
  -- totally quiet, shouldn't trigger
  ActionPotentialTasklet {
     numLeadingOff = 0,
     actionPotentials = [],
     numTrailingOff = 20,
     numOutputSteps = 0,
     shouldSpike = False },
  -- one spike, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 10,
    actionPotentials = [(1,0)],
    numTrailingOff = 10,
    numOutputSteps = 4,
    shouldSpike = False },
  -- back-to-back spikes, should trigger
  ActionPotentialTasklet {
    numLeadingOff = 5,
    actionPotentials = [(1,0),(1,0)],
    numTrailingOff = 9,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two close spikes, should trigger
  ActionPotentialTasklet {
    numLeadingOff = 8,
    actionPotentials = [(1,2),(1,0)],
    numTrailingOff = 20,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two very close spikes, should trigger, but with an odd spike beforehand
  ActionPotentialTasklet {
    numLeadingOff = 10,
    actionPotentials = [(1,8),(1,1),(1,0)],
    numTrailingOff = 6,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two closeish spikes, should trigger
  ActionPotentialTasklet {
    numLeadingOff = 10,
    actionPotentials = [(1,3),(1,0)],
    numTrailingOff = 20,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two separated spikes, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 5,
    actionPotentials = [(1,6),(1,0)],
    numTrailingOff = 8,
    numOutputSteps = 4,
    shouldSpike = False },
  -- two widely separated spikes, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 7,
    actionPotentials = [(1,8),(1,0)],
    numTrailingOff = 10,
    numOutputSteps = 4,
    shouldSpike = False }
  ]


-- | Tasklets in which there's a very long period of inactivity
-- during which the synaptic connection should "weaken".
-- Afterward, the cell should only respond to back-to-back
-- inputs.

cdTaskletsWeaken :: [ActionPotentialTasklet]

cdTaskletsWeaken = [
  -- very long wait followed by back-to-back spikes, should trigger
  ActionPotentialTasklet {
     numLeadingOff = 100,
     actionPotentials = [(1,0),(1,0)],
     numTrailingOff = 10,
     numOutputSteps = 4,
     shouldSpike = True },
  -- long wait, a loose spike, followed by back-to-back spikes, should trigger
  ActionPotentialTasklet {
     numLeadingOff = 50,
     actionPotentials = [(1,50),(1,0),(1,0)],
     numTrailingOff = 10,
     numOutputSteps = 4,
     shouldSpike = True },
  -- very long wait followed by two close spikes, should NOT trigger
  ActionPotentialTasklet {
    numLeadingOff = 100,
    actionPotentials = [(1,4),(1,0)],
    numTrailingOff = 20,
    numOutputSteps = 4,
    shouldSpike = False },
  -- long wait,a loose spike, followed by two close spikes, should NOT trigger
  ActionPotentialTasklet {
    numLeadingOff = 50,
    actionPotentials = [(1,50),(1,4),(1,0)],
    numTrailingOff = 20,
    numOutputSteps = 4,
    shouldSpike = False }
  ]

-- | Short tasklets for creating a period of high activity during
-- which the synaptic connection should "strengthen".  Afterward,
-- the cell should spike after slightly separated input pulses
-- that it shouldn't otherwise respond to.

cdTaskletsHighActivity :: [ActionPotentialTasklet]

cdTaskletsHighActivity = [
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,0),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,1),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,2),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True }
  {-
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,3),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,4),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,5),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True } -}
  ]


-- | Tasklets for use after high activity.  These are different
-- from 'tasklets0', because in these tasklets, input spikes
-- separated by 6 should trigger an action potential.

cdTaskletsAfterStrengthen :: [ActionPotentialTasklet]

cdTaskletsAfterStrengthen = [
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,6),(1,0)],
     numTrailingOff = 3,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,8),(1,0)],
     numTrailingOff = 3,
     numOutputSteps = 3,
     shouldSpike = False }
  ]







--- Problem templates ---

instance Default CoincidenceDetectorProblem where

  def = CoincidenceDetectorProblem {
    neuronProblem = def,
    correctOutputPoints = 64,
    excessOutputPoints = 4,
    falsePositivePoints = 0,
    falseNegativePoints = 32,
    inputFailurePoints = 4,
    quietPoints = 1,
    trainingTasks = Custom [],
    reactionCostFactor = 0,
    deltaCostFactor = 0
    }


problemTemplateCDMedium :: CoincidenceDetectorProblem
problemTemplateCDMedium =
  def {trainingTasks = Medium}

problemTemplateCDLong :: CoincidenceDetectorProblem
problemTemplateCDLong =
  def {trainingTasks = Long}

problemTemplateCDLearnMedium :: CoincidenceDetectorProblem
problemTemplateCDLearnMedium =
  def {trainingTasks = LearnMedium}

problemTemplateCDLearnLong :: CoincidenceDetectorProblem
problemTemplateCDLearnLong =
  def {trainingTasks = LearnLong}


---- Problem instance ----
-- This has to go at the bottom because of stuff not being in
-- scope !?!?

instance Problem CoincidenceDetectorProblem where

  type Input CoincidenceDetectorProblem = CoincidenceDetectorTask

  type Output CoincidenceDetectorProblem = [Bool]

  type Rating CoincidenceDetectorProblem = Double

  initialColonySpec CoincidenceDetectorProblem{..} =
    singleNeuronColonySpec neuronProblem

  problemInputToColonyInputStream
    CoincidenceDetectorProblem{..}
    (CoincidenceDetectorTask _ input) =
    boolsToColonyInputStream neuronProblem boolInput
    where
      boolInput = toBools input

  colonyOutputsToProblemOutput
    CoincidenceDetectorProblem{..}
    colonyOutputSeq =
    colonyOutputsToBools neuronProblem colonyOutputSeq

  minLandmarkRating _ = 0

  inputsForRating CoincidenceDetectorProblem{..} =
    case trainingTasks of
      Custom ts ->
        ts
      Medium ->
        tsMedium
      Long ->
        tsLong
      LearnMedium ->
        tsLearn
      LearnLong ->
        tsLearnLong

  rateResult problem (CoincidenceDetectorTask scoreFactor input) output
    totalReactionCount totalReactionDeltas =
    scoreFactor * rawScore
    where
      CoincidenceDetectorProblem{..} = problem
      tallies = cdTaskTally input output
      tallyTotal = foldl' (\s t -> s + tallyScore problem t) 0 tallies
      reactionCountCost =
        reactionCostFactor * fromIntegral totalReactionCount
      reactionDeltaCost =
        deltaCostFactor * fromIntegral totalReactionDeltas
      reactionPenalty = reactionCountCost + reactionDeltaCost
      rawScore = tallyTotal - reactionPenalty

instance Sniplet CoincidenceDetectorProblem CoincidenceDetectorTask where
  toInlines _ t =
    code $ boolsToDrawing $ toBools t


instance Sniplet CoincidenceDetectorProblem [Bool] where
  toInlines _ bs =
    code $ boolsToDrawing $ bs

$(deriveJSON myJSONOpts ''CoincidenceDetectorProblem)
$(deriveJSON myJSONOpts ''ActionPotentialTasklet)
$(deriveJSON myJSONOpts ''CoincidenceDetectorTask)
$(deriveJSON myJSONOpts ''TaskletTally)
$(deriveJSON myJSONOpts ''TrainingSet)

$(derive makeNFData ''ActionPotentialTasklet)
$(derive makeNFData ''CoincidenceDetectorTask)
