{-# LANGUAGE RecordWildCards, NamedFieldPuns, NoMonomorphismRestriction #-}

{-|

Module: ColonyToStyledNetwork

Description: Tools for drawing a Colony as a GraphML document

Copyright: © 2016 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : POSIX


-}

module ColonyToStyledNetwork (
  colonyToStyledNetwork,
  efficaciousAnalysisToStyledNetwork
  ) where

import Control.Applicative
import qualified Data.Foldable as Foldable
import Data.IntMap (IntMap)
import qualified Data.Graph.Inductive as Graph
import qualified Data.IntMap as IntMap
import Data.IntSet (IntSet)
import qualified Data.IntSet as IntSet
import Data.List
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Maybe
import Data.Sequence (Seq)
import qualified Data.Sequence as Seq
import Data.Set (Set)
import qualified Data.Set as Set

import Text.Printf
import Text.XML.Light

import StyledNetwork

import Colony
import Compiler
import EfficaciousAnalysis
import Instruction

colonyAsStyledNetwork_vcid :: String
colonyAsStyledNetwork_vcid = "$Id: ColonyToStyledNetwork.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"

pattern :: Int -> Int -> String
pattern cellIx p =
  printf "P|%d:%d" cellIx p


-- Negative input indices are interpreted as background inputs
input :: Int -> String
input inputIx =
  if inputIx >= 0
  then printf "I|%d" inputIx
  else printf "B|%d" (-inputIx)



output :: Int -> String
output outputIx =
  printf "O|%d" outputIx


edge :: String -> String -> String
edge fromKey toKey =
  printf "%s>%s" fromKey toKey

reaction :: Int -> String -> String -> String
reaction index fromKey toKey =
  printf "R|%d:%s" index (edge fromKey toKey)

network :: Int -> String
network cellIx =
  printf "C|%d" cellIx


-- | (Private) Make a 'StyledEdge' representing a synapse.  The
-- input is structured to match the result of 'synapseSpecs'.
-- The result is a complicated tuple:
--
-- * (cell index, pattern) for presynaptic end
--
-- * styled presynaptic node
--
-- * (cell index, pattern) for postsynaptic end
--
-- * styled postsynaptic node
--
-- * styled edge

synapseToStyled ::
  (Int, Int, SynapseSpec)
  -- ^ (index of presynaptic cell,
  --    index of postsynaptic cell,
  --    'SynapseSpec')
  -> ((Int, Int), StyledNode, (Int, Int), StyledNode, StyledEdge)

synapseToStyled labeledEdge =
  ((fromCellIx, sourcePattern),
   StyledNode {
      nodeKey = pattern fromCellIx sourcePattern,
      nodeLabel = show sourcePattern,
      nodeStyle = SynapseSourceNode
      },
   (toCellIx, targetPattern),
   StyledNode{
      nodeKey = pattern toCellIx targetPattern,
      nodeLabel = show targetPattern,
      nodeStyle = SynapseTargetNode
      },
   StyledEdge{..})
  where
    (fromCellIx, toCellIx, synSpec) = labeledEdge
    SynapseSpec{..} = synSpec
    EmitterSpec{..} = emitterSpec
    ReceptorSpec{..} = receptorSpec
    fromKey = pattern fromCellIx sourcePattern
    toKey = pattern toCellIx targetPattern
    edgeKey = edge fromKey toKey
    edgeLabel = printf "≥%d⇒%+d [%d]" sourceThreshold targetDelta delay
    edgeStyle = if targetDelta > 0 then ActivationEdge else InhibitionEdge



-- | (Private) The input is structured to match the result of
-- applying 'IntMap.toList' to 'inputSpecs'.  The output is a
-- tuple:
--
-- * Maybe an input node.  These go in a boxed subgraph.
--
-- * Maybe a background input node.  These are particular to each
-- target and sit outside boxed subgraphs.  Each cell is assumed
-- to have at most one of these, and it's index is the negative
-- of the cell's index.
--
-- * (cell, pattern) of input target.  This node is styled specially.
--
-- * Styled target node.
--
-- * Styled edge.
--
-- One of the input nodes is always 'Nothing' and the other is
-- 'Just' the styled input node.  The point of doing it this way
-- is that a boxed subgraph is built for the ordinary inputs, and
-- it's easy enough to add a 'Mabye' to a running list.

inputToStyled ::
  (Int, (String, Int, ReceptorSpec)) -- ^ matches associations in 'inputSpecs'
  -> (Maybe StyledNode, Maybe StyledNode, (Int, Int), StyledNode, StyledEdge)

inputToStyled (inputIx, (inputName, targetCellIx, ReceptorSpec{..})) =
  (if isOrdinary then Just styledSource else Nothing,
   if isOrdinary then Nothing else Just styledSource,
   (targetCellIx, targetPattern),
   styledTarget,
   StyledEdge{..})
  where
    isOrdinary = inputIx >= 0
    styledSource = StyledNode {
      nodeKey = input inputIx,
      nodeLabel = inputName,
      nodeStyle = if isOrdinary then InputSourceNode else ReducedNode
      }
    styledTarget = StyledNode {
      nodeKey = pattern targetCellIx targetPattern,
      nodeLabel = show targetPattern,
      nodeStyle = if isOrdinary then InputTargetNode else BackgroundInputTargetNode
      }
    fromKey = nodeKey styledSource
    toKey = nodeKey styledTarget
    edgeKey = edge fromKey toKey
    edgeLabel = ""
    edgeStyle = if targetDelta >= 0 then ActivationEdge else InhibitionEdge


-- | (Private) The input is structured to match 'emitterSpecs'.
-- The output of this is a tuple:
--
-- * (cell, pattern) of emitter source.  This node is styled specially.
--
-- * An output node.  These go in a boxed subgraph.  If the
-- message is -1, then the output node is the stop node and is
-- styled specially.
--
-- * Styled edge.  These are always draw as activation edges.

emitterToStyled :: (String, Int, EmitterSpec, Int) -> ((Int, Int), StyledNode, StyledNode, StyledEdge)
emitterToStyled (outputName, sourceCellIx, EmitterSpec{..}, message)  =
  ((sourceCellIx, sourcePattern),
   styledSource, styledTarget, StyledEdge{..})
  where
    styledSource = StyledNode {
      nodeKey = pattern sourceCellIx sourcePattern,
      nodeLabel = show sourcePattern,
      nodeStyle = OutputSourceNode
      }
    styledTarget = StyledNode {
      nodeKey = output message,
      nodeLabel = if message == -1 then "" else show message,
      nodeStyle = if message == -1 then StopNode else OutputTargetNode
      }
    fromKey = nodeKey styledSource
    toKey = nodeKey styledTarget
    edgeKey = edge fromKey toKey
    edgeLabel = ""
    edgeStyle = ActivationEdge


-- | (Private) Returns a tuple of
--
-- * (cell index, source pattern)
--
-- * (cell index, target pattern)
--
-- * Styled edge.
--
-- The point of returning tuples instead of styled nodes is that
-- some pattern nodes are specially styled, and most pattern
-- nodes participate in many reactions.  So it makes more sense
-- to create those nodes separately.

reactionToStyled ::
  Int
  -> Reaction
  -> ((Int, Int), (Int, Int), StyledEdge)

reactionToStyled cellIx Reaction{..} =
  ((cellIx, sPat), (cellIx, tPat), StyledEdge{..})
  where
    fromKey = pattern cellIx sPat
    toKey = pattern cellIx tPat
    edgeKey = reaction index fromKey toKey
    edgeLabel = printf "≥%d⇒%+d" thr delta
    edgeStyle = if delta >= 0 then ActivationEdge else InhibitionEdge



-- | (Private) Return a 'StyledNetwork' for all patterns in this
-- cell involved in any of the given 'Reaction's.

cellToStyled ::
  String
  -- ^ label for the resulting 'StyledNetwork'
  -> [Reaction]
  -- ^ reactions to consider
  -> Map (Int, Int) StyledNode
  -- ^ map from (cell index, pattern) to 'StyledNode' for special
  -- patterns.  All other patterns will be given a 'StyleNode'
  -- with style 'OrdinaryNode'.
  -> Int
  -- ^ index of this cell
  -> StyledNetwork

cellToStyled networkLabel reactions specialPatternTable cellIx =
  StyledNetwork{..}
  where
    -- rxnSources :: [(Int, Int)]
    -- rxnTargets :: [(Int, Int)]
    (rxnSources, rxnTargets, rxnEdges) = unzip3 (map (reactionToStyled cellIx) reactions)
    referencedPatterns = Set.fromList rxnSources `Set.union` Set.fromList rxnTargets

    thisCell (n, p) _ = (cellIx == n)
    specialNodeMap = Map.filterWithKey thisCell specialPatternTable
    specialPatterns = Map.keysSet specialNodeMap
    specialNodes = Map.elems specialNodeMap
    ordinaryPatterns = referencedPatterns `Set.difference` specialPatterns
    ordinaryNodes = [
      StyledNode {nodeStyle = OrdinaryNode, ..}
      | (cellIx, p) <- Set.toList ordinaryPatterns, let nodeKey = pattern cellIx p, let nodeLabel = show p]


    networkKey = network cellIx
    nodes = ordinaryNodes ++ specialNodes
    edges = rxnEdges
    subnetworks = []



colonyToStyledNetwork ::
  String -- ^ overall label for the network
  -> ColonySpec
  -> IntMap [Reaction] -- ^ map from cell index to reactions to include
  -> StyledNetwork

colonyToStyledNetwork networkLabel ColonySpec{..} cellToReactions =
  StyledNetwork{
    networkKey = "N|" ++ networkLabel,
    nodes = bgSourceNodes,
    edges = concat [inEdges, outEdges, synEdges],
    subnetworks = inSubnet : outSubnet : cellNetworks,
    ..
    }
  where

    (synSources, synSourceNodes,
     synTargets, synTargetNodes,
     synEdges) = unzip5 (map synapseToStyled synapseSpecs)

    (mInSourceNodes, mBGSourceNodes,
     inTargets, inTargetNodes,
     inEdges) = unzip5 (map inputToStyled (IntMap.toList inputSpecs))

    inSourceNodes = catMaybes mInSourceNodes
    bgSourceNodes = catMaybes mBGSourceNodes

    inSubnet = StyledNetwork {
      networkKey = "N|input",
      networkLabel = "Input",
      nodes = inSourceNodes,
      edges = [], -- no edges interior to input subnet
      subnetworks = []
      }

    (outSources, outSourceNodes,
     outTargetNodes,
     outEdges) = unzip4 (map emitterToStyled emitterSpecs)

    outSubnet = StyledNetwork {
      networkKey = "N|output",
      networkLabel = "Output",
      nodes = outTargetNodes,
      edges = [], -- no edges interior to output subnet
      subnetworks = []
      }

    specialPatternTable =
      (Map.fromList
       (concat
        [zip synSources synSourceNodes,
         zip synTargets synTargetNodes,
         zip inTargets inTargetNodes,
         zip outSources outSourceNodes]))

    cellNetworks = do
      (cellIx, cellLabel) <- IntMap.toList cellNames
      let reactions = IntMap.findWithDefault [] cellIx cellToReactions
      return (cellToStyled cellLabel reactions specialPatternTable cellIx)



efficaciousAnalysisToStyledNetwork ::
  EfficaciousAnalysis
  -> String
  -> ColonySpec
  -> StyledNetwork

efficaciousAnalysisToStyledNetwork EfficaciousAnalysis{..} networkLabel colonySpec =
  colonyToStyledNetwork networkLabel colonySpec cellToReactions
  where
    cellToReactions =
      IntMap.map (\(_, rxnSet, _) -> Set.toList rxnSet) perCellEfficacious
