{-# LANGUAGE RecordWildCards, NamedFieldPuns, DeriveDataTypeable, DeriveGeneric, ScopedTypeVariables, ConstraintKinds, FlexibleContexts, OverloadedStrings, Arrows #-}

{-|
Module      : Copy21PhyloExtract

Description : Extract useful genes for doing phylogenetics experiments

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Extract genes from a sample run.  This is all specialized for my
normal Copy21 setup with beacon.

-}

module Main where

import Control.Applicative
import Control.DeepSeq
import Control.Monad
import Control.Monad.Trans

import Data.Aeson
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Maybe
import Data.Set (Set)
import qualified Data.Set as Set
import Data.Typeable

import Options.Applicative
import Options.Applicative.Arrows

import System.IO
import System.FilePath

import Text.Printf

import AesonUtil
import Agent
import BitString
import Database
import Genome
import Instruction
import Job
import Logging
import PopulationDatabase
import PopulationSpec
import PredicateTools
import Problem
import ProblemInventory
import Query
import qualified QueryRequest as Q
import qualified QueryResult as R
import Report
import ReportFormatTypes
import ShortShow

import EngineUtil

-- | Command line.
data PhyloExtractJob =
  PhyloExtractJob {
    samplesDir :: FilePath,
    sampleName :: String,
    minRating :: Int
    }
  deriving (Show)

pejobParser :: ParserInfo PhyloExtractJob
pejobParser =
  (info pejobOpts
   (progDesc "Copy21-problem-specific phylogenetic information extraction; give --help for more help"))

pejobOpts :: Parser PhyloExtractJob
pejobOpts = runA $ proc () -> do
  samplesDir <-
    (asA $ argument str
     (metavar "SAMPLES_DIR"
      <> help "Data directory storing sample runs")
    ) -< ()
  sampleName <-
    (asA $ argument str
     (metavar "SAMPLE_NAME"
      <> help "Name of the sample (subdirectory)")
    ) -< ()
  mMinRating <-
    (asA $ optional $ option auto
     (long "min-rating"
      <> metavar "RATING"
      <> help "Consider all agents with at least this rating")
    ) -< ()
  let minRating = fromMaybe 470320 mMinRating
  returnA -< PhyloExtractJob{..}


-- | Main routine.  Load a sample run of an FPCopy21BE problem
-- (or similar) and examine each agent with the given minimum
-- rating.  In the sample run's directory, for each such agent,
-- save the genes for each gene key in a file named @Agent/{agent
-- ID number}/-/{gene key}/.csv@.

main = do
  PhyloExtractJob{..} <- execParser pejobParser

  let environmentDirectory = samplesDir </> sampleName
      databaseSpec = DatabaseSpec{..}
      configFile = samplesDir </> "Config.yaml"
      genomeFile = samplesDir </> "Genome.yaml"
      problemFile = samplesDir </> "Problem.yaml"
  assertDirectoryExists samplesDir
  config@Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture
  withDatabase databaseSpec $ withPopCacheX genomeSpec problem $ do
    aMap <- doSelect (RatedAtleast (toJSON minRating))
    forM_ (Map.elems aMap) $ \a -> do
      let geneAscs = analyzeAgent genomeSpec problem a
      forM_ geneAscs $ \(key, genes) -> do
        let gFile =
              (environmentDirectory
               </> printf "A%d-%s.txt" (identifier a) key)
            content = unlines (map bsTo01String genes)
        liftIO $ do
          writeFile gFile content
  return ()

-- | (Subroutine) Given a 'GenomeSpec', 'Problem', and 'Agent',
-- return a list of pairs (gene key, genes from the agent that
-- fit that key).

analyzeAgent ::
  (GenomeSpec gs,
   Problem prob,
   ShortShow (Interval gs),
   NFData (Output prob),
   NFData (Input prob),
   NFData (Rating prob)) =>
  gs
  -> prob
  -> Agent (Genome gs) (Rating prob)
  -> [(String, [BitString])]

analyzeAgent genomeSpec problem a@Agent{egg=Egg{..}, ..} = do
  (key, check) <- copy21Elements
  let relevantGenes =
        [nucleotides
        | (j, (instruction, nucleotides)) <- ixedInstructionDetails,
          check (j, instruction)]
  return (key, relevantGenes)
  where
    -- I originally did this with the whole genome.  Need to be
    -- smarter about this... Some agents have genes that refer to
    -- the pattern of interest but are not the genes I'm looking
    -- for (not efficacious)

    -- > R.GenomeDetails{..} = queryAgent genomeSpec problem Q.GenomeDetails a
    R.Efficacious{..} = queryAgent genomeSpec problem Q.Efficacious a
    effInstIxs = Set.fromList [j | (j, _) <- allEfficaciousInstructions]
    ixedInstructionDetails = [ (j, (instruction, nucleotideBits))
      | (j, GeneDetails{..}) <- zip [0..] (geneDetails genomeSpec genome),
        j `Set.member` effInstIxs ]


-- | This is a table of gene keys and tests that define them.
-- They are basically the names of different functions genes
-- might have.  They are
--
-- * B: beacon gene, activates pattern 31 (the output beacon)
--
-- * P0, P1: gene switched by patterns 8 or 9 (where the sender receives input)
--
-- * S: gene targets pattern 3 (sender side of synapse)
--
-- * R: gene switched by pattern 4 (receiver side of synapse)
--
-- * Q0, Q1: gene targets patterns 12 or 13 (where the receiver generates output)
--
-- * H: gene targets pattern 0 (receiver's stop signal, as in halt)

copy21Elements :: [(String, (Int, Instruction) -> Bool)]
copy21Elements = [
  ("B", instTMatch 31),
  ("P0", instSMatch 8),
  ("P1", instSMatch 9),
  ("S", instTMatch 3),
  ("R", instSMatch 4),
  ("Q0", instTMatch 12),
  ("Q1", instTMatch 13),
  ("H", instTMatch 0)]



-- | Return whether the given instruction is switched on or off
-- by the given pattern.

instSMatch ::
  Int -- ^ switch pattern to test
  -> (Int, Instruction) -- ^ program index and instruction
  -> Bool

instSMatch s (j, instruction) =
  switchPattern == s
  where
    Instruction{..} = instruction
    reactions = splitInstruction (j, instruction)



-- | Return whether the given instruction includes a reaction
-- that targets the given pattern

instTMatch ::
  Int -- ^ target pattern to test
  -> (Int, Instruction) -- ^ program index and instruction
  -> Bool

instTMatch t (j, instruction) =
  t `elem` targetPatterns
  where
    Instruction{..} = instruction
    reactions = splitInstruction (j, instruction)
    targetPatterns = map tPat reactions
