{-# LANGUAGE RecordWildCards, NamedFieldPuns, DeriveDataTypeable, DeriveGeneric, ScopedTypeVariables, ConstraintKinds, FlexibleContexts, OverloadedStrings #-}

{-|
Module      : Engine

Description : Main simulation engine

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Main where

import GHC.Generics

import Control.Concurrent
import Control.DeepSeq
import Control.Exception
import Control.Monad
import Control.Monad.State
import Control.Monad.Trans

import Data.Aeson hiding (Array)
import qualified Data.Aeson as J
import qualified Data.Aeson.Types as JT
import Data.Binary (Binary)
import qualified Data.ByteString as BS
import Data.ByteString.Builder
import qualified Data.ByteString.Lazy as BSL
import qualified Data.ByteString.Lazy.UTF8 as LU
import Data.Data
import Data.Default
import Data.Either
import Data.List (find)
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Time.Clock
import Data.Time.Format
import Data.Time.LocalTime
import Data.Typeable
import Data.Word
import qualified Data.Yaml as Yaml
import Options.Applicative
import Options.Applicative.Arrows
import System.Directory
import System.Environment
import System.Exit
import System.FilePath
import System.IO
import System.Locale
import qualified System.Log.Handler as LH
import qualified System.Log.Handler.Simple as LHS
import System.Log.Formatter
import System.Log.Logger
import System.Posix.Signals
import Text.Printf

import Agent
import Database
import Evolution
import GeneralName
import Genome
import Job
import Logging
import PopulationDatabase
import PopulationSpec
import PredicateTools
import Problem
import ProblemInventory
import Query
import qualified QueryRequest as Q
import qualified QueryResult as R
import Report
import ReportFormatTypes
import MonadUtil
import RandomST

import EngineUtil

main = do
  job@(Args cOpts@CommonOpts{..} cmd) <- execParser jobParser
  -- To start with, the global logger notices everything, but its
  -- only handler sends WARNING and higher messages to stderr
  wHandler <- logToHandle stderr WARNING
  updateGlobalLogger rootLoggerName (
    setLevel DEBUG
    . setHandlers [wHandler]
    )
  -- Do work
  (doCmd cOpts cmd
   `catches`
   [Handler handleTargetAchievedEx,
    Handler handleConfigurationEx])


doCmd :: CommonOpts -> Command -> IO ()

doCmd cOpts List = do
  printf "Problem templates:\n"
  showKeys (Map.keys problemTemplates)
  printf "Genome templates:\n"
  showKeys (Map.keys genomeSpecTemplates)
  where
    prefix = " "
    showKeys keys = do
      forM_ keys $ \key -> do
        putStr prefix
        putStrLn key

doCmd cOpts Template{..} = do
  createDirectoryIfMissing True samplesDir
  Box prob <- getProblemTemplate problemTemplateName
  Box (genomeSpec :: gs) <- getGenomeSpecTemplate genomeArchitectureTemplateName
  let problemName = generalName prob
      genomeArchitecture = generalName genomeSpec
      archive = Never
      target = def
      config = Config{..}
  yamlEncodeFileSafe (samplesDir </> "Alignment.yaml") (def :: AlignmentSpec gs)
  yamlEncodeFileSafe (samplesDir </> "Config.yaml") config
  yamlEncodeFileSafe (samplesDir </> "Genome.yaml") genomeSpec
  yamlEncodeFileSafe (samplesDir </> "Problem.yaml") prob
  yamlEncodeFileSafe (samplesDir </> "Selection.yaml") (def :: SelectionSpec)

doCmd CommonOpts{..} Init{..} = do
  let environmentDirectory = samplesDir </> sampleName
      databaseSpec = DatabaseSpec{..}
      alignmentFile = samplesDir </> "Alignment.yaml"
      configFile = samplesDir </> "Config.yaml"
      genomeFile = samplesDir </> "Genome.yaml"
      problemFile = samplesDir </> "Problem.yaml"
      selectionFile = samplesDir </> "Selection.yaml"
  assertDirectoryExists samplesDir
  sendLogToFile (samplesDir </> "Log.txt") logLevel
  config@Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture
  alignmentSpec <- checkedLoadYAML alignmentFile
  selectionSpec <- checkedLoadYAML selectionFile
  let populationSize = resultingPopulationSize selectionSpec
      populationSpec = PopulationSpec{..}
  (genomes, randomSeed) <- case (identical, importFrom) of
    (False, "") -> do
      seed0 <- createRandomSeed
      let (genomes, randomSeed) = withSeed seed0 $ do
            pullSampleList populationSize genomeSpec
      return (genomes, randomSeed)
    (False, importPath) -> do
      assertFileNotEmpty importPath
      importBytes <- BS.readFile importPath
      let genomes = checkImportGenomes genomeSpec problem importBytes
      randomSeed <- createRandomSeed
      return (genomes, randomSeed)
    (True, "") -> do
      seed0 <- createRandomSeed
      let (prototype, randomSeed) = withSeed seed0 $ do
            pullSample genomeSpec
      return (replicate populationSize prototype, randomSeed)
    (True, path) -> do
      throwIO $ ConflictingSettings {
        message = "Request for identical starting agents conflicts with request to import from " ++ path
        }
  deepseq genomes $ createDirectoryIfMissing True environmentDirectory
  initializeFirstGeneration populationSpec problem genomes randomSeed archive


doCmd CommonOpts{..} Reset{..} = do
  let environmentDirectory = samplesDir </> sampleName
      databaseSpec = DatabaseSpec{..}
      alignmentFile = samplesDir </> "Alignment.yaml"
      configFile = samplesDir </> "Config.yaml"
      genomeFile = samplesDir </> "Genome.yaml"
      problemFile = samplesDir </> "Problem.yaml"
      selectionFile = samplesDir </> "Selection.yaml"
  assertDirectoryExists environmentDirectory
  assertFileNotEmpty (environmentDirectory </> "problemName.txt")
  sendLogToFile (environmentDirectory </> "Log.txt") logLevel
  config@Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box (genomeSpec :: gs) <- loadGenomeSpec genomeFile genomeArchitecture
  alignmentSpec <- checkedLoadYAML alignmentFile
  selectionSpec <- checkedLoadYAML selectionFile
  let populationSize = resultingPopulationSize selectionSpec
      populationSpec = PopulationSpec{..}
      initRating = minLandmarkRating problem
      recordGenerations =
        case archive of
          Never -> False
          _ -> True
  ir <- (reinitializePopulationDatabase
         environmentDirectory initRating
         recordGenerations archive)
  let InitializationRec{..} = ir
      newEggsRatings =
        [(Egg [] g, rateGenome genomeSpec problem g) | g <- genomes]
      delta = PopulationBirthDeathDelta{toKill = [], ..}
  withDatabase databaseSpec $ withPopCacheX genomeSpec problem $ do
      applyBirthDeathDelta delta
      nextTime
  return ()


doCmd CommonOpts{..} Run{..} = do
  let environmentDirectory = samplesDir </> sampleName
      databaseSpec = DatabaseSpec{..}
      alignmentFile = samplesDir </> "Alignment.yaml"
      configFile = samplesDir </> "Config.yaml"
      genomeFile = samplesDir </> "Genome.yaml"
      problemFile = samplesDir </> "Problem.yaml"
      selectionFile = samplesDir </> "Selection.yaml"
  assertDirectoryExists samplesDir
  config@Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture
  alignmentSpec <- checkedLoadYAML alignmentFile
  selectionSpec <- checkedLoadYAML selectionFile
  let populationSize = resultingPopulationSize selectionSpec
      populationSpec = PopulationSpec{..}

  checkSig <- makeSigStopper
  checkClock <- maybeHelperIO maxSeconds makeClockChecker

  createDirectoryIfMissing True environmentDirectory
  sendLogToFile (environmentDirectory </> "Log.txt") logLevel

  withDatabase databaseSpec $ withPopCache $ do
    untilTrue $ do
      quickReport
      writeStatus
      checkSig
      checkClock
      checkTarget genomeSpec problem target
      fullProblemNextGeneration populationSpec problem
      return False
  return ()

doCmd CommonOpts{..} Continue{..} = do
  let environmentDirectory = samplesDir </> sampleName
      databaseSpec = DatabaseSpec{..}
      alignmentFile = samplesDir </> "Alignment.yaml"
      configFile = samplesDir </> "Config.yaml"
      genomeFile = samplesDir </> "Genome.yaml"
      problemFile = samplesDir </> "Problem.yaml"
      selectionFile = samplesDir </> "Selection.yaml"
  assertDirectoryExists samplesDir
  config@Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture
  alignmentSpec <- checkedLoadYAML alignmentFile
  selectionSpec <- checkedLoadYAML selectionFile
  let populationSize = resultingPopulationSize selectionSpec
      populationSpec = PopulationSpec{..}

  assertDirectoryExists environmentDirectory
  sendLogToFile (environmentDirectory </> "Log.txt") logLevel

  checkSig <- makeSigStopper
  
  withDatabase databaseSpec $ withPopCache $ do
    n <- getCurrentTimeStep
    let goalGenerations = n + moreGenerations
        checkDone = do
          t <- getCurrentTimeStep
          (msgDebug
           (printf
            "Checking current generation number %d against goal %d"
            t goalGenerations))
          when (goalGenerations <= t) $ liftIO $ throwIO MaxGenerationsReached

    untilTrue $ do
      quickReport
      writeStatus
      checkDone
      checkSig
      fullProblemNextGeneration populationSpec problem
      return False
  return ()



doCmd CommonOpts{..} Query{..} = do
  let environmentDirectory = samplesDir </> sampleName
      databaseSpec = DatabaseSpec{..}
      configFile = samplesDir </> "Config.yaml"
      genomeFile = samplesDir </> "Genome.yaml"
      problemFile = samplesDir </> "Problem.yaml"
  assertDirectoryExists samplesDir
  config@Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture

  aMap <- withDatabase databaseSpec $ withPopCacheX genomeSpec problem $ do
    doSelect select

  let rMap = fmap (queryAgent genomeSpec problem request) aMap
  writeReports genomeSpec problem outputFormat outputDestination rMap


initializeFirstGeneration ::
  (UsefulProblem prob,
   UsefulGenomeSpec gs) =>
  PopulationSpec gs
  -> prob
  -> [Genome gs]
  -> [Word32]
  -> CompoundMPredicate ArchivePredicate
  -> IO ()
initializeFirstGeneration popSpec@PopulationSpec{..}
  problem genomes randomSeed archivePredicate =
  let
    problemName = generalName problem
    recordGenerations =
      case archivePredicate of
        Never -> False
        _ -> True
    initRec = InitializationRec{..}
    initRating = minLandmarkRating problem
    newEggsRatings =
      [(Egg [] g, rateGenome genomeSpec problem g) | g <- genomes]
    delta = PopulationBirthDeathDelta{toKill = [], ..}
  in do
    initializePopulationDatabase databaseSpec initRating initRec
    withDatabase databaseSpec $ withPopCacheX genomeSpec problem $ do
      applyBirthDeathDelta delta
      nextTime


data ImportException =
  ImportException String
  deriving (Show, Eq, Typeable, Data)

instance Exception ImportException

-- | The problem argument is a dummy to get the types to work out
checkImportGenomes ::
  forall gs prob.
  (UsefulGenomeSpec gs,
   UsefulProblem prob) =>
  gs -> prob -> BS.ByteString -> [Genome gs]
checkImportGenomes gs problem bytes =
  if null errors
  then successes
  else
    throw (ImportException (head errors))
  where
    exports :: Map Integer (R.AgentResultX gs prob)
    exports = case Yaml.decodeEither' bytes of
      Left exc ->
        throw exc
      Right exports ->
        exports
    parseAttempts =
      [JT.parseEither (importGenomeDataParser gs) genome
        | R.Export{..} <- Map.elems exports]
    errors = lefts parseAttempts
    successes = rights parseAttempts


-- | Exceptions thrown when a run reaches is target.
data TargetAchievedException =
  MaxGenerationsReached
  | LandmarkBornReached
  | LandmarkDiedReached
  deriving (Read, Show, Eq, Data, Typeable, Generic)

instance Exception TargetAchievedException

handleTargetAchievedEx :: TargetAchievedException -> IO ()
handleTargetAchievedEx ex =
  msgInfo $ case ex of
    MaxGenerationsReached ->
      "Reached maximum number of generations"
    LandmarkBornReached ->
      "Requested landmark agent has been born"
    LandmarkDiedReached ->
      "Requested landmark agent has died"


-- | Throw a 'TargetAchievedException' if one of the targets has
-- been achieved.
checkTarget ::
  forall gs prob.
  (UsefulGenomeSpec gs,
   UsefulProblem prob) =>
  gs
  -> prob
  -> Target
  -> PopMonad (Genome gs) (Rating prob) ()

checkTarget genomeSpec problem Target{..} = do

  case maxGenerations of
    Nothing -> return ()
    Just tLast -> do
      t <- getCurrentTimeStep
      (msgDebug
       (printf
        "Checking current generation number %d against max generations %d"
        t tLast))
      when (tLast <= t) $ liftIO $ throwIO MaxGenerationsReached

  case landmarkBorn of
    Nothing -> return ()
    Just targetRatingV ->
      case fromJSON targetRatingV of
        J.Success targetRating ->
          withLandmarkRatedMoreThan targetRating $ \agent -> do
            liftIO $ throwIO $ LandmarkBornReached
        J.Error msg ->
            liftIO $ throwIO $ BadRating {
              badRating = targetRatingV,
              parseErrorMessage = msg
              }

  case landmarkDied of
    Nothing -> return ()
    Just targetRatingV ->
      case fromJSON targetRatingV of
        J.Success targetRating ->
          withLandmarkRatedMoreThan targetRating $ \agent -> do
            case lifespan agent of
              Left _ -> return ()
              Right (_, deathTime) ->
                liftIO $ throwIO $ LandmarkDiedReached
        J.Error msg ->
            liftIO $ throwIO $ BadRating {
              badRating = targetRatingV,
              parseErrorMessage = msg
              }

withLandmarkRatedMoreThan targetRating axn = do
  landmarks <- getLandmarks
  let mTargetLandmark = find (\x -> rating x >= targetRating) (Map.elems landmarks)
  case mTargetLandmark of
    Nothing -> return ()
    Just agent ->
      axn agent


quickReport ::
  (Show rating) =>
  PopMonad genome rating ()
quickReport = do
  seed <- getCurrentRandomSeed
  t <- getCurrentTimeStep
  n <- getCurrentIdentifier
  r <- getBestRating
  msgInfo ("At time " ++ show t ++ " agent #" ++ show n
           ++ " best rating so far: " ++ show r)
  msgDebug ("random seed: " ++ show (take 4 seed))

writeStatus ::
  (ToJSON rating,
  Binary rating,
  Binary genome) =>
  PopMonad genome rating ()

writeStatus = do
  t <- getCurrentTimeStep
  n <- getCurrentIdentifier
  r <- getBestRating
  lm <- getLandmarks
  now <- liftIO getCurrentTime
  let (newestLandmarkID, Agent{..}) = Map.findMax lm
  let rObj = object [
        "generation" .= t,
        "nextAgentID" .= n,
        "bestRatingSoFar" .= r,
        "newestLandmark" .= (newestLandmarkID, rating),
        "timeStamp" .= toJSON now
        ]
  (d :: Database TableName) <- lift get
  let Database{dbSpec = DatabaseSpec{environmentDirectory} } = d
  liftIO $ Yaml.encodeFile (environmentDirectory </> "Status.yaml") rObj





-- | Advance to the next generation

fullProblemNextGeneration ::
  (UsefulGenomeSpec gs,
   UsefulProblem prob)
  =>
  PopulationSpec gs
  -> prob
  -> PopMonad (Genome gs) (Rating prob) ()

fullProblemNextGeneration PopulationSpec{..} problem =
  (nextGeneration
   selectionSpec
   genomeSpec
   (rateGenome genomeSpec problem))


