{-# LANGUAGE RecordWildCards, NamedFieldPuns, DeriveDataTypeable, DeriveGeneric, ScopedTypeVariables, ConstraintKinds, FlexibleContexts, OverloadedStrings #-}

{-|
Module      : EngineUtil

Description : Various loading, checking, and handling utilities for Engine and related executables.

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module EngineUtil where

import GHC.Generics

import Control.Concurrent
import Control.DeepSeq
import Control.Exception
import Control.Monad
import Control.Monad.State
import Control.Monad.Trans

import Data.Aeson hiding (Array)
import qualified Data.Aeson as J
import qualified Data.Aeson.Types as JT
-- import Data.Binary
import qualified Data.ByteString as BS
import Data.ByteString.Builder
import qualified Data.ByteString.Lazy as BSL
import qualified Data.ByteString.Lazy.UTF8 as LU
import Data.Data
import Data.Default
import Data.Either
import Data.List (find)
import Data.Map (Map)
import qualified Data.Map as M
import Data.Time.Clock
import Data.Time.Format
import Data.Time.LocalTime
import Data.Typeable
import Data.Word
import qualified Data.Yaml as Yaml
import Options.Applicative
import Options.Applicative.Arrows
import System.Directory
import System.Environment
import System.Exit
import System.FilePath
import System.IO
import qualified System.Log.Handler as LH
import qualified System.Log.Handler.Simple as LHS
import System.Log.Formatter
import System.Log.Logger
import System.Posix.Signals
import Text.Printf

import Agent
import Database
import Evolution
import GeneralName
import Genome
import Job
import Logging
import PopulationDatabase
import PopulationSpec
import PredicateTools
import Problem
import ProblemInventory
import Query
import qualified QueryRequest as Q
import qualified QueryResult as R
import Report
import ReportFormatTypes
import MonadUtil
import RandomST

-- | Make a log handler that sends messages to a 'Handle'
logToHandle :: Handle -> Priority -> IO (LHS.GenericHandler Handle)
logToHandle dest logLevel = do
  h0 <- LHS.streamHandler dest logLevel
  let h1 = LH.setFormatter h0
           (simpleLogFormatter "$prio $time $loggername:\n>> $msg")
      h = LH.setLevel h1 logLevel
  return h

-- | Make a log handler that sends messages to a file.
logToFile :: FilePath -> Priority -> IO (LHS.GenericHandler Handle)
logToFile path logLevel = do
  dest <- openFile path AppendMode
  hSetBuffering dest LineBuffering
  h <- logToHandle dest logLevel
  return h{LHS.closeFunc = hClose}

-- | Add a handler to the global logger that sends messages to a
-- file.
sendLogToFile :: FilePath -> Priority -> IO ()
sendLogToFile path logLevel = do
  h <- logToFile path logLevel
  updateGlobalLogger rootLoggerName $ addHandler h

handleConfigurationEx :: ConfigurationException -> IO any
handleConfigurationEx ex = do
  msgError "Error while interpreting configuration:"
  case ex of
    UnknownObject{..} ->
      msgError $ printf "Unknown %s requested: %s" objectType objectName
    DoNotOverwriteFile{..} ->
      msgError $ printf "File %s exists but shouldn't be overwritten" path
    FileDoesNotExist{..} ->
      msgError $ printf "File %s does not exist" path
    DirectoryDoesNotExist{..} ->
      msgError $ printf "Directory %s does not exist" path
    FileIsEmpty{..} ->
      msgError $ printf "File %s exists but is empty" path
    BadRating{..} ->
      (msgError
       (printf "Unable to interpret JSON object as rating\n%s\n%s"
        (LU.toString $ J.encode badRating)
        parseErrorMessage))
    ConflictingSettings{..} ->
      msgError ("Conflicting settings: " ++ message)

  exitWith (ExitFailure 10)


-- | Exception signalling a problem reading the configuration
data ConfigurationException =
  -- | An unknown object was requested
  UnknownObject {
    objectType :: String, -- ^ type (problem, genome architecture, etc)
    objectName :: String -- ^ name of requested object
    }
  |
  -- | A file exists that would be overwritten
  DoNotOverwriteFile {
    path :: FilePath -- ^ path to a file that exists but shouldn't
    }
  |
  -- | A file that needs to be read does not exit
  FileDoesNotExist {
    path :: FilePath -- ^ path to a file that should exist but doesn't
    }
  |
  -- | A directory that needs to be read does not exit
  DirectoryDoesNotExist {
    path :: FilePath -- ^ path to a directory that should exist but doesn't
    }
  |
  -- | A file that needs to be read is empty
  FileIsEmpty {
    path :: FilePath -- ^ path to a file that exists but is empty
    }
  |
  -- | Could not interpret a rating
  BadRating {
    badRating :: Value, -- ^ JSON value that could not be parsed
    parseErrorMessage :: String
    }
  |
  -- | Conflicting settings
  ConflictingSettings {
    message :: String
    }
  deriving (Show, Eq, Typeable, Data)

instance Exception ConfigurationException


checkedLookup :: Map String t -> String -> String -> IO t
checkedLookup inventory objectType name = do
  when (not (M.member name inventory)) $ do
    throwIO (UnknownObject objectType name)
  return (inventory M.! name)


getProblemTemplate :: String -> IO (Box UsefulProblem)
getProblemTemplate = checkedLookup problemTemplates "problem"

getGenomeSpecTemplate :: String -> IO (Box UsefulGenomeSpec)
getGenomeSpecTemplate = checkedLookup genomeSpecTemplates "genome architecture"

assertDoNotOverwrite :: MonadIO m => FilePath -> m ()
assertDoNotOverwrite path = liftIO $ do
  x <- doesFileExist path
  when x $ throwIO DoNotOverwriteFile{..}


assertFileExists :: MonadIO m => FilePath -> m ()
assertFileExists path = liftIO $ do
  x <- doesFileExist path
  unless x $ throwIO FileDoesNotExist{..}

assertDirectoryExists :: MonadIO m => FilePath -> m ()
assertDirectoryExists path = liftIO $ do
  x <- doesDirectoryExist path
  unless x $ throwIO DirectoryDoesNotExist{..}

assertFileNotEmpty :: MonadIO m => FilePath -> m ()
assertFileNotEmpty path = liftIO $ do
  assertFileExists path
  n <- withFile path ReadMode $ \h -> do
    hFileSize h
  when (n == 0) $ throwIO FileIsEmpty{..}

checkedLoadYAML ::
  (MonadIO m,
   FromJSON t) =>
  FilePath -> m t
checkedLoadYAML filePath = liftIO $ do
  assertFileNotEmpty filePath
  readYAMLFile filePath

checkedLoadObject ::
  forall c t m. (FromJSON t, c t, MonadIO m)
  =>
  FilePath
  -> Impl c t
  -> m (Box c)
checkedLoadObject filePath Impl = liftIO $ do
  assertFileNotEmpty filePath
  (p :: t) <- readYAMLFile filePath
  return (Box p)

readYAMLFile :: FromJSON t => FilePath -> IO t
readYAMLFile filePath = do
  mp <- Yaml.decodeFileEither filePath
  case mp of
    Left pError ->
      throwIO pError
    Right p ->
      return p

loadProblem :: FilePath -> String -> IO (Box UsefulProblem)
loadProblem filePath name = do
  (withNamedObject
   problemInventory
   (checkedLoadObject filePath)
   (throwIO (UnknownObject "problem" name))
   name)

loadGenomeSpec :: FilePath -> String -> IO (Box UsefulGenomeSpec)
loadGenomeSpec filePath name = do
  (withNamedObject
   genomeArchitectureInventory
   (checkedLoadObject filePath)
   (throwIO (UnknownObject "genome architecture" name))
   name)


-- | Exceptions thrown when a run hits some constraint, such as
-- running out of time.
data RunConstraintException =
  ExternalStopSignal
  | MaxSecondsElapsed
  deriving (Read, Show, Eq, Data, Typeable, Generic)

instance Exception RunConstraintException


-- | Install signal handlers.  When a signal is received, set a
-- flag (to stop gracefully) and remove this signal handler, so
-- the next signal of this type is likely to cause the program to
-- end immediately.  Make an action called a /sig checker/ that
-- throw 'ExternalStopSignal' if the flag has been set.
makeSigStopper ::
  (MonadIO m) =>
  IO (m ())
makeSigStopper = do
  stopper <- newMVar False
  let sigs = [sigINT, sigHUP, sigUSR1, sigTERM]
      stop sig = do
        msgInfo (printf "Caught signal %s; calling for stop" (show sig))
        v <- swapMVar stopper True
        installHandler sig Default Nothing
        return ()

  forM_ sigs $ \sig ->
    installHandler sig (Catch (stop sig)) Nothing

  return $ liftIO $ do
    b <- readMVar stopper
    when b (throwIO ExternalStopSignal)


-- | Given a 'Maybe' and an action, if the 'Maybe' object is
-- 'Just x', perform the action, otherwise do a do-nothing
-- action.  This is used to build checker actions out of
-- configuration that may or may not specify a contstraint to
-- check.

maybeHelperIO ::
  (MonadIO m) =>
  Maybe t
  -> (t -> IO (m ()))
  -> IO (m ())
maybeHelperIO m axn =
  case m of
    Nothing -> return (return ())
    Just x ->
      axn x

-- | Make an action called a /clock checker/ that throws
-- 'MaxSecondsElapsed' if the current time is more than
-- @maxClockSeconds@ from when the checker was made.

makeClockChecker ::
  (MonadIO m, Integral a) =>
  a -- ^ @maxClockSeconds@
  -> IO (m ())

makeClockChecker maxClockSeconds = do
  startTime <- getCurrentTime
  let stopTime = addUTCTime (fromIntegral maxClockSeconds) startTime
  stopTimeInLocalZone <- utcToLocalZonedTime stopTime
  msgInfo ("Running until " ++ formatTime defaultTimeLocale "%c" stopTimeInLocalZone)
  return $ liftIO $ do
    now <- getCurrentTime
    when (now >= stopTime) (throwIO MaxSecondsElapsed)


-- | Encode an object as YAML to the given file path, using
-- 'assertDoNotOverwrite' to avoid overwriting an existing file.
yamlEncodeFileSafe :: ToJSON a => FilePath -> a -> IO ()
yamlEncodeFileSafe path obj = do
  assertDoNotOverwrite path
  Yaml.encodeFile path obj


-- | This does exactly the same thing as 'withPopCache' but by
-- accepting a 'UsefulGenomeSpec' and a 'UsefulProblem' as dummy
-- arguments, the type checking works a lot better, no silly
-- ambiguities.
withPopCacheX ::
  (UsefulGenomeSpec gs,
   UsefulProblem prob) =>
  gs -> prob -> PopMonad (Genome gs) (Rating prob) c -> DBMonad TableName c

withPopCacheX genomeSpec problem axn =
  withPopCache axn

