{-# LANGUAGE RecordWildCards, NamedFieldPuns, RankNTypes, FlexibleInstances, MultiParamTypeClasses, FlexibleContexts, UndecidableInstances, ScopedTypeVariables, DeriveDataTypeable, DeriveGeneric, TemplateHaskell #-}

{-|

Module: Evolution

Description : Abstractions for representing a deterministic-rating-based evolutionary problem

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : POSIX


This is for some general functions for dealing with the
selection-mutation process.

An interesting detail: The 'SelectFromBlocks' are processed in
order, and I generally list the most restrictive selection blocks
first (only affects the "weak" protocol so far).  That means
those agents have lower identifiers, and are considered older
than agents bred from less restrictive blocks.  That means the
high scoring children from the most restrictive block are at a
slight disadvantage as they are less likely to be among survivors
in future generations.  This might exaggerate the
innovation-via-outliers effect.

-}

module Evolution (
  SelectFromBlock(..),
  SelectionSpec(..),
  EvaluationException(..),
  resultingPopulationSize,
  nextGeneration,
  rateGenome
  ) where

import GHC.Generics

import Control.DeepSeq
import Control.Exception

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.Default
import Data.DeriveTH
import Data.Foldable
import Data.Map (Map)
import qualified Data.Map as M
import Data.Sequence (Seq, ViewL(..), ViewR(..))
import qualified Data.Sequence as Seq
import Data.Typeable
--import Debug.Trace
--import Text.Printf

import AesonUtil
import Agent
import Instruction
import Compiler
import Genome
import Logging
import PopulationDatabase
import Problem
import RandomST
import Util

evolution_vcid :: String
evolution_vcid = "$Id: Evolution.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"

-- | Counts for how many agents from the top of the population
-- list to breed from, and how many offspring to generate.

data SelectFromBlock =
  UniformBlock {
    numOffspring :: !Int,
    -- ^ generate this many offspring
    numToBreed :: !Int
    -- ^ breed from this many of the top rated individuals
    }
  | Tournament {
    numOffspring :: !Int,
    -- ^ generate this many offspring
    pTakeHigher :: !Double
    -- ^ pick pairs and keep the higher-rated individual with this probability
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

$(derive makeNFData ''SelectFromBlock)
$(derive makeBinary ''SelectFromBlock)
$(deriveJSON myJSONOpts ''SelectFromBlock)

data SelectionSpec =
  SelectionSpec {
    selectFromBlocks :: [SelectFromBlock],
    -- ^ breed according to these blocks
    numToKeep :: Int
    -- ^ keep this many top rated individuals from the original
    -- population
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

-- | This is weak selection
instance Default SelectionSpec where
  def = SelectionSpec{..}
    where
      selectFromBlocks =
        [UniformBlock {
            numOffspring = 200,
            numToBreed = 100},
         UniformBlock {
            numOffspring = 100,
            numToBreed = 500}]
      numToKeep = 200

$(derive makeNFData ''SelectionSpec)
$(derive makeBinary ''SelectionSpec)
$(deriveJSON myJSONOpts ''SelectionSpec)


resultingPopulationSize SelectionSpec{..} =
  numToKeep + Prelude.sum (map numOffspring selectFromBlocks)

-- Agent.hs defines
-- Sample (gs, genome, genome) => (Sample (gs, Agent, Agent) Egg)

instance
  (Sample (gs, genome, genome) genome,
   Show rating) =>
  Sample
  (SelectFromBlock, gs, Seq (Agent genome rating))
  [Egg genome] where

  pullSample (UniformBlock {..}, gs, agentsSorted) =
    pullSampleList numOffspring (gs, Seq.take numToBreed agentsSorted)

  pullSample (Tournament {..}, gs, agentsSorted) =
    pullSampleList numOffspring (TournamentPair pTakeHigher, gs, agentsSorted)

data TournamentPair = TournamentPair Double

instance Sample (TournamentPair, Seq x) x where

  pullSample (TournamentPair p, xs) = do
    let n = Seq.length xs
    j1 <- pullSample (UniformDistribution (0, n - 1))
    j2 <- pullSample (UniformDistribution (0, n - 1))
    q <- pullSample (BernoulliDistribution p)
    let k = if q then min j1 j2 else max j1 j2
        -- j1 < j2 means rating a1 >= rating a2
    return (Seq.index xs k)

instance
  (Sample (gs, genome, genome) genome,
   Show rating) =>
  Sample
  (TournamentPair, gs, Seq (Agent genome rating))
  (Egg genome) where

    pullSample (t, gs, agentsSorted) = do
      a1 :: Agent genome rating <- pullSample (t, agentsSorted)
      a2 :: Agent genome rating <- pullSample (t, agentsSorted)
      e :: Egg genome <- pullSample (gs, a1, a2)
      return e

-- | Pull a sample of new genomes and such from a parent
-- population, using a deterministic genome rating function.

instance (Sample (gs, genome, genome) genome,
          Ord genome,
          Ord rating,
          Show rating) =>
         Sample
         (SelectionSpec,
          gs,
          genome -> rating,
          Map Integer (Agent genome rating))
         (PopulationBirthDeathDelta genome rating) where

  pullSample (SelectionSpec {..}, gs, rateGenome,
              cachedLivingAgents) =
    let
      rateEgg e = rateGenome (genome e)

      -- new improved version
      agentsSorted = Seq.sort $ Seq.fromList $ M.elems cachedLivingAgents
      toKill = reverse $ map identifier $ toList
               $ Seq.drop numToKeep agentsSorted

    in do
      newEggss <- (mapM
                   (\ sfb ->
                     pullSample (sfb, gs, agentsSorted))
                   selectFromBlocks)
      let newEggs = Prelude.concat newEggss
      let newEggsRatings = keepMap rateEgg newEggs
      return PopulationBirthDeathDelta{..}


rateGenome ::
  (GenomeSpec gs,
   -- ToJSON (Genome gs),
   Problem problem,
   Num (Rating problem),
   NFData (Input problem),
   NFData (Output problem),
   NFData (Rating problem))
  =>
  gs -> problem -> Genome gs -> Rating problem

rateGenome gs problem genome =
  -- trace ("Rating genome " ++ take 40 (show $ toJSON genome)) $
  rateColonyAdd problem program
  where
    instructions = interpretGenome gs genome
    program =
      case compile (getInstructionSpec gs) instructions of
        Left errors ->
          throw (CompileErrors instructions errors)
        Right progAndInfo ->
          progAndInfo


-- | This is the main selection-mutation process step.

nextGeneration ::
  (Binary genome,
   Ord genome,
   Sample (gs, genome, genome) genome,
   NFData genome,
   Show genome,
   Typeable genome,
   Binary rating,
   Ord rating,
   NFData rating,
   Show rating,
   Typeable rating) =>
  SelectionSpec
  -> gs
  -> (genome -> rating)
  -> PopMonad genome rating ()

nextGeneration selectionSpec genomeSpec ratingFunction = do
  msgDebug "Evolution.nextGeneration begin"
  cachedLivingAgents <- getLivingAgents
  bdDelta <- (withCurrentRandomSeed $
              pullSample (selectionSpec, genomeSpec,
                          ratingFunction, cachedLivingAgents))
  msgDebug "Evolution.nextGeneration birth-death-delta built"
  applyBirthDeathDelta bdDelta
  msgDebug "Evolution.nextGeneration birth-death-delta applied"
  nextTime
  msgDebug "Evolution.nextGeneration done"


data EvaluationException =
  CompileErrors [Instruction] [CompileError]
  deriving (Read, Show, Data, Typeable, Generic)

instance Exception EvaluationException
