{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables, FlexibleContexts, MultiParamTypeClasses, FlexibleInstances, ParallelListComp, TypeFamilies, DeriveDataTypeable, OverloadedStrings, TemplateHaskell, DeriveGeneric #-}

{-|

Module: FPCopyProblem

Description : Problem for copying bits through a narrow synapse

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Note: This is mostly the same as "CopyProblem" except that the
rating is a floating point number, hence the FP.

Here's the traditional set up I've been using, although the
problem is configurable in many ways:

There are two cells. Cell 1 is the sender and gets role input on
@A[1]@.  Cell 2 is the receiver and gets role input on @A[2]@.
This was originally specialized to the Copy 2 1 problem, but it's
now a lot more general.

Colonies are allowed a maximum number of time steps (normally
100).  The min time steps parameter is normally 20 and is used
only for scoring.  There is no bonus for finishing faster than
that.

The problem includes a beacon feature, in which points can be
awarded for emitting a beacon signal from the receiver.

Normal configuration:

Synapse channels: 3 --> 4, etc.

Output channels:
@
31 --> -2 (beacon)
0 --> -1 (stop signal)
12 --> 0
13 --> 1 etc.
@

Input channels:
@
0 --> 8
1 --> 9 etc.
@

All thresholds are set at 10 for channels.

Scoring: Each colony is presented with each possible input word.
One correctness point is given for each output bit that matches
the corresponding input bit.  The total correctness score is the
number of correctness points multiplied by the max time steps
(normally 100) and the correctness factor (normally 100) (for a
net factor of 10000).  The timing score is /max time steps minus
(actual time - min time steps)/, totaled over all input words.

For each input word, if the beacon signal is emitted, the score
includes 'beaconPoints' extra.  The default value of
'beaconPoints' is 0, which effectively disables this feature.

-}

module FPCopyProblem where

import GHC.Generics

import Control.Applicative
import Control.DeepSeq
import Control.Monad

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.DeriveTH
import Data.Default
import qualified Data.IntMap as IM
import qualified Data.Foldable as F
import Data.Monoid
import qualified Data.Sequence as Seq
import Data.Sequence (Seq, ViewL(..), ViewR(..), viewl, viewr)
import Data.String
import qualified Data.Text as T
import Data.Typeable
import Text.Pandoc.Builder
import Text.Printf

import Debug.Trace

import AesonUtil
import Colony
import GeneralName
import Problem
import Snip
import Util


fpCopyProblem_vcid :: String
fpCopyProblem_vcid = "$Id: FPCopyProblem.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"

-- | Colony consists of a sender and a receciver.
--
-- Accept input bit @j@ on @A[j+'inputWordIndexToPattern']@ in sender
--
-- Emit stop as -1 when @A[0] ≥ 'stopThreshold'@ in receiver
--
-- Emit beacon as -2 when @A[`beaconPattern`] ≥
-- `beaconThreshold`@ in receiver
--
-- Set output bit j when @A[j+'outputWordIndexToPattern'] ≥
-- 'outputThreshold'@ in receiver
--
-- Synapse connects sender patterns 'synapseInputIndexToPattern'...
-- to receciver patterns 'synapseOutputIndexToPattern'...

data FPCopyProblem = FPCopyProblem {
  wordSize :: Int,
  synapseSize :: Int,
  minSteps :: Int,
  maxSteps :: Int,
  correctnessFactor :: Double,
  inputWordIndexToPattern :: Int,
  -- ^ add this to an index into the input word to get the
  -- pattern in the sender that gets incremented
  inputActivation :: Int,
  -- ^ amount added to an input pattern in the sender when the
  -- corresponding bit in the input word is on
  synapseInputIndexToPattern :: Int,
  -- ^ add this to a bit index into the synaptic word to get the
  -- pattern in the sender that gets read
  synapseOutputIndexToPattern :: Int,
  -- ^ add this to a bit index into the synaptic word to get the
  -- pattern in the receiver that gets incremented
  synapseThreshold :: Int,
  -- ^ the synapse fires when the activation of a synapse pattern
  -- in the sender reaches this threshold
  synapseCost :: Int,
  -- ^ the activation subtracted when the synapse spikes
  synapseDelay :: Int,
  -- ^ how many time steps it takes for a spike to reach the
  -- receiver
  synapseActivation :: Int,
  -- ^ the activation added to a pattern in the receiver when the
  -- synapse spikes
  outputWordIndexToPattern :: Int,
  -- ^ add this to an index into the output word to get the
  -- pattern in the receiver that gets read
  outputThreshold :: Int,
  -- ^ an output bit is set on when the activation of an output
  -- pattern in the receiver reaches this threshold
  outputCost :: Int,
  -- ^ the activation subtracted when an output bit is set
  beaconPattern :: Int,
  -- ^ this pattern is read for the beacon signal
  beaconThreshold :: Int,
  -- ^ in the receiver cell, if @A[`beaconPattern`] ≥
  -- `beaconThreshold`@ then the beacon signal is emitted
  beaconCost :: Int,
  -- ^ amount to subtract from @A[`beaconPattern`]@ in the
  -- receiver cell when the beacon signal is emitted
  beaconPoints :: Double,
  -- ^ points awarded for emitting the beacon signal
  minReactions :: Double,
  -- ^ no cost for averaging up to this many reactions per time step
  reactionCostFactor :: Double,
  -- ^ an energy cost; multiply this by ((the number of reactions
  -- used divided by the number of time steps) less
  -- 'minReactions') and subtract that from the total score
  minTotalDeltas :: Double,
  -- ^ no cost for averaging up to this total of reaction deltas
  -- per time step
  deltaCostFactor :: Double,
  -- ^ an energy cost; multiply this by ((the total of all
  -- reaction deltas applied divided by the number of time steps)
  -- less 'minTotalDeltas') and subtract that from the total score
  roleActivation :: Int
  -- ^ how much activation is added to @A[1]@ in the sender and
  -- @A[2]@ in the receiver on each time step to indicate their
  -- roles.
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

$(derive makeNFData ''FPCopyProblem)
$(derive makeBinary ''FPCopyProblem)
$(deriveJSON myJSONOpts ''FPCopyProblem)

-- | The default is the @Copy21@ problem with no beacon cost and
-- no energy costs.
instance Default FPCopyProblem where
  def = FPCopyProblem {
    wordSize = 2,
    synapseSize = 1,
    minSteps = 20,
    maxSteps = 100,
    correctnessFactor = 100,
    inputWordIndexToPattern = 8,
    inputActivation = 1,
    synapseInputIndexToPattern = 3,
    synapseOutputIndexToPattern = 4,
    synapseThreshold = 10,
    synapseCost = 0,
    synapseDelay = 0,
    synapseActivation = 1,
    outputWordIndexToPattern = 12,
    outputThreshold = 10,
    outputCost = 0,
    beaconPattern = 31,
    beaconThreshold = 10,
    beaconCost = 0,
    beaconPoints = 0,
    minReactions = 0,
    reactionCostFactor = 0,
    minTotalDeltas = 0,
    deltaCostFactor = 0,
    roleActivation = 1
    }

instance GeneralName FPCopyProblem where
  generalName _ = "FPCopy"

instance Problem FPCopyProblem where

  -- | Interpreted as a list of inputs bits set
  type Input FPCopyProblem = [Int]

  -- | Intepreted as (time steps, outputs set when stopped)
  type Output FPCopyProblem = (Int, [Int])

  type Rating FPCopyProblem = Double

  initialColonySpec FPCopyProblem{..} =
    let
      numCells = 2
      sender = 0
      receiver = 1

      cellNames = IM.fromList [
        (sender, "Sender"),
        (receiver, "Receiver")]

      -- Background input
      -- Sender gets input into 1
      -- Receiver gets input into 2
      bgInput = IM.fromList [
        (-1,
         ("S", sender,
          ReceptorSpec{targetPattern = 1, targetDelta = roleActivation})),
        (-2,
         ("R", receiver,
          ReceptorSpec{targetPattern = 2, targetDelta = roleActivation}))
        ]

      -- One input item for each bit in the input word
      inputSpecs = bgInput `IM.union` IM.fromList [
        (j, (show j, sender, ReceptorSpec{
                targetPattern = inputWordIndexToPattern + j,
                targetDelta = inputActivation}))
        | j <- [0 .. wordSize - 1]]

      -- One synapse object for each bit in the communication
      -- channel
      synapseSpecs = [
        (sender, receiver,
         SynapseSpec{
           emitterSpec =
              EmitterSpec{
                sourcePattern = synapseInputIndexToPattern + j,
                sourceThreshold = synapseThreshold,
                activationDelta = -synapseCost },
           delay = synapseDelay,
           receptorSpec =
             ReceptorSpec{
               targetPattern = synapseOutputIndexToPattern + j,
               targetDelta = synapseActivation }})
        | j <- [0 .. synapseSize - 1]]


      stopSignal =
        ("Stop",
         receiver,
         EmitterSpec{
           sourcePattern = 0,
           sourceThreshold = outputThreshold,
           activationDelta = -outputCost},
         -1)

      beaconSignal =
        ("Beacon",
         receiver,
         EmitterSpec{
           sourcePattern = beaconPattern,
           sourceThreshold = beaconThreshold,
           activationDelta = -beaconCost},
         -2)

      -- One output item for each bit in the input word
      emitterSpecs =
        stopSignal : beaconSignal : [
        (show j,
         receiver,
         EmitterSpec{
           sourcePattern = outputWordIndexToPattern + j,
           sourceThreshold = outputThreshold,
           activationDelta = -outputCost},
         j)
        | j <- [0 .. wordSize - 1]]

    in
     ColonySpec {..}



  problemInputToColonyInputStream FPCopyProblem{..} input =
    replicate maxSteps ([-1, -2] ++ input)


  colonyOutputsToProblemOutput FPCopyProblem{..} colonyOutputSeq =
    let
      whenStopped = Seq.length colonyOutputSeq

      _ :> lastEmitted = viewr colonyOutputSeq
    in
     (whenStopped, lastEmitted)


  stopCheck FPCopyProblem{..} colonyOutputs =
    -1 `elem` colonyOutputs

  inputsForRating FPCopyProblem{..} =
    allInputBitLists wordSize

  minLandmarkRating _ = 0

  rateResult FPCopyProblem{..} input (whenStopped, output) totalReactionCount totalReactionDeltas =
    let
      correctnessScore =
        countCorrectCopiedBits wordSize input output
      timeScore = maxSteps - (minSteps `max` whenStopped)
      beaconScore =
        if -2 `elem` output
        then beaconPoints
        else 0
      averageReactionCount =
        fromIntegral totalReactionCount / fromIntegral whenStopped
      averageTotalDeltas =
        fromIntegral totalReactionDeltas / fromIntegral whenStopped
      reactionCountCost =
        max 0 (reactionCostFactor * (averageReactionCount - minReactions))
      reactionDeltaCost =
        max 0 (deltaCostFactor * (averageTotalDeltas - minTotalDeltas))
    in
     (correctnessFactor * fromIntegral (maxSteps * correctnessScore)
      + fromIntegral timeScore
      + beaconScore
      - reactionCountCost
      - reactionDeltaCost)


-- | (Private) Produce all sublists of 0 .. n-1
allInputBitLists :: Int -> [[Int]]
allInputBitLists n = reverse (allSubLists [n-1, n-2 .. 0])

-- | (Private) Count the number of elements of 0 .. n-1 common to
-- both lists.
countCorrectCopiedBits :: Int -> [Int] -> [Int] -> Int
countCorrectCopiedBits n inputs outputs =
  sum [fromEnum ((j `elem` inputs) == (j `elem` outputs))
      | j <- [0 .. n-1] ]


-- | Convert input to string of 0s and 1s.
instance Sniplet FPCopyProblem [Int] where
  toInlines FPCopyProblem{..} pInput =
    code $ patternIxsTo01s wordSize pInput

-- | Convert output to
instance Sniplet FPCopyProblem (Int, [Int]) where
  toInlines FPCopyProblem{..} (n, pOutput) =
    code
    $ (printf "%s (%d steps)"
       (patternIxsTo01s wordSize pOutput)
       n)


-- | Given a list of indexes, return a string where the
-- characters at those indices are 1 and the rest are 0.
patternIxsTo01s :: Int -> [Int] -> String
patternIxsTo01s nMax ks =
  [ if j `elem` ks then '1' else '0'
  | j <- [0 .. (nMax - 1)]]


