{-# LANGUAGE RecordWildCards, NamedFieldPuns, FlexibleContexts, DeriveDataTypeable, DeriveGeneric, TemplateHaskell #-}

{-|

Module: GraphML

Description: Tools for building GraphML files from StyledNetworks

Copyright: © 2016 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : POSIX

<http://graphml.graphdrawing.org/ GraphML> is a file format read
by the <http://www.yworks.com YED> graph editor.  As well as
basic GraphML, the functions in this module incorporate styling
and other information particular to YED.

To do eventually:

- The styled network data types should probably go in their own
module.

- The most important functions should be re-written as a class

- That way we can convert a genome to a styled network, then
convert that to GraphML or any other format.

-}

module GraphML (
  networkToGraphMLDocument
  ) where

import Debug.Trace

import Text.XML.Light

import StyledNetwork

infix 3 <:>, +>

graphML_vcid :: String
graphML_vcid = "$Id: GraphML.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"

-- | Build an attribute setting

(+>) :: String -> String -> Attr
key +> value = Attr (unqual key) value


-- | Build a qualified name

(<:>) :: String -> String -> QName
pre <:> name =
  blank_name {qPrefix = Just pre, qName = name}


-- | Build a node from an id, a label to display in the node,
-- a shape string, and a color

prettyNode :: Node t => String -> t -> String -> String -> Element
prettyNode nodeId label shapeType color =
  (unode "node"
   (["id" +> nodeId],
    nodeGraphicsData $ shapeNode label shapeType color))


-- | Build an element containing node data as expected by YED.

nodeGraphicsData :: Node ([Attr], t) => t -> Element
nodeGraphicsData dataContent =
  unode "data" (["key" +> "NG"], dataContent)


-- | Build a shape node from a label, a shape string, and a fill
-- color string

shapeNode :: Node t => t -> String -> String -> Element
shapeNode label shapeType color =
  (node ("y" <:> "ShapeNode")
   [node ("y" <:> "NodeLabel") label,
    node ("y" <:> "Fill")
    ["color" +> color,
     "transparent" +> "false"],
    node ("y" <:> "BorderStyle")
    ["color" +> "#000000"],
    node ("y" <:> "Shape")
    ["type" +> shapeType]])


-- | Build an edge given the edge ID, source ID, the targetID, the edge
-- label, and whether the link should be styled to indicate
-- positive or negative action on the target

prettyEdge ::
  (Node t,
   Node ([Attr], t))
  =>
  String -> String -> String -> Maybe t -> Bool -> Element
prettyEdge edgeId sourceId targetId mLabel positive =
  (unode "edge"
   (["id" +> edgeId,
     "source" +> sourceId,
     "target" +> targetId],
    edgeGraphicsData $ polyLineEdge mLabel positive))

-- | Build an element containing edge data as expected by YED.
edgeGraphicsData :: Node ([Attr], t) => t -> Element
edgeGraphicsData dataContent =
  unode "data" (["key" +> "EG"], dataContent)


-- | Build a polyline edge from a label and whether it should be
-- styled to indicate positive or negative action on its
-- target.

polyLineEdge ::
  (Node t,
   Node ([Attr], t))
  =>
  Maybe t -> Bool -> Element
polyLineEdge mLabel positive =
  (node ("y" <:> "PolyLineEdge")
   ([node ("y" <:> "LineStyle")
     (if positive
      then ["color" +> "#000000", "type" +> "line", "width" +> "1.0"]
      else ["color" +> "#000000", "type" +> "dashed", "width" +> "1.0"]),
     node ("y" <:> "Arrows")
     (if positive
      then ["source" +> "none", "target" +> "standard"]
      else ["source" +> "none", "target" +> "white_delta"])] -- "t_shape"
    ++
    case mLabel of
      Nothing -> []
      Just label ->
        [node ("y" <:> "EdgeLabel")
         (["backgroundColor" +> "#ffffff"],
          label)]
   ))


-- | Make a directed graph element from an ID and content
digraphElement :: Node ([Attr], t) => String -> t -> Element
digraphElement graphId nodesAndEdgesContent =
  unode "graph"
  (["id" +> graphId,
    "edgedefault" +> "directed"],
   nodesAndEdgesContent)


-- | Wrap the given data in a GraphML document

scaffold :: Element -> Element
scaffold documentContent =
  (unode "graphml"
   (["xmlns" +> "http://graphml.graphdrawing.org/xmlns",
     "xmlns:xsi" +> "http://www.w3.org/2001/XMLSchema-instance",
     "xsi:schemaLocation" +> "http://graphml.graphdrawing.org/xmlns http://www.yworks.com/xml/schema/graphml/1.1/ygraphml.xsd",
     "xmlns:y" +> "http://www.yworks.com/xml/graphml",
     "xmlns:yed" +> "http://www.yworks.com/xml/yed/3"],
    -- These elements declare YED-specific attributes
    [unode "key"
     ["for" +> "node",
      "id" +> "NG",
      "yfiles.type" +> "nodegraphics"],
     unode "key"
     ["for" +> "edge",
      "id" +> "EG",
      "yfiles.type" +> "edgegraphics"],
     documentContent]))



nodeToGraphML StyledNode{..} =
  prettyNode nodeKey nodeLabel shape color
  where
   (shape, color) = nodeStyleToShapeAndColor nodeStyle

nodeStyleToShapeAndColor s =
  case s of
    OrdinaryNode -> ("ellipse", "#ffcc00") -- yellow
    BackgroundInputTargetNode -> ("hexagon", "#ffcccc") -- pink
    InputSourceNode -> ("trapezoid2", "#ccffcc") -- green
    InputTargetNode -> ("trapezoid2", "#ccccff") -- blue
    OutputSourceNode -> ("trapezoid", "#ccccff")
    OutputTargetNode -> ("trapezoid", "#ccffcc")
    SynapseSourceNode -> ("ellipse", "#ddddcc") -- orange
    SynapseTargetNode -> ("ellipse", "#ddccdd") -- purple
    StopNode -> ("octagon", "#ff0000")
    ReducedNode -> ("diamond", "#ffcccc") -- pink

edgeToGraphML StyledEdge{..} =
  (prettyEdge
   edgeKey
   fromKey
   toKey
   (if null edgeLabel then Nothing else Just edgeLabel)
   (edgeStyle == ActivationEdge))

networkToGraphML StyledNetwork{..} =
  (digraphElement
   networkKey
   (concat [map nodeToGraphML nodes,
            map edgeToGraphML edges,
            map embeddedNetworkToGraphML subnetworks]))

embeddedNetworkToGraphML :: StyledNetwork -> Element
embeddedNetworkToGraphML n@StyledNetwork{..} =
  unode "node"
  (["id" +> "NW:" ++ networkKey,
    "yfiles.foldertype" +> "group"],
   [unode "data"
    (["key" +> "NG"],
     node ("y" <:> "ProxyAutoBoundsNode")
     (node ("y" <:> "Realizers")
      (["active" +> "0"],
       [node ("y" <:> "GroupNode")
        [node ("y" <:> "Fill")
         ["color" +> "#f8ecc9"],
         node ("y" <:> "NodeLabel")
         (["alignment" +> "right",
           "autoSizePolicy" +> "node_width",
           "backgroundColor" +> "#404040",
           "textColor" +> "#ffffff",
           "fontSize" +> "16",
           "modelName" +> "internal",
           "modelPosition" +> "t",
           "x" +> "0.0",
           "y" +> "0.0"],
          networkLabel)],
        node ("y" <:> "GroupNode") () ]))),
    networkToGraphML n])

networkToGraphMLDocument :: StyledNetwork -> Element
networkToGraphMLDocument network =
  scaffold (networkToGraphML network)
