{-# LANGUAGE BangPatterns, RecordWildCards, NamedFieldPuns, DisambiguateRecordFields, TypeFamilies, ParallelListComp, FlexibleInstances, MultiParamTypeClasses, DeriveDataTypeable, OverloadedStrings, TemplateHaskell, DeriveGeneric #-}


module HaploidSingletonGenome (
  HaploidSingletonGenomeSpec(..),
  makeDefaultHaploidSingletonGenomeSpec,
  makeDefaultHaploidAlignmentSpec,
  geneSize
  ) where

import GHC.Generics

import Control.DeepSeq
import Control.Exception

import Data.Aeson hiding (Array)
import Data.Aeson.Types hiding (Array)
import Data.Aeson.TH
--import Data.Array
import Data.Binary
import Data.Data
import Data.Default
import Data.DeriveTH
import qualified Data.Text as T
import Data.Typeable

--import Debug.Trace
import Text.Printf

import AesonUtil
import BitSource
import RandomST
import BitString
import qualified RangeMap as R
import ShortShow

import GenomeDiagram
import GeneralName

import Instruction
import Genome
import Chromosome
import SmithWaterman


haploidSingletonGenome_vcid :: String
haploidSingletonGenome_vcid = "$Id: HaploidSingletonGenome.hs,v 93fba847b3a8 2017-01-20 21:13:26Z garrett $"

-- | The exported format is a little-endian string of @0@s and
-- @1@s.

data HaploidSingletonGenomeSpec =
  HaploidSingletonGenomeSpec {
    expansionFactor :: Int, -- ^ how many bits in a codon
    instructionSpec :: InstructionSpec,

    crossoverSpec :: CrossoverSpec,

    mutationSpec :: SingleChromosomeMutationSpec,

    initialNumGenes :: Int
    -- ^ number of genes in an initial genome
    }
  deriving (Read, Show, Eq, Ord, Typeable, Data)

-- | A 'HaploidSingletonGenomeSpec' with an expansion factor of
-- 3, 32 initial genes, and the default 'InstructionSpec'
instance Default HaploidSingletonGenomeSpec where
  def = makeDefaultHaploidSingletonGenomeSpec 3 32 def

-- | Make a 'HaploidSingletonGenomeSpec' from some basic
-- information.  The important computation this does is the block
-- size.
makeDefaultHaploidSingletonGenomeSpec ::
  Int -- ^ expansion factor
  -> Int -- ^ initial number of genes
  -> InstructionSpec
  -> HaploidSingletonGenomeSpec
makeDefaultHaploidSingletonGenomeSpec expansionFactor initialNumGenes instructionSpec =
  HaploidSingletonGenomeSpec {..}
  where
    crossoverSpec = def {crossoverBlockSize = blockSize}
    mutationSpec = def {mutationBlockSize = blockSize}
    blockSize = expansionFactor * bitsPerInstruction instructionSpec


data HGenome =
  Genome {
    chromosome :: Chromosome,
    mutations :: [SingleChromosomeMutation],
    source :: IntervalSource HaploidSingletonGenomeSpec
    }
  deriving (Read, Show, Eq, Ord, Typeable, Data)

data HInterval =
    Interval !Int !Int -- ^ @[begin,end)@ bit indices of part of the chromosome
    deriving (Read, Show, Eq, Ord, Typeable, Data)

data HMutation =
    Mutation SingleChromosomeMutation
    deriving (Read, Show, Eq, Ord, Typeable, Data)

instance GeneralName HaploidSingletonGenomeSpec where
  generalName _ = "Haploid"

instance GenomeSpec HaploidSingletonGenomeSpec where

  type Genome HaploidSingletonGenomeSpec = HGenome
  type Interval HaploidSingletonGenomeSpec = HInterval
  type Mutation HaploidSingletonGenomeSpec = HMutation

  getInstructionSpec = instructionSpec

  genomeLength _ Genome{..} =
    toInteger (chromosomeLength chromosome)

  geneDetails genomeSpec@(HaploidSingletonGenomeSpec {..}) genome@(Genome {..}) =
    let
      codeOrder = (expansionFactor - 1) `quot` 2
      -- gene size in nucleotide bits
      gsNucBits = geneSize genomeSpec
      -- gene size in codon bits
      (gsCodonBits, rCB) = gsNucBits `quotRem` expansionFactor

      nNucBits = chromosomeLength chromosome
      (numGenes, rNG) = nNucBits `quotRem` gsNucBits
      chromosomeCodonBits = bsMajorityDecode codeOrder 0 nNucBits chromosome
    in
      assert (0 == rNG) $
      assert (0 == rCB) $
      [GeneDetails {..}
      | j <- [0 .. numGenes - 1],
        let kNB = j * gsNucBits,
        let interval = Interval kNB (kNB + gsNucBits),
        let nucleotideBits = chromosomeSubstring kNB (kNB + gsNucBits) chromosome,
        let kCB = j * gsCodonBits,
        let codonBits = bsSubstring kCB (kCB + gsCodonBits) chromosomeCodonBits,
        let (instruction, leftovers) = readFromBitList instructionSpec codonBits,
        let geneSource = intervalSource genomeSpec genome interval,
        let instructionEncodingDetails =
             (majorityCodeCompleteEncodingDetails
              expansionFactor
              nucleotideBits
              codonBits
              (partialEncodingDetails instructionSpec codonBits))]

  exportGenomeData _ Genome{..} =
    toJSON (bsTo01String chromosome)

  importGenomeDataParser _ v =
    let
      -- Aeson parser
      binaryChromosomeP = withText "string of bits" $ \t ->
        case bsFrom01StringChecked (T.unpack t) of
          Left eMsg ->
            fail eMsg
          Right bs ->
            return bs

    in do
      c <- binaryChromosomeP v
      return $ Genome {
           chromosome = c,
           mutations = [],
           source = R.RangeMap [((0, bsNumBits c), R.empty)] }


  intervalSource _ Genome{..} (Interval begin end) =
    R.keyIntersection (begin, end) source
    -- error (printf
    --           "HaploidSingletonGenome intervalSource: begin = %d, end = %d\n source = %s"
    --          begin end (show source))

  allPossibleMutations HaploidSingletonGenomeSpec{..} Genome{..} =
    let
      genomeLength = chromosomeLength chromosome
    in
     map Mutation (allPossibleSingleChromosomeMutations mutationSpec genomeLength)

  applyMutations HaploidSingletonGenomeSpec{..} wrappedMutations Genome{..} =
      let
        ms = [m | Mutation m <- wrappedMutations]
        (newChromosome, annotations) = mutateChromosome chromosome ms
        source = R.RangeMap [(r, FromChromosome 0 (Interval pBegin pEnd) flips)
                            | (r, (pBegin, pEnd), flips, _) <- annotations]
      in
       Genome {chromosome = newChromosome, ..}

  intervalsChanged genomeSpec@HaploidSingletonGenomeSpec{..} (Mutation m) Genome{..} =
    let
      (begin, end) = mutationInterval m
      n = chromosomeLength chromosome
    in
     assert (0 <= begin) $
     assert (begin < end) $
     if begin < n
     then [Interval begin (n `min` end)]
     else []

instance R.Range HInterval where

  empty = Interval 0 0

  length (Interval begin end) =
    end - begin

  take n (Interval begin end) =
    let
      (nBegin, nEnd) = R.take n (begin, end)
    in
     Interval nBegin nEnd

  drop n (Interval begin end) =
    let
      (nBegin, nEnd) = R.drop n (begin, end)
    in
     Interval nBegin nEnd


instance R.BoundedRange HInterval where

  bounds (Interval begin end) = (begin, end)

instance ShortShow HInterval where

  shortShow (Interval begin end) =
    printf "[%d,%d)" begin end

instance IntervalCoordinates HInterval where

  intervalSegmentCoordinates (Interval begin end) =
    (fromIntegral begin, fromIntegral end, 0.0)



geneSize :: HaploidSingletonGenomeSpec -> Int
geneSize (HaploidSingletonGenomeSpec {..}) =
  expansionFactor * bitsPerInstruction instructionSpec



-- | Random initialization
instance Sample HaploidSingletonGenomeSpec HGenome where
  pullSample genomeSpec@(HaploidSingletonGenomeSpec {..}) =
    let
      SingleChromosomeMutationSpec {..} = mutationSpec
      nBits = initialNumGenes * geneSize genomeSpec
    in do
      chromosome <- pullSample (UniformBits nBits)
      let
        mutations = []
        source = R.RangeMap [((0, nBits), R.empty)]
        in
       return (Genome {..})



instance Sample
         (HaploidSingletonGenomeSpec, HGenome, HGenome)
         HGenome where

  pullSample (HaploidSingletonGenomeSpec {..}, gLeft, gRight) =
    let
      leftChromosome = chromosome gLeft
      rightChromosome = chromosome gRight
      rightParentLength = chromosomeLength rightChromosome
    in do
      Crossover {..} <- pullSample (crossoverSpec, leftChromosome, rightChromosome)
      mutations <- pullSample (mutationSpec, leftEnd + rightParentLength - rightBegin)
      let
        cCrossed = (chromosomeJoin
                    (chromosomeTake leftEnd leftChromosome)
                    (chromosomeDrop rightBegin rightChromosome))
        (chromosome, intervalMap) = mutateChromosome cCrossed mutations
        leftSources =
          [((rBegin, rBegin + xLength),
            FromChromosome 0 (Interval pBegin xEnd) rFlips)
          | ((rBegin, rEnd), (pBegin, pEnd), flips, chunk) <- intervalMap,
            let xEnd = min leftEnd pEnd,
            let xLength = xEnd - pBegin,
            let rFlips = [n | n <- flips, pBegin + n < leftEnd],
            pBegin < leftEnd]
        rightSources =
          [((rBegin + offset, rEnd),
            FromChromosome 1 (Interval xBegin pEnd) rFlips)
          | ((rBegin, rEnd), (pBegin, pEnd), flips, chunk) <- intervalMap,
            let xBegin = max rightBegin pBegin,
            let offset = xBegin - pBegin,
            let rFlips = [n - offset | n <- flips, pBegin + n >= rightBegin],
            pEnd > rightBegin]
        source = R.RangeMap (leftSources ++ rightSources)
        in
       return (Genome {..})

data HAlignmentSpec =
  AlignmentSpec BlockAlignmentSpec
  deriving (Read, Show, Eq, Data, Typeable, Generic)

data HAlignmentResult =
  AlignmentResult {
    resultSW :: SmithWatermanResult BitString
    } deriving (Show, Typeable, Generic)


instance Alignable HaploidSingletonGenomeSpec where

  type AlignmentSpec HaploidSingletonGenomeSpec = HAlignmentSpec

  type AlignmentResult HaploidSingletonGenomeSpec = HAlignmentResult

  setBlockSize n (AlignmentSpec as) =
    AlignmentSpec (setAlignmentBlockSize n as)

  bestAlignment _ (AlignmentSpec as@(BlockAlignmentSpec {..})) x y =
    let
      resultSW = blockAlignSmithWaterman as (chromosome x) (chromosome y)
    in
     AlignmentResult {..}

  alignmentScore _ _ (AlignmentResult {..}) =
    let
      SmithWatermanResult {..} = resultSW
      z = bestScore
    in
     z

  showAlignment _ (AlignmentSpec (BlockAlignmentSpec {..})) (AlignmentResult {..}) =
    let
      SmithWatermanResult {..} = resultSW
      showBlock b = ' ' : bsTo01String b
      (top, bottom) = prettyShowAlignment (1 + alignmentBlockSize) showBlock workspace
    in
     unlines [top, bottom]

  alignmentMatrix _ (AlignmentSpec as@(BlockAlignmentSpec {..})) genomeAr =
    let
      chromosomeAr = fmap chromosome genomeAr
    in
     fmap bestScore (blockSelfAlignmentMatrix as chromosomeAr)


instance Default HAlignmentSpec where
  def = AlignmentSpec def

makeDefaultHaploidAlignmentSpec ::
  Int -- ^ block size
  -> HAlignmentSpec
makeDefaultHaploidAlignmentSpec blockSize =
  AlignmentSpec (makeDefaultBlockAlignmentSpec blockSize)

$(derive makeBinary ''HaploidSingletonGenomeSpec)
$(derive makeBinary ''HGenome)
$(derive makeBinary ''HInterval)
$(derive makeBinary ''HMutation)

$(derive makeNFData ''HaploidSingletonGenomeSpec)
$(derive makeNFData ''HGenome)
$(derive makeNFData ''HInterval)
$(derive makeNFData ''HMutation)

$(deriveJSON myJSONOpts ''HaploidSingletonGenomeSpec)
$(deriveJSON myJSONOpts ''HInterval)
$(deriveJSON myJSONOpts ''HMutation)
$(deriveJSON myJSONOpts ''HAlignmentSpec)

instance FromJSON (FromChromosome HaploidSingletonGenomeSpec)
instance ToJSON (FromChromosome HaploidSingletonGenomeSpec)

$(deriveJSON myJSONOpts ''HGenome)
