{-# LANGUAGE RecordWildCards, NamedFieldPuns, DeriveDataTypeable, TemplateHaskell, Arrows, DeriveGeneric, StandaloneDeriving #-}

{-|
Module      : Job

Description : Parsing the command line

Copyright   : © 2017 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : POSIX

-}

module Job where

import GHC.Generics

import Control.Applicative
import Control.Exception
import Control.Monad

import Data.Aeson
import Data.Aeson.TH
import Data.Array
import qualified Data.Attoparsec.ByteString as AP
import qualified Data.ByteString.UTF8 as BSU
import Data.Data
import Data.Default
import Data.DeriveTH
import Data.Maybe
import Data.Typeable
import Data.Version
--import Data.Yaml
import Options.Applicative
import Options.Applicative.Arrows
import System.Log.Logger
import Text.Regex.Base
import Text.Regex.PCRE.String (Regex)

-- Generated by cabal
import qualified Paths_Sim as P

import AesonUtil
import ArchivePredicate
import PredicateTools
import Query
import QueryRequest
import ReportFormatTypes

-- This got left out of System.Log.Logger
deriving instance Typeable Priority
deriving instance Data Priority

data CommonOpts =
  CommonOpts {
    logLevel :: Priority
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

data Target =
  Target {
    maxGenerations :: Maybe Integer,
    landmarkBorn :: Maybe Value,
    landmarkDied :: Maybe Value
    }
  deriving (Show, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''Target)


instance Default Target where
  def = Target{..}
    where
      maxGenerations = Nothing
      landmarkBorn = Nothing
      landmarkDied = Nothing


data Config =
  Config {
    problemName :: String,
    genomeArchitecture :: String,
    archive :: CompoundMPredicate ArchivePredicate,
    target :: Target
    } deriving (Show, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''Config)


data Args = Args CommonOpts Command
  deriving (Show, Data, Typeable, Generic)

data Command =
  List
  | Template {
    samplesDir :: FilePath, -- ^ diretory where samples will be kept
    problemTemplateName :: String, -- ^ name of the problem to solve
    genomeArchitectureTemplateName :: String -- ^ name of the genome architecture
    }
  | Init {
    samplesDir :: FilePath, -- ^ diretory where samples will be kept
    sampleName :: String, -- ^ name of the sample (subdirectory)
    importFrom :: FilePath,
    identical :: Bool -- ^ whether all initial genomes should be the same
    }
  | Reset {
    samplesDir :: FilePath, -- ^ diretory where samples will be kept
    sampleName :: String -- ^ name of the sample (subdirectory)
    }
  | Run {
    samplesDir :: FilePath, -- ^ diretory where samples will be kept
    sampleName :: String, -- ^ name of the sample (subdirectory)
    maxSeconds :: Maybe Integer -- ^ run for no more than this many seconds
    }
  | Continue {
    samplesDir :: FilePath, -- ^ diretory where samples will be kept
    sampleName :: String, -- ^ name of the sample (subdirectory)
    moreGenerations :: Integer -- ^ run this many more generations
    }
  | Query {
    samplesDir :: FilePath, -- ^ diretory where samples are kept
    sampleName :: String, -- ^ name of the sample (subdirectory)
    select :: SelectAgents, -- ^ which agents
    request :: AgentRequest, -- ^ what information
    outputFormat :: OutputFormat, -- ^ which format
    outputDestination :: OutputDestination -- ^ where to write it
    }
  deriving (Show, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''Command)


jobOpts :: Parser Args
jobOpts = runA $ proc () -> do
  cmds <- ((asA . hsubparser)
           (command "list"
            (info listOpts
             (progDesc "List available problems, genome architectures, etc."))
            <>
            command "template"
            (info templateOpts
             (progDesc "Create configuration files from templates"))
            <>
            command "init"
            (info initOpts
             (progDesc "Initialize a sample run"))
            <>
            command "reset"
            (info resetOpts
             (progDesc "Reset (reinitialize) a sample run"))
            <>
            command "run"
            (info runOpts
             (progDesc "Run a sample to completion target in Config.yaml"))
            <>
            command "continue"
            (info continueOpts
             (progDesc "Continue a sample for more generations"))
            <>
            command "query"
            (info queryOpts
             (progDesc "Query a sample"))
           )) -< ()
  opts <- asA commonOpts -< ()
  A helper -< Args opts cmds


commonOpts :: Parser CommonOpts
commonOpts =
  (CommonOpts <$>
   (option auto
    (long "log-level"
     <> metavar "LEVEL"
     <> help "Append log messages of this level (ERROR, WARNING, NOTICE, INFO, DEBUG) to an appropriate file. Default is NOTICE."
     <> value NOTICE))
  )

listOpts :: Parser Command
listOpts = runA $ proc () -> do
  returnA -< List

templateOpts :: Parser Command
templateOpts = runA $ proc () -> do
  samplesDir <-
    (asA $ argument str
      (metavar "SAMPLES_DIR"
       <> help "Data directory storing sample runs")) -< ()
  problemTemplateName <-
    (asA $ argument str
     (metavar "PROBLEM"
      <> help "Problem name")) -< ()
  genomeArchitectureTemplateName <-
    (asA $ argument str
     (metavar "ARCHITECTURE"
      <> help "Genome architecture")) -< ()
  returnA -< Template{..}

initOpts :: Parser Command
initOpts = runA $ proc () -> do
  samplesDir <-
    (asA $ argument str
     (metavar "SAMPLES_DIR"
      <> help "Data directory storing sample runs")
    ) -< ()
  sampleName <-
    (asA $ argument str
     (metavar "SAMPLE_NAME"
      <> help "Name of the sample (subdirectory)")
    ) -< ()
  importFrom <-
    (asA $ option str
     (long "import"
      <> metavar "GENOME_FILE"
      <> value ""
      <> help "Import genomes exported from another population")
     ) -< ()
  identical <-
    (asA $ switch
     (long "identical"
      <> help "All genomes are created identical")) -< ()

  returnA -< Init{..}


resetOpts :: Parser Command
resetOpts = runA $ proc () -> do
  samplesDir <-
    (asA $ argument str
     (metavar "SAMPLES_DIR"
      <> help "Data directory storing sample runs")
    ) -< ()
  sampleName <-
    (asA $ argument str
     (metavar "SAMPLE_NAME"
      <> help "Name of the sample (subdirectory)")
    ) -< ()
  returnA -< Reset{..}

runOpts :: Parser Command
runOpts = runA $ proc () -> do
  samplesDir <-
    (asA $ argument str
     (metavar "SAMPLES_DIR"
      <> help "Data directory storing sample runs")
    ) -< ()
  sampleName <-
    (asA $ argument str
      (metavar "SAMPLE_NAME"
       <> help "Name of the sample (subdirectory)")
    ) -< ()
  maxSeconds <-
    (asA $ optional $ option auto
      (long "max-seconds"
       <> metavar "SECONDS"
       <> help "Run for no more than this many seconds")
    ) -< ()
  returnA -< Run{..}


continueOpts :: Parser Command
continueOpts = runA $ proc () -> do
  samplesDir <-
    (asA $ argument str
     (metavar "SAMPLES_DIR"
      <> help "Data directory storing sample runs")
    ) -< ()
  sampleName <-
    (asA $ argument str
      (metavar "SAMPLE_NAME"
       <> help "Name of the sample (subdirectory)")
    ) -< ()
  moreGenerations <-
    (asA $ argument auto
      (metavar "GENERATIONS"
       <> help "Run this many more generations")
    ) -< ()
  returnA -< Continue{..}

queryOpts :: Parser Command
queryOpts = runA $ proc () -> do
  samplesDir <-
    (asA $ argument str
     (metavar "SAMPLES_DIR"
      <> help "Data directory storing sample runs")
    ) -< ()
  sampleName <-
    (asA $ argument str
      (metavar "SAMPLE_NAME"
       <> help "Name of the sample (subdirectory)")
    ) -< ()

  -- select agents
  mlandmarks <-
    (asA $ flag Nothing (Just Landmarks)
     (long "landmarks"
      <> help "Query landmark agents (default)")) -< ()
  mbyID <-
    (asA $ optional $ strOption
     (long "by-id"
      <> metavar "IDs"
      <> help "Query agents with these IDs (can be comma or space separated)")) -< ()
  mtop <-
    (asA $ flag Nothing (Just TopRated)
     (long "top-rated"
      <> help "Query top rated living agents")) -< ()
  mratedAtLeastStr <-
    (asA $ optional $ strOption
     (long "rated-at-least"
      <> help "Query living agents with at least this rating (in JSON format)")) -< ()
  mliving <-
    (asA $ flag Nothing (Just Living)
     (long "living"
      <> help "Query all living agents")) -< ()
  -- request information
  msummary <-
    (asA $ flag Nothing (Just Summary)
     (long "summary"
      <> help "Request a summary (default)")) -< ()
  magent <-
    (asA $ flag Nothing (Just JustAgent)
     (long "agent"
      <> help "Request the whole agent structure")) -< ()
  mexport <-
    (asA $ flag Nothing (Just Export)
     (long "export"
      <> help "Request an export record")) -< ()
  mprogram <-
    (asA $ flag Nothing (Just Program)
     (long "program"
      <> help "Request information about the agent's program")) -< ()
  mgenome <-
    (asA $ flag Nothing (Just GenomeDetails)
     (long "genome"
      <> help "Request information about the agent's genome")) -< ()
  mratingInputs <-
    (asA $ flag Nothing (Just RatingInputs)
     (long "rating-inputs"
      <> help "Request summary of inputs used for rating")) -< ()
  mefficacious <-
    (asA $ flag Nothing (Just Efficacious)
     (long "efficacious"
      <> help "Request efficacious analysis of the agent's genome")) -< ()
  mfull <-
    (asA $ flag Nothing (Just FullDetails)
     (long "full"
      <> help "Request full analysis of the agent's genome")) -< ()
  mreactionNetwork <-
    (asA $ flag Nothing (Just ReactionNetwork)
     (long "reaction-network"
      <> help "Request GraphML showing efficacious reactions")) -< ()

  -- output formats
  mtext <-
    (asA $ flag Nothing (Just PlainTextOutput)
     (long "text"
      <> help "Output as plain text (default)")) -< ()
  mmarkdown <-
    (asA $ flag Nothing (Just MarkdownOutput)
     (long "markdown"
      <> help "Output as Markdown")) -< ()
  myaml <-
    (asA $ flag Nothing (Just YAMLOutput)
     (long "yaml"
      <> help "Output as YAML")) -< ()
  mjson <-
    (asA $ flag Nothing (Just JSONOutput)
     (long "json"
      <> help "Output as JSON")) -< ()
  mgraphml <-
    (asA $ flag Nothing (Just GraphMLOutput)
     (long "graphml"
      <> help "Output as GraphML")) -< ()
  overwrite <-
    (asA $ switch
     (long "overwrite"
      <> help "Overwrite existing files")) -< ()

  -- output destinations
  mtoStdout  <-
    (asA $ flag Nothing (Just AllToStdout)
     (long "stdout"
      <> help "Send output to stdout (default)")) -< ()
  mpathTemplate <-
    (asA $ optional $ strOption
     (long "path-template"
      <> metavar "TEMPLATE"
      <> help "Send output for each agent to its own file using the printf TEMPLATE for the file path, which should include a %d for the agent's id")
    ) -< ()
  let mOnePerFile = fmap (ToFilePerAgentId overwrite) mpathTemplate
      mIDs = do
        idsStr <- mbyID
        return (IDs $ parseCommaSeparatedNumbers idsStr)
      mratedAtLeast = do
        s <- mratedAtLeastStr
        case stringToJSONValue s of
          Left msg ->
            throw (MalformedJSONError msg)
          Right v ->
            return (RatedAtleast v)
  returnA -< Query{
    select = mpick Landmarks [mIDs, mlandmarks, mtop, mratedAtLeast, mliving],
    request = mpick Summary [magent, msummary, mexport, mprogram,
                             mgenome, mratingInputs, mefficacious, mfull,
                             mreactionNetwork],
    outputFormat = mpick PlainTextOutput [mtext, mmarkdown, mjson, myaml, mgraphml],
    outputDestination = mpick AllToStdout [mtoStdout, mOnePerFile],
    ..}

mpick :: t -> [Maybe t] -> t
mpick x ms =
  fromMaybe x (msum ms)

jobParser :: ParserInfo Args
jobParser =
  (info jobOpts
   (progDesc
    (concat ["Artificial life simulation engine, version ",
             showVersion P.version,
             "; give --help for more help"])))


--- some odds and ends ---

parseCommaSeparatedNumbers :: String -> [Integer]
parseCommaSeparatedNumbers s =
  let
    intRX :: Regex
    intRX = makeRegex "[0-9]+"
    ms = matchAllText intRX s
    intStrs = [fst (a ! 0) | a <- ms]
  in
   map read intStrs
