{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables, FlexibleContexts, FlexibleInstances, ParallelListComp, TypeFamilies, DeriveDataTypeable, OverloadedStrings, TemplateHaskell, DeriveGeneric #-}

{-|

Module: NeuronProblem

Description: Foundations for single neuron problems

Copyright: © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Each colony has one cell that receives role input into one
pattern.  It receives a stream of input spikes (on or off at any
given time step) into a configurable channel, and produces a
stream of output spikes.

-}

module NeuronProblem where

import GHC.Generics

import Control.Applicative
import Control.Monad

import Data.Aeson
import Data.Aeson.TH
import Data.Data
import Data.Default
import Data.Foldable (Foldable)
import qualified Data.Foldable as F
import qualified Data.IntMap as IM
import Data.Monoid
import qualified Data.Sequence as Seq
import Data.String
import Data.Text (Text)
import qualified Data.Text as T
import Text.Pandoc.Builder
import Data.Typeable

import Debug.Trace

import AesonUtil
import Colony
import Util

neuronProblem_vcid :: String
neuronProblem_vcid = "$Id: NeuronProblem.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"

data NeuronProblem = NeuronProblem {
  inputPattern :: Int,
  -- ^ the pattern that gets incremented when the neuron receives
  -- input
  inputActivation :: Int,
  -- ^ how much activation to add when the neuron receives input
  outputPattern :: Int,
  -- ^ when the activation of this pattern goes high enough, the
  -- neuron spikes
  outputThreshold :: Int,
  -- ^ the threshold that the output pattern must reach to spike
  outputCost :: Int
  -- ^ the activation to be subtracted when the neuron spikes,
  -- should be /positive/
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''NeuronProblem)

instance Default NeuronProblem where

  def = NeuronProblem {
    inputPattern = 3, -- encodes as 010000g
    inputActivation = 8,
    outputPattern = 41, -- encodes as 101111g
    outputThreshold = 2,
    outputCost = 2
    }

data NeuronProblemTraceSummary = NeuronProblemTraceSummary {
  npts_in :: String,
  -- ^ input word as a string of 0s and 1s
  npts_out :: String
  -- ^ output word as a string of 0s and 1s
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)


$(deriveJSON myJSONOpts{fieldLabelModifier = drop 5}
  ''NeuronProblemTraceSummary)

singleNeuronColonySpec NeuronProblem{..} =
  ColonySpec{
    inputSpecs = IM.fromList [inputSpec],
    emitterSpecs = [emitterSpec],
    ..}
  where
    numCells = 1
    cellNames = IM.fromList [(0, "Neuron")]
    synapseSpecs = []

    inputSpec =
      (1,     -- message is 1
       ("In", -- named "In"
        0, -- feeds to cell 0
        ReceptorSpec{
          targetPattern = inputPattern,
          targetDelta = inputActivation}))

    emitterSpec =
      ("Out", -- named "Out"
       0, -- reads cell 0
       EmitterSpec{
         sourcePattern = outputPattern,
         sourceThreshold = outputThreshold,
         activationDelta = -outputCost },
       1) -- message is 1


-- | Convert a list of 'Bool's to one character for 'False' and
-- another for 'True'
boolsToDisplay ::
  Char -- ^ character for 'True'
  -> Char -- ^ character for 'False'
  -> [Bool] -> String

boolsToDisplay on off xs =
  [if x then on else off | x <- xs]

-- | Convert a list of 'Bool's to "0" for 'False' and "1" for 'True'
boolsTo01s :: [Bool] -> String
boolsTo01s =
  boolsToDisplay '1' '0'

-- | Convert a list of 'Bool's to ╌ (0x254c) for 'False' and ┸ (0x2538) for 'True'
boolsToDrawing :: [Bool] -> String
boolsToDrawing =
  boolsToDisplay '┸' '╌'

-- | Things that can be represented as a sequence of 'Bool', as
-- in on and off
class ToBools t where
  toBools :: t -> [Bool]


-- This is a little weird, but it works.
instance ToBools Bool where
  toBools b = [b]

instance ToBools elt => ToBools [elt] where
  toBools = concatMap toBools


-- | Convert a list of 'Bool' indicating when the input is high
-- to the correct input format for a single neuron 'Colony'

boolsToColonyInputStream ::
  NeuronProblem
  -> [Bool]
  -> [[Int]]

boolsToColonyInputStream NeuronProblem {..} input =
  [if x
   then [1]
   else []
  | x <- input]


-- | Convert the output from a single neuron 'Colony' to a list
-- of 'Bool's for when the output is high.  The absurdly general
-- type is in case I later decide to change the container type
-- (list) from 'Sequence'

colonyOutputsToBools ::
  Foldable list =>
  NeuronProblem
  -> list [emission] -> [Bool]

colonyOutputsToBools NeuronProblem{..} colonyOutputSeq =
  [not (null emitted)
  | emitted <- F.toList colonyOutputSeq ]
