{-# LANGUAGE BangPatterns, RecordWildCards, NamedFieldPuns, ScopedTypeVariables, NoMonomorphismRestriction, Rank2Types, FlexibleInstances, DeriveDataTypeable, FlexibleContexts, UndecidableInstances, StandaloneDeriving, MultiParamTypeClasses #-}

{-|

Module: PopulationDatabase

Description : Big monad for dealing with the whole population database.

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}


module PopulationDatabase (
  PopulationDatabaseException(..),
  InitializationRec(..),
  initializePopulationDatabase,
  reinitializePopulationDatabase,
  getProblemName,
--  initializeCurrentState,
  PopMonad,
  TableName,
  withPopCache,
  ArchivePredicate(..),
  getAgents,
  getAgent,
  getLivingAgents,
  putAgent,
  getCurrentIdentifier,
  getCurrentTimeStep,
  getBestRating,
  getRecordGenerations,
  getArchivePredicate,
  nextTime,
  getLandmarks,
  getGeneration,
  getCurrentRandomSeed,
  withCurrentRandomSeed,
  PopulationBirthDeathDelta(..),
  PopCache(..),
  applyBirthDeathDelta
  ) where

--import Debug.Trace

import Control.Monad
import Control.Monad.Trans
import Control.Monad.Trans.State
import Control.Exception
import qualified Control.Exception as E
import Control.DeepSeq

import Data.Binary (Binary)
import Data.Data
import qualified Data.List as L
import Data.Map (Map)
import qualified Data.Map as M
import Data.Maybe
import Data.Set (Set)
import qualified Data.Set as S
import Data.Typeable
import Data.Word

import System.IO.Error
import System.FilePath
import System.Directory

import Text.Printf

import Agent
import ArchivePredicate
import DBStorable
import Database
import qualified EventHandler as Ev
import Logging
import MapUtil
import ParUtil
import PredicateTools
import RandomST
import ShortShow




populationDatabase_vcid :: String
populationDatabase_vcid = "$Id: PopulationDatabase.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"

data TableName =
  InitialState
  | CurrentState
  | Landmark
  | Generation
  | Archive
    deriving (Read, Show, Eq, Ord)

nonInitialTableNames :: [TableName]
nonInitialTableNames = [CurrentState, Generation, Landmark, Archive]

data PopulationDatabaseException =
  DuplicateIdentifier Integer
  | LivingIDsInconsistent
  | BadBirthDeathDeltaToKill
  | BadAgentRecord Integer
  | DirectoryAlreadyInitialized FilePath
  | DirectoryNotInitialized FilePath
  deriving (Eq, Ord, Read, Show, Typeable)

instance E.Exception PopulationDatabaseException

-- Interesting things that can happen in the database:

-- | An 'Agent' has been created and added to the current population
data EventAgentBorn agent =
  AgentBorn agent
  deriving (Read, Show, Eq, Ord, Typeable, Data)

instance
  (Show agent, Typeable agent)
    => Ev.Event (EventAgentBorn agent)

-- | An 'Agent' has died and been removed from the current population
data EventAgentDied agent =
  AgentDied agent
  deriving (Read, Show, Eq, Ord, Typeable, Data)

instance
  (Show agent, Typeable agent)
    => Ev.Event (EventAgentDied agent)

-- | A new landmark has been found
data EventNewLandmark agent =
  NewLandmark agent
  deriving (Read, Show, Eq, Ord, Typeable, Data)

instance
  (Show agent, Typeable agent)
    => Ev.Event (EventNewLandmark agent)


-- | A 'PopulationBirthDeathDelta' has been applied, but the
-- current time step hasn't been advanced yet.
data EventGenerationComplete =
  GenerationComplete
  deriving (Read, Show, Eq, Data, Typeable)
instance Ev.Event EventGenerationComplete

-- | Initialization fields

data InitializationRec genome = InitializationRec {
  problemName :: String,
  randomSeed :: [Word32],
  recordGenerations :: Bool,
  -- ^ whether to maintain the Generations database of which
  -- 'Agent's are in each generation,
  archivePredicate :: CompoundMPredicate ArchivePredicate,
  genomes :: [genome]
  }

instance (Binary genome) =>
         DBStorable (InitializationRec genome) where
  dbStore (InitializationRec {..}) prefix tableName = do
    let prefixer = (++) prefix
    dbPut (prefixer "problemName") problemName tableName
    dbPut (prefixer "randomSeed") randomSeed tableName
    dbPut (prefixer "recordGenerations") recordGenerations tableName
    dbPut (prefixer "archivePredicate") archivePredicate tableName
    dbPut (prefixer "genomes") genomes tableName
  dbLoad prefix tableName = do
    let prefixer = (++) prefix
    problemName <- dbUnsafeGet (prefixer "problemName") tableName
    randomSeed <- dbUnsafeGet (prefixer "randomSeed") tableName
    recordGenerations <- dbUnsafeGet (prefixer "recordGenerations") tableName
    archivePredicate <- dbUnsafeGet (prefixer "archivePredicate") tableName
    genomes <- dbUnsafeGet (prefixer "genomes") tableName
    return (InitializationRec {..})

-- | Current state fields

data StateRec rating = StateRec {
  bestRating :: rating,
  currentTimeStep :: Integer, -- ^ time step to be used for any new 'Agent's
  currentIdentifier :: Integer, -- ^ identifier to be used for next new 'Agent'
  currentRandomSeed :: [Word32],
  livingAgentIDs :: [Integer]
  } deriving (Read, Show, Eq, Ord)


instance
  (Binary rating)
  =>
  DBStorable (StateRec rating) where
  dbStore (StateRec {..}) prefix tableName = do
    let prefixer = (++) prefix
    dbPut (prefixer "bestRating") bestRating tableName
    dbPut (prefixer "currentTimeStep") currentTimeStep tableName
    dbPut (prefixer "currentIdentifier") currentIdentifier tableName
    dbPut (prefixer "currentRandomSeed") currentRandomSeed tableName
    dbPut (prefixer "livingAgentIDs") livingAgentIDs tableName
  dbLoad prefix tableName = do
    let prefixer = (++) prefix
    bestRating <- dbUnsafeGet (prefixer "bestRating") tableName
    currentTimeStep <- dbUnsafeGet (prefixer "currentTimeStep") tableName
    currentIdentifier <- dbUnsafeGet (prefixer "currentIdentifier") tableName
    currentRandomSeed <- dbUnsafeGet (prefixer "currentRandomSeed") tableName
    livingAgentIDs <- dbUnsafeGet (prefixer "livingAgentIDs") tableName
    return (StateRec {..})


isPopulationDatabaseInitialized :: FilePath -> IO Bool
isPopulationDatabaseInitialized environmentDirectory = do
  doesFileExist (environmentDirectory </> "InitialState")


assertNotInitialized :: String -> FilePath -> IO ()
assertNotInitialized caller environmentDirectory = do
  init <- isPopulationDatabaseInitialized environmentDirectory
  when init
    (msgExc (DirectoryAlreadyInitialized environmentDirectory)
     (caller
      ++ ": Directory already initialized: "
      ++ environmentDirectory))


assertInitialized :: String -> FilePath -> IO ()
assertInitialized caller environmentDirectory = do
  init <- isPopulationDatabaseInitialized environmentDirectory
  when (not init)
    (msgExc (DirectoryNotInitialized environmentDirectory)
     (caller
      ++ ": Directory already initialized: "
      ++ environmentDirectory))


-- | (High level) Create an empty population database on the file system.

initializePopulationDatabase ::
  forall genome rating.
  (Binary genome, Binary rating) =>
  DatabaseSpec
  -- ^ database directory
  -> rating
  -- ^ minimum rating for landmarks
  -> InitializationRec genome
  -> IO ()

initializePopulationDatabase
  dbSpec@DatabaseSpec{..}
  bestRating
  initializationRec@(InitializationRec {..}) = do
  createDirectoryIfMissing True environmentDirectory
  assertNotInitialized "initializePopulationDatabase" environmentDirectory
  -- Create a DoNotIndex file in the directory so that I can make
  -- sure various file indexing programes aren't going through
  -- gigs of stuff:
  writeFile (environmentDirectory </> "DoNotIndex") ""
  writeFile (environmentDirectory </> "problemName.txt") problemName
  withDatabase dbSpec  $ do
    let currentTimeStep = 0
        currentIdentifier = 0
        currentRandomSeed = randomSeed
        livingAgentIDs = []
        sr :: StateRec rating
        sr = StateRec {..} -- captures 'bestRating'
    dbStore sr "" CurrentState
    dbTouch Archive
    dbTouch Landmark
    dbStore initializationRec "" InitialState


-- | (High level) Reinitialize the database, keeping the previous
-- 'InitializationRec' except for the possibility to change
-- 'recordGenerations'.  This removes all nonessential databases, etc.  It
-- returns the 'InitializationRec' so the 'GenomeSpec' can be
-- specified.

reinitializePopulationDatabase ::
  forall genome rating.
  (Binary genome,
   Binary rating) =>
  FilePath
  -- ^ database directory
  -> rating
  -- ^ minimum rating for landmarks
  -> Bool
  -- ^ record generations flag
  -> CompoundMPredicate ArchivePredicate
  -- ^ when to record 'Agent's in the Archive database
  -> IO (InitializationRec genome)

reinitializePopulationDatabase environmentDirectory bestRating
  recordGenerations archivePredicate = do

  assertInitialized "reinitializePopulationDatabase" environmentDirectory
  withDatabase DatabaseSpec{..} $ do
    oldInitRec <- dbLoad "" InitialState
    let newInitRec = oldInitRec {
          recordGenerations = recordGenerations,
          archivePredicate = archivePredicate
          }
    dbStore newInitRec "" InitialState
    liftIO $ mapM_ (sturdyRemoveTable environmentDirectory) nonInitialTableNames
    let currentTimeStep = 0
        currentIdentifier = 0
        currentRandomSeed = randomSeed newInitRec
        livingAgentIDs = []
        sr :: StateRec rating
        sr = StateRec {..} -- captures 'bestRating'
    dbStore sr "" CurrentState
    dbTouch Archive
    dbTouch Landmark
    return newInitRec

-- | Remove a file if it exists, or its compressed equivalent, or
-- do nothing if the file doesn't exist.

sturdyRemoveTable :: FilePath -> TableName -> IO ()
sturdyRemoveTable environmentDirectory t = do
  let path = environmentDirectory </> show t
  msgDebug ("Removing table " ++ show t)
  _ <- tryJust (exFilter isDoesNotExistError) $ removeFile path
  _ <- tryJust (exFilter isDoesNotExistError) $ removeFile (path `addExtension` ".bz2")
  return ()

-- | A filter adapter for use with exceptions and 'tryJust'.
exFilter :: (t -> Bool) -> t -> Maybe ()
exFilter pred e = if pred e then Just () else Nothing

-- | Read the problem name from its file.
getProblemName :: FilePath -- ^ database directory
                   -> IO String
getProblemName dir = do
  probNameContents <- readFile (dir </> "problemName.txt")
  return (head (words probNameContents))


{-

-- | Initialize the 'CurrentState' table from @InitialState@

initializeCurrentState ::
  forall genome rating.
  (Binary genome,
   Binary rating)
  =>
  rating ->
  DBMonad TableName (InitializationRec genome)
initializeCurrentState bestRating = do
  initializationRec@(InitializationRec {..}) <- dbLoad "" InitialState
  let
      currentTimeStep = 0
      currentIdentifier = 0
      currentRandomSeed = randomSeed
      livingAgentIDs = []
      sr :: StateRec rating
      sr = StateRec {..} -- captures 'bestRating'
  dbStore sr "" CurrentState
  dbTouch Archive
  dbTouch Landmark
  return initializationRec
-}

type PopMonad genome rating =
  StateT (PopCache genome rating) (DBMonad TableName)

data PopCache genome rating =
  PopCache {
  initializationRec :: InitializationRec genome,
  stateRec :: StateRec rating,
  livingAgents :: Map Integer (Agent genome rating),
  eventReceivers :: [Ev.Receiver (PopMonad genome rating) ()]
  }


dispatch ::
  Ev.Event e =>
  e
  -> PopMonad genome rating ()
dispatch event = do
  rs <- gets eventReceivers
  _ <- Ev.dispatch rs event
  return ()

addEventReceiver ::
  (Ev.Event e) =>
  (e -> PopMonad genome rating ()) -> PopMonad genome rating ()
addEventReceiver r = do
  PopCache {..} <- get
  put (PopCache {eventReceivers = (Ev.Receiver r):eventReceivers, ..})


instance
  (Binary genome,
   Show genome,
   Typeable genome)
  =>
  (MPredicate
   (PopMonad genome rating)
   ArchivePredicate
   (Agent genome rating)) where

    check a (BornOnMultiple n) =
      let
        birthTime = case lifespan a of
          Left m -> m
          Right (m, _) -> m
      in
       return (birthTime `mod` n == 0)

    check a (DiedOnMultiple n) =
      case lifespan a of
        Right (_, m) ->
          return (m `mod` n == 0)
        Left _ ->
          return False

    check a (AliveOnMultiple n) =
      case lifespan a of
        Right (born, died) ->
          let
            (q, r) = quotRem died n
          in
           return (born <= n * q)
        Left born -> do
          t <- getCurrentTimeStep
          let (q, r) = quotRem t n
          return (born <= n * q)

-- | Perform an action (in the 'PopMonad') on a population
-- contained in a database (in the 'DBMonad').

withPopCache ::
  forall genome rating c.
  (Binary genome,
   Show genome,
   Typeable genome,
   Binary rating,
   Show rating,
   Typeable rating) =>
  PopMonad genome rating c -> DBMonad TableName c

withPopCache axn0 = do
  initializationRec <- dbLoad "" InitialState
  stateRec@(StateRec {..}) <- dbLoad "" CurrentState
  items <- dbGetMany livingAgentIDs CurrentState
  let livingAgents = M.fromList items
  when (livingAgentIDs /= M.keys livingAgents) $
    let
      missingIDs = livingAgentIDs L.\\ M.keys livingAgents
    in
     (msgExc
      LivingIDsInconsistent
      ("Unable to load living agents with the following IDs: \n"
       ++ unlines (map show missingIDs)))
  let eventReceivers = []
  let popCache0 = PopCache {..}
  (result, popCache1) <- runStateT axn1 popCache0
  -- rebind stateRec to the new current state
  let PopCache {..} = popCache1
  dbStore initializationRec "" InitialState
  dbStore stateRec "" CurrentState
  return result
  where
    -- Update Generations after each generation is completed
    recordGenerationHandler GenerationComplete =
      recordGeneration

    -- Store dead agents in the Archive
    keepDeadAgent ::
      CompoundMPredicate ArchivePredicate
      -> EventAgentDied (Agent genome rating)
      -> PopMonad genome rating ()
    keepDeadAgent p (AgentDied a) = do
      f <- check a p
      when f $ lift $ do
        liftIO $ msgDebug ("Archiving agent " ++ show (identifier a))
        dbPut (identifier a) a Archive

    axn1 = do
      f <- getRecordGenerations
      when f $ do
        addEventReceiver recordGenerationHandler

      p <- getArchivePredicate
      unless (p == Never) $ do
        addEventReceiver (keepDeadAgent p)
      axn0



-- | Get as many of the requested agents as possible from the
-- current population or the archive.

getAgents ::
  (Binary genome,
   Binary rating)
  =>
  [Integer]
  -> PopMonad genome rating (Map Integer (Agent genome rating))

getAgents ns = do
  mAgents <- mapM getAgent ns
  let agents = catMaybes mAgents
  return (M.fromList [ (identifier a, a) | a <- agents])


-- | Read the agent from the cache, the landmark table, or the
-- archive.

getAgent ::
  (Binary genome,
   Binary rating)
  =>
  Integer -> PopMonad genome rating (Maybe (Agent genome rating))

getAgent n = do
  living <- getLivingAgents
  let mA = M.lookup n living
  case mA of
    Just _ -> return mA
    Nothing -> do
      mL <- lift $ dbGet n Landmark
      case mL of
        Just _ -> return mL
        Nothing -> do
          lift $ dbGet n Archive


-- | Return an indexed set of all living agents

getLivingAgents ::
  (Binary genome,
   Binary rating)
  =>
  PopMonad genome rating (Map Integer (Agent genome rating))
getLivingAgents =
  gets livingAgents


-- | (Private) Write the agent to the current state table.

putAgent ::
  (Binary genome,
   Binary rating) =>
  Agent genome rating -> PopMonad genome rating ()

putAgent a@(Agent {..}) = do
  lift $ dbPut identifier a CurrentState


-- | (Private) Delete an agent from the current state table.

deleteAgent ::
  (Binary genome,
   Binary rating)
  =>
  Agent genome rating -> PopMonad genome rating ()

deleteAgent (Agent {..}) = do
  lift $ dbDelete identifier CurrentState


-- | (Private) Update the table of living agents in the
-- 'PopCache' state

updateLivingAgents ::
  (Map Integer (Agent genome rating)) -> PopMonad genome rating ()

updateLivingAgents newLivingAgents = do
  PopCache {..} <- get
  let StateRec {..} = stateRec
  let livingAgents = newLivingAgents
  let livingAgentIDs = M.keys livingAgents
  let stateRec = StateRec {..}
  put (PopCache {..})


-- | (Private) Increment the current identifier in the cache,
-- return the old value.
--
-- Implementation note: I'm going to change this on 2013-07-15 to
-- return the old value rather than the new value.  I have to
-- re-run a bunch of stuff anyway, so I figured I'd fix things up
-- so that everything is indexed from 0 rather than some from 1,
-- some from 0.

nextIdentifier :: PopMonad genome rating Integer
nextIdentifier = do
  PopCache {..} <- get
  let StateRec {..} = stateRec
  let m = currentIdentifier
  let currentIdentifier = m + 1
  let stateRec = StateRec {..}
  put (PopCache {..})
  return m


-- | Fetch the current identifier (that will be used for the next new agent)
getCurrentIdentifier :: PopMonad genome rating Integer
getCurrentIdentifier = gets (currentIdentifier . stateRec)


-- | Fetch the current time from the cache
getCurrentTimeStep :: PopMonad genome rating Integer
getCurrentTimeStep =
  gets (currentTimeStep . stateRec)


-- | Fetch the current random seed
getCurrentRandomSeed :: PopMonad genome rating [Word32]
getCurrentRandomSeed =
  gets (currentRandomSeed . stateRec)


-- | Fetch the current best rating
getBestRating :: PopMonad genome rating rating
getBestRating =
  gets (bestRating . stateRec)


-- | (Private) Update the best rating in the monadic state
updateBestRating :: rating -> PopMonad genome rating ()
updateBestRating r = do
  PopCache {..} <- get
  let StateRec {..} = stateRec
  let bestRating = r
  let stateRec = StateRec {..}
  put (PopCache {..})


getRecordGenerations :: PopMonad genome rating Bool
getRecordGenerations =
  gets (recordGenerations . initializationRec)


getArchivePredicate ::
  PopMonad genome rating (CompoundMPredicate ArchivePredicate)
getArchivePredicate =
  gets (archivePredicate . initializationRec)


-- | (Private) Dispatch the 'GenerationComplete' event.  Then
-- increment the current time in the cache and write current
-- state to the database.  The population should have been bred,
-- new agents stored etc. before running this action.

nextTime ::
  (Binary rating)
  =>
  PopMonad genome rating ()
nextTime = do
  PopCache {..} <- get
  let StateRec {..} = stateRec
  dispatch GenerationComplete
  let now = currentTimeStep
  let currentTimeStep = now + 1
  let stateRec = StateRec {..}
  lift $ dbStore stateRec "" CurrentState
  put (PopCache {..})
  -- continue with the next generation


recordGeneration :: PopMonad genom rating ()
recordGeneration = do
  PopCache {..} <- get
  let StateRec {..} = stateRec
  -- msgDebug (printf "recordGeneration: recording generation %d" currentTimeStep)
  lift $ dbPut currentTimeStep livingAgentIDs Generation


-- | Put a new agent into the database, intialize its lifespan,
-- and return the new 'Agent' with its new 'identifier'.
-- Dispatch a 'AgentBorn' event.

newAgent ::
  (Binary genome,
   Show genome,
   Typeable genome,
   Binary rating,
   Show rating,
   Typeable rating)
  =>
  (Egg genome, rating)
  -> PopMonad genome rating (Agent genome rating)

newAgent (egg, rating) = do
  identifier <- nextIdentifier
  -- This had better not be in the database already
  mA <- getAgent identifier
  when (isJust mA) (throw (DuplicateIdentifier identifier))
  t <- getCurrentTimeStep
  let a = Agent {lifespan = Left t, ..}
  putAgent a
  dispatch (AgentBorn a)
  return a


-- | Kill an agent, possibly keeping it in the archive, but
-- removing it from the current state.  Dispatch an 'AgentDied'
-- event.

endLifespan ::
  (Binary genome,
   Show genome,
   Typeable genome,
   Binary rating,
   Show rating,
   Typeable rating)
  =>
  Agent genome rating
  -> PopMonad genome rating ()

endLifespan a0@(Agent {lifespan = Left birth, ..}) = do
  t <- getCurrentTimeStep
  let a = (a0 {lifespan = Right (birth, t)})
  -- The agent may be archived as a landmark with no death date,
  -- so we might have to update it:
  updateArchive a Landmark
  -- msgDebug ("endLifespan: About to delete agent " ++ show identifier)
  deleteAgent a
  dispatch (AgentDied a)

endLifespan a =
  msgExc (BadAgentRecord (identifier a))
    ("endLifespan: unexpected agent " ++ show a)


-- | Return a tabl of landmark 'Agent's

getLandmarks ::
  (Binary genome,
   Binary rating)
  =>
  PopMonad genome rating (Map Integer (Agent genome rating))
getLandmarks =
  lift $ withCursor Landmark $ \c -> do
    landmarksAsPairs <- cursorForEach c $ \agentID agent ->
      return (agentID, agent)
    return (M.fromList landmarksAsPairs)


--- landmark maintenance ---

-- | (Private) If the agent was stored for a special reason (a
-- landmark, landmark's parent), store it again so that the
-- lifespan is correct.

updateArchive ::
  (Binary genome,
   Binary rating)
  =>
  Agent genome rating
  -> TableName
  -> PopMonad genome rating ()

updateArchive a@(Agent {..}) tableName = do
  agentInArchive <- lift $ dbHasKey identifier tableName
  if agentInArchive
    then do
    msgDebug ("updateArchive: updating " ++ show identifier
              ++ " in " ++ show tableName)
    lift $ dbPut identifier a tableName
    else do
      -- msgDebug ("updatingArchive: no record of " ++ show identifier ++ " in " ++ show tableName)
      return ()


-- | Record a landmark when this 'Agent' has the best rating so
-- far.  That is, save the 'Agent' and its parents to the
-- 'Landmark' database.  Dispatch a 'NewLandmark' event.

maybeRecordLandmark ::
  (Binary genome,
   Show genome,
   Typeable genome,
   Binary rating,
   Show rating,
   Ord rating,
   Typeable rating)
  =>
  Agent genome rating
  -> PopMonad genome rating ()

maybeRecordLandmark agent = do
  bestRating <- getBestRating
  let r = rating agent
  when (r > bestRating) $ do
    msgInfo ("maybeRecordLandmark: new best rating of "
             ++ show r ++ " agent #" ++ show (identifier agent))
    lift $ dbPut (identifier agent) agent Landmark
    updateBestRating r
    -- We also have to save the parents in the archive.  Which
    -- means you have to call this BEFORE deleting the parents.
    forM_ (parentIds $ egg agent) $ \n -> do
      Just p <- getAgent n
      lift $ dbPut n p Archive
    dispatch (NewLandmark agent)


--- really public ---

-- | Return a 'Map' of 'Agent's that were alive at generation 'n'

getGeneration ::
  (Binary genome,
   Binary rating)
  =>
  Integer -- ^ n
  -> PopMonad genome rating (Map Integer (Agent genome rating))

getGeneration n = do
  mIds <- lift $ dbGet n Generation
  case mIds of
    Nothing ->
      return M.empty
    Just ids ->
      getAgents ids


-- | Perform an action in the 'RandomST' monad using the current
-- random seed from 'PopMonad', storing the new random seed in
-- 'PopMonad' for future use

withCurrentRandomSeed ::
  (forall s. RandomST s value)
  -> PopMonad genome rating value

withCurrentRandomSeed axn = do
  PopCache {..} <- get
  let StateRec {..} = stateRec
  let seed0 = currentRandomSeed
  let (val, seed1) = withSeed seed0 axn
  let currentRandomSeed = seed1
  let stateRec = StateRec {..}
  put (PopCache {..})
  return val


data PopulationBirthDeathDelta genome rating =
  PopulationBirthDeathDelta {
    toKill :: [Integer],
    newEggsRatings :: [(Egg genome, rating)]  }

-- | Give a birth-death delta, change the cache and database to
-- hold the new population, and store the new ratings.

applyBirthDeathDelta ::
  (Binary genome,
   NFData genome,
   Show genome,
   Typeable genome,
   Binary rating,
   NFData rating,
   Ord rating,
   Show rating,
   Typeable rating) =>
  PopulationBirthDeathDelta genome rating
  -> PopMonad genome rating ()

applyBirthDeathDelta (PopulationBirthDeathDelta {..}) = do
  PopCache {..} <- get
  -- consistency check
  let StateRec {..} = stateRec
  when (livingAgentIDs /= (M.keys livingAgents))
    (throw LivingIDsInconsistent)
  when (not (S.isSubsetOf (S.fromList toKill) (S.fromList livingAgentIDs) ))
    (throw BadBirthDeathDeltaToKill)
  -- consistency check end
  let agentsToKeep = mapDifference livingAgents toKill
  let agentsToKill = mapIntersection livingAgents toKill
      -- evaluate ratings / eggs in parallel
  newAgents <- mapM newAgent (parListDeep newEggsRatings)
  mapM_ maybeRecordLandmark newAgents
  let newAgentIdentifiers = map identifier newAgents
  let newAgentsMap = M.fromList (zip newAgentIdentifiers newAgents)
  let combined = agentsToKeep `M.union` newAgentsMap
  updateLivingAgents combined
  mapM_ endLifespan (M.elems agentsToKill)
