{-# LANGUAGE TypeFamilies, FlexibleContexts, RecordWildCards, NamedFieldPuns #-}

{-|

Module: Problem

Description : Abstractions for representing a deterministic-rating-based evolutionary problem

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

I'm decoupling this from "Genome".  It uses only the
'ReactionProgram' type from "Compiler".

-}

module Problem (
  Problem(..),
--  Reportable(..),
  traceColonyOn,
  traceColonyOnAll,
  traceColonyOnAllExtra,
  runColonyOn,
  runColonyOnAll,
  rateColonyAdd
  ) where

import Control.DeepSeq

import Data.Sequence (Seq)
import qualified Data.Sequence as Seq

import Colony
import Compiler (ProgramInformation, ReactionProgram)
import Instruction
import ParUtil

problem_vcid :: String
problem_vcid = "$Id: Problem.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"

class Problem problem where
  type Input problem :: *
  type Output problem :: *
  type Rating problem :: *

  -- | Fetch the problem's initial 'ColonySpec'
  initialColonySpec :: problem -> ColonySpec

  problemInputToColonyInputStream ::
    problem -> Input problem -> ColonyInputStream

  colonyOutputsToProblemOutput ::
    problem -> Seq ColonyOutput -> Output problem

  -- | Say whether a run of a colony should stop at the end of a
  -- time step even if more input is available (function of its
  -- current outputs).  The default implementation always returns
  -- 'False', meaning don't stop until the end of an input stream
  -- is reached.
  stopCheck ::
    problem -> ColonyOutput -> Bool

  stopCheck _ _ = False


  -- | Rating below which landmarks are not recognized
  minLandmarkRating ::
    problem -> Rating problem

  -- | List all inputs that are used for rating purposes
  inputsForRating ::
    problem -> [Input problem]


  -- | Give a (numerical) rating for yielding the output given
  -- this input.
  rateResult ::
    problem
    -> Input problem -- ^ a given input
    -> Output problem -- ^ the resulting output
    -> Int -- ^ total number of reactions performed
    -> Int -- ^ total of absolute value of all reaction deltas
    -> Rating problem -- ^ resulting rating



traceColonyOn ::
  (Problem problem)
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> Input problem
  -> (TraceResult, (Int, Int, Seq TraceStep))
  -- ^ (trace result, (total reaction count, total reaction deltas, colony outputs))

traceColonyOn problem progAndInfo input =
  let
    colonyInput :: ColonyInputStream
    colonyInput = problemInputToColonyInputStream problem input
  in
   (traceColony
    (initialColonySpec problem)
    (stopCheck problem)
    progAndInfo
    colonyInput)


-- | Trace a colony on all inputs used for rating it.

traceColonyOnAll ::
  (Problem problem,
   NFData (Input problem),
   NFData (Output problem),
   NFData (Rating problem))
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> [(Input problem, Output problem, Int, Int, TraceResult)]
  -- ^ list of input, output, total reaction count, total reaction deltas,
  -- trace results

traceColonyOnAll problem progAndInfo =
  parListDeep $ map f (inputsForRating problem)
  where
    f input =
      (input, output, totalReactionCount, totalReactionDeltas, traceResult)
      where
        (traceResult, (totalReactionCount, totalReactionDeltas, traceSteps)) =
          traceColonyOn problem progAndInfo input
        output = colonyOutputsToProblemOutput problem (fmap tsOutput traceSteps)


-- | Trace a colony on all inputs used for rating it, including
-- all the 'TraceStep' objects (which can tie up a lot of memory)

traceColonyOnAllExtra ::
  (Problem problem,
   NFData (Input problem),
   NFData (Output problem),
   NFData (Rating problem))
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> [(Input problem, Output problem, Int, Int, TraceResult, Seq TraceStep)]
  -- ^ list of input, output, total reaction count, total reaction deltas,
  -- trace results, trace steps
traceColonyOnAllExtra problem progAndInfo =
  parListDeep $ map f (inputsForRating problem)
  where
    f input =
      (input, output, totalReactionCount, totalReactionDeltas, traceResult, traceSteps)
      where
        (traceResult, (totalReactionCount, totalReactionDeltas, traceSteps)) =
          traceColonyOn problem progAndInfo input
        output = colonyOutputsToProblemOutput problem (fmap tsOutput traceSteps)


runColonyOn ::
  (Problem problem)
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> Input problem
  -> (Output problem, Int, Int)
  -- ^ problem output, total reaction count, total reaction deltas

runColonyOn problem progAndInfo input =
  (problemOutput, totalReactionCount, totalReactionDeltas)
  where
    colonyInput =
      problemInputToColonyInputStream problem input
    (totalReactionCount, totalReactionDeltas, colonyOutput) =
      (runColony
       (initialColonySpec problem)
       (stopCheck problem)
       progAndInfo
       colonyInput)
    problemOutput =
      colonyOutputsToProblemOutput problem colonyOutput


-- | Run a colony on all inputs used for rating it.
runColonyOnAll ::
  (Problem problem,
   NFData (Input problem),
   NFData (Output problem),
   NFData (Rating problem))
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> [(Input problem, Output problem, Int, Int, Rating problem)]

runColonyOnAll problem progAndInfo =
  parListDeep $
  [(input, output, totalReactionCount, totalReactionDeltas,
    rateResult problem input output totalReactionCount totalReactionDeltas)
  | input <- inputsForRating problem,
    let (output, totalReactionCount, totalReactionDeltas) =
          runColonyOn problem progAndInfo input]


-- | Rate the colony on each possible input and add the ratings
-- together.
rateColonyAdd ::
  (Num (Rating problem),
   Problem problem,
   NFData (Input problem),
   NFData (Output problem),
   NFData (Rating problem)) =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> Rating problem

rateColonyAdd problem progAndInfo =
  sum [r | (_, _, _, _, r) <- runColonyOnAll problem progAndInfo]
