{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables, DeriveDataTypeable, StandaloneDeriving, FlexibleContexts, FlexibleInstances, GADTs, ExistentialQuantification, NoMonomorphismRestriction, DeriveGeneric, TemplateHaskell, ImpredicativeTypes, RankNTypes, UndecidableInstances, ConstraintKinds, KindSignatures #-}

{-|
Module      : ProblemInventory

Description : Collection of all available Problems and other configuration

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : POSIX

-}

module ProblemInventory where

import GHC.Exts (Constraint)
import GHC.Generics()

import Control.DeepSeq

import Data.Aeson hiding (Array)
--import Data.Aeson.TH()
import Data.Binary (Binary)
import Data.Data
import Data.Default
import Data.Map (Map)
import qualified Data.Map as M
--import Data.Text
import Data.Typeable()
--import qualified Data.Yaml as Y


import Chromosome
import GeneralName
import Genome
import qualified HaploidSingletonGenome as Hap
import qualified PolyploidGenome as Poly
--import qualified PopulationSpec as PopSpec
--import PopulationDatabase
import Problem
--import qualified RangeMap as R
import RandomST
import ShortShow
import Snip

import FPCopyProblem
import CoincidenceDetectorProblem
--import SwitchedOscillatorProblem
--import BurstingProblem
--import NeuronProblem

problemInventory_vcid :: String
problemInventory_vcid = "$Id: ProblemInventory.hs,v 8cbecca913b8 2017-01-22 21:10:35Z garrett $"

-- | I've been wanting this for a long time: Box up an object and
-- a class implementation.
data Box (c :: * -> Constraint) =
  forall t. c t => Box t


--- The weirdest thing I've written in a long time ---

data Impl (c :: * -> Constraint) t = Impl

data ImplWrap (c :: * -> Constraint) =
  forall t. c t => ImplWrap (Impl c t)

withNamedObject ::
  Map String (ImplWrap c)
  -> (forall t. c t => Impl c t -> result)
  -> result
  -> String
  -> result
withNamedObject inventory axn ifUnknown name =
  case M.lookup name inventory of
    Just (ImplWrap impl) ->
      axn impl
    Nothing ->
      ifUnknown

--- end of weirdness ---

-- | All the type classes required for a problem.  Note that
-- using a type synonym doesn't work, because the compiler
-- insists all applications of synonyms must be fully saturated,
-- but it doesn't require that for classes. *shrug*
--
-- The constraint FromJSON (Output problem) is needed because we
-- have to export genomes and then import them again, and if I
-- don't tie up all the type variables, the compiler doesn't know
-- how to handle imports using 'QueryResult.AgentResult'.

class
  (Problem problem,
   ToJSON problem,
   FromJSON problem,
   GeneralName problem,
   Show problem,
   Read (Input problem),
   Show (Input problem),
   ToJSON (Input problem),
   FromJSON (Input problem),
   Sniplet problem (Input problem),
   NFData (Input problem),
   Show (Output problem),
   ToJSON (Output problem),
   FromJSON (Output problem),
   NFData (Output problem),
   Sniplet problem (Output problem),
   Num (Rating problem),
   Binary (Rating problem),
   FromJSON (Rating problem),
   NFData (Rating problem),
   Ord (Rating problem),
   Read (Rating problem),
   Show (Rating problem),
   ToJSON (Rating problem),
   Typeable (Rating problem))
  => UsefulProblem problem

instance UsefulProblem FPCopyProblem
instance UsefulProblem CoincidenceDetectorProblem

problemTemplates :: Map String (Box UsefulProblem)
problemTemplates =
  M.fromList   [
    ("Copy21", Box (def :: FPCopyProblem)),
    -- Coincidence detectors
    ("CDMedium", Box problemTemplateCDMedium),
    ("CDLong", Box problemTemplateCDLong),
    ("CDLearnMedium", Box problemTemplateCDLearnMedium),
    ("CDLearnLong", Box problemTemplateCDLearnLong)
  ]


problemInventory :: Map String (ImplWrap UsefulProblem)
problemInventory =
  M.fromList $ map addProblemName [
    ImplWrap (Impl :: Impl UsefulProblem FPCopyProblem),
    ImplWrap (Impl :: Impl UsefulProblem CoincidenceDetectorProblem)
    ]

problemName :: ImplWrap UsefulProblem -> String
problemName (ImplWrap (Impl :: Impl UsefulProblem t)) =
      generalName (undefined :: t)

addProblemName :: ImplWrap UsefulProblem -> (String, ImplWrap UsefulProblem)
addProblemName pw = (problemName pw, pw)

----------

class
  (GenomeSpec gs,
   Show gs,
   ToJSON gs,
   FromJSON gs,
   GeneralName gs,
   Show (Genome gs),
   Typeable (Genome gs),
   NFData (Genome gs),
   Binary (Genome gs),
   ToJSON (Genome gs),
   FromJSON (Genome gs),
   Ord (Genome gs),
   Sample gs (Genome gs),
   Sample (gs, Genome gs, Genome gs) (Genome gs),
   ToJSON (AlignmentSpec gs),
   FromJSON (AlignmentSpec gs),
   Default (AlignmentSpec gs),
   ShortShow (Interval gs)) =>
  UsefulGenomeSpec gs

instance UsefulGenomeSpec Hap.HaploidSingletonGenomeSpec
instance UsefulGenomeSpec Poly.PolyploidGenomeSpec

genomeArchitectureInventory :: Map String (ImplWrap UsefulGenomeSpec)
genomeArchitectureInventory =
  M.fromList[
    ("Haploid",
     ImplWrap (Impl :: Impl UsefulGenomeSpec Hap.HaploidSingletonGenomeSpec)),
    ("Polyploid",
     ImplWrap (Impl :: Impl UsefulGenomeSpec Poly.PolyploidGenomeSpec))
    ]

genomeSpecTemplates :: Map String (Box UsefulGenomeSpec)
genomeSpecTemplates =
  M.fromList [
    ("haploid-32-aligned",
     Box (def :: Hap.HaploidSingletonGenomeSpec)),
    ("haploid-32-length-neutral",
     Box (def {
             Hap.crossoverSpec = def {
                crossoverType = SinglePointLengthNeutral}}
          :: Hap.HaploidSingletonGenomeSpec)),
    ("haploid-32-aligned-short",
     Box shortHaploidSpec),
    ("haploid-32-length-neutral-short",
     Box (shortHaploidSpec {
             Hap.crossoverSpec = def {
                crossoverType = SinglePointLengthNeutral}})),
    ("diploid-2-16-aligned",
     Box (def :: Poly.PolyploidGenomeSpec))
    ]


-- | 'InstructionSpec' like the default but with only 4 pattern
-- bits
shortInstructionSpec :: InstructionSpec
shortInstructionSpec =
  def {
    numPatternBits  = 4
    }

shortHaploidSpec :: Hap.HaploidSingletonGenomeSpec
shortHaploidSpec =
  Hap.makeDefaultHaploidSingletonGenomeSpec 3 32 shortInstructionSpec
