{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables, DeriveDataTypeable, DeriveGeneric, TemplateHaskell, FlexibleContexts #-}

module Query where

import GHC.Generics

import Control.DeepSeq
import Control.Exception
import Control.Monad.Trans

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap
import Data.IntSet (IntSet)
import qualified Data.IntSet as IntSet
import Data.Map (Map)
import qualified Data.Map as Map
-- import Data.Set (Set)
-- import qualified Data.Set as S
import Data.Typeable
import Text.Printf

import AesonUtil
import Agent
import ColonyToStyledNetwork
import Compiler
import EfficaciousAnalysis
import Genome as G
import Instruction
import Problem
import PopulationDatabase
import qualified QueryRequest as Q
import QueryResult
import ShortShow

query_vcid :: String
query_vcid = "$Id: Query.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"

-- | Ways of getting a list of 'Agent's out of a sample run
data SelectAgents =
  Living
  | Landmarks
  | Generation Integer
  | TopRated
  | RatedAtleast Value
  | IDs [Integer]
  deriving (Show, Eq, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''SelectAgents)

-- | Exceptions that can arise during a query
data QueryException =
  CompilationErrors [CompileError] [Instruction]
  | MalformedJSONError String
  | ParseRatingError String
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance Exception QueryException


-- | Fetch 'Agent's from a database as specified by a 'Query'
doSelect ::
  (Binary rating,
   Ord rating,
   FromJSON rating,
   Binary genome)
  =>
     SelectAgents
     -> PopMonad genome rating (Map Integer (Agent genome rating))

doSelect Living = getLivingAgents

doSelect Landmarks = getLandmarks

doSelect (Generation n) = getGeneration n

doSelect (IDs ns) = getAgents ns

doSelect TopRated = do
  living <- getLivingAgents
  let maxRating =
        (maximum
         (map Agent.rating (Map.elems living)))
      topRated =
        (Map.filter
         (\a -> Agent.rating a == maxRating)
         living)
  return topRated

doSelect (RatedAtleast v)  = do
  living <- getLivingAgents
  case fromJSON v of
    Error msg ->
      liftIO $ throwIO (ParseRatingError msg)
    Success r ->
      (return
       (Map.filter
        (\a -> Agent.rating a >= r)
        living))


-- | Process an 'AgentRequest'
queryAgent ::
  (GenomeSpec gs,
   Problem prob,
   ShortShow (Interval gs),
   NFData (Input prob),
   NFData (Output prob),
   NFData (Rating prob)
  ) =>
  gs
  -> prob
  -> Q.AgentRequest
  -> Agent (Genome gs) (Rating prob)
  -> AgentResultX gs prob

queryAgent genomeSpec problem Q.JustAgent a =
  JustAgent a

queryAgent genomeSpec problem Q.Summary a@Agent{egg=Egg{..}, ..} =
  Summary{
    id = identifier,
    genomeLength = G.genomeLength genomeSpec genome,
    ..}

queryAgent genomeSpec problem Q.Export a@Agent{egg=Egg{..}, ..} =
  Export{
    id = identifier,
    genome = exportGenomeData genomeSpec genome
    }

queryAgent genomeSpec problem Q.Program a@Agent{egg=Egg{..}, ..} =
  Program {
    id = identifier,
    genomeLength = G.genomeLength genomeSpec genome,
    instructions = interpretGenome genomeSpec genome,
    ..}

queryAgent genomeSpec problem Q.GenomeDetails a@Agent{egg=Egg{..}, ..} =
  GenomeDetails {
    id = identifier,
    genomeLength = G.genomeLength genomeSpec genome,
    geneDescriptions = map shortShow geneDetails,
    instructionDetails =
      [ (instruction, codonBits, nucleotideBits, instructionEncodingDetails)
      | G.GeneDetails{..} <- geneDetails],
    ..}
  where
    geneDetails = G.geneDetails genomeSpec genome

queryAgent genomeSpec problem Q.RatingInputs a@Agent{egg=Egg{..}, ..} =
  RatingInputs{
    id = identifier,
    ..}
  where
    instructions = interpretGenome genomeSpec genome

    progAndInfo =
      case compile (getInstructionSpec genomeSpec) instructions
      of
        Right progAndInfo ->
          progAndInfo
        Left errors ->
          throw (CompilationErrors errors instructions)

    allInputs = runColonyOnAll problem progAndInfo

queryAgent genomeSpec problem Q.Efficacious a =
  Efficacious{..}
  where
    FullDetails{..} = queryAgent genomeSpec problem Q.FullDetails a

queryAgent genomeSpec problem Q.FullDetails a@Agent{egg=Egg{..}, ..} =
  FullDetails {
    id = identifier,
    genomeLength = G.genomeLength genomeSpec genome,
    geneDescriptions = map shortShow geneDetails,
    instructionDetails =
      [ (instruction, codonBits, nucleotideBits, instructionEncodingDetails)
      | G.GeneDetails{..} <- geneDetails],
    ..}
  where
    instructions = interpretGenome genomeSpec genome

    progAndInfo@(programInfo, _) =
      case compile (getInstructionSpec genomeSpec) instructions
      of
        Right progAndInfo ->
          progAndInfo
        Left errors ->
          throw (CompilationErrors errors instructions)

    instructionMap = IntMap.fromList (zip [0..] instructions)

    geneDetails = G.geneDetails genomeSpec genome
    geneDetailsMap = IntMap.fromList (zip [0..] geneDetails)

    traceResultsPlus = traceColonyOnAll problem progAndInfo
    traceResults = [tr | (input, output, _, _, tr) <- traceResultsPlus]

    allInputs =
      [(input, output, totalReactionCount, totalReactionDeltas,
        rateResult problem input output totalReactionCount totalReactionDeltas)
      | (input, output, totalReactionCount,
         totalReactionDeltas, tr) <- traceResultsPlus]

    ea@EfficaciousAnalysis{
      perCellEfficacious,
      allEfficaciousInstructionIxs,
      allEfficaciousReactions} = efficaciousAnalysis programInfo traceResults


    listInstructions :: IntSet -> [(Int, Instruction)]
    listInstructions ixSet =
      map (\j -> (j, instructionMap IntMap.! j)) (IntSet.toList ixSet)

    listGeneDescriptions :: IntSet -> [String]
    listGeneDescriptions ixSet =
      (map
       (\j -> shortShow $ geneDetailsMap IntMap.! j)
       (IntSet.toList ixSet))


    efficaciousReactions =
      IntMap.map (\(_, rxns, _) -> rxns) perCellEfficacious
    efficaciousInstructions =
      IntMap.map (\(instIxs, _, _) -> listInstructions instIxs) perCellEfficacious
    efficaciousGeneDescriptions =
      IntMap.map (\(instIxs, _, _) -> listGeneDescriptions instIxs) perCellEfficacious
    allEfficaciousInstructions =
      listInstructions allEfficaciousInstructionIxs
    allEfficaciousGeneDescriptions =
      listGeneDescriptions allEfficaciousInstructionIxs


queryAgent genomeSpec problem Q.ReactionNetwork a@Agent{egg=Egg{..}, ..} =
  ReactionNetwork{
    id = identifier,
    reactionNetwork = (efficaciousAnalysisToStyledNetwork
                       effAn
                       (printf "Agent %d" identifier)
                       (initialColonySpec problem))
    }
  where
    instructions = interpretGenome genomeSpec genome

    progAndInfo@(programInfo, _) =
      case compile (getInstructionSpec genomeSpec) instructions
      of
        Right progAndInfo ->
          progAndInfo
        Left errors ->
          throw (CompilationErrors errors instructions)

    traceResultsPlus = traceColonyOnAll problem progAndInfo
    traceResults = [tr | (input, output, _, _, tr) <- traceResultsPlus]

    effAn = efficaciousAnalysis programInfo traceResults
