{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables, DeriveDataTypeable, DeriveGeneric, TemplateHaskell, FlexibleContexts #-}

{-|

Module: QueryResult

Description : Data from querying agents

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Possible results of querying an agent.  These datatypes are
declared in their own module to avoid name clashes with
"QueryRequest".

-}

module QueryResult where

import GHC.Generics

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.IntMap (IntMap)
-- import qualified Data.IntMap as IM
-- import Data.Map (Map)
-- import qualified Data.Map as Map
import Data.Set (Set)
-- import qualified Data.Set as Set
import Data.Typeable
import Prelude hiding (id)

import AesonUtil
import Agent
import BitString
import Compiler
import EfficaciousAnalysis
import Genome
import Instruction
import Problem
import StyledNetwork

queryResult_vcid :: String
queryResult_vcid = "$Id: QueryResult.hs,v 5a02784b68c3 2017-01-19 18:10:41Z garrett $"


-- | Information that can be retrieved from an 'Agent'.  Normally
-- I use 'identifier' for an agent's id number so as not to clash
-- with the identity function in the prelude.  I'm using 'id'
-- here rather than 'identifier' because a lot of this stuff will
-- get exported to JSON and those extra letters could add up to a
-- lot of unnecessary data.  The various types are passed in
-- separately rather than using a 'GenomeSpec' and 'Problem'
-- because of limitations that would result when using the
-- generic and template derivations from packages @aeson@.
data AgentResult genome rating input output =
  JustAgent (Agent genome rating)
  | Summary {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer
    }
    -- ^ quick summary: id, rating, lifespan, parents, genome length
  | Export {
    id :: Integer,
    genome :: Value
    }
    -- ^ genome information only; for moving genomes from one
    -- simulation to another, even one that solves a different
    -- problem, has different settings, etc.
  | Program {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer,
    instructions :: [Instruction]
    }
    -- ^ summary of the agent plus a listing of its instructions,
    -- but no bit-level genome data; requires no significant
    -- computation
  | GenomeDetails {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer,
    geneDescriptions ::  [String],
    instructionDetails :: [(Instruction, BitString, BitString, EncodingDetails)]
    -- ^ architecture-independent instruction information, @(the
    -- instruction, codon bits, nucleotide bits, encoding details)@
    }
    -- ^ summary of the agent plus a complete listing of the
    -- genome and additional information; requires no
    -- significant computation
  | FullDetails {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer,
    allInputs :: [(input, output, Int, Int, rating)],
    -- ^ table of all inputs used for rating, and the resulting output
    programInfo :: ProgramInformation,
    -- ^ instructions, reactions, and patterns referenced
    geneDescriptions ::  [String],
    -- ^ same as in 'GenomeDetails'
    instructionDetails :: [(Instruction, BitString, BitString, EncodingDetails)],
    -- ^ same as in 'GenomeDetails'
    efficaciousInstructions :: IntMap [(Int, Instruction)],
    -- ^ map from cell index to indexed efficacious instructions
    allEfficaciousInstructions :: [(Int, Instruction)],
    -- ^ all efficacious instructions
    efficaciousReactions :: IntMap (Set Reaction),
    -- ^ map from cell index to efficacious reactions
    allEfficaciousReactions :: Set Reaction,
    -- ^ all efficacious reactions
    efficaciousGeneDescriptions :: IntMap [String],
    -- ^ map from cell index to gene details of efficacious
    -- instructions
    allEfficaciousGeneDescriptions :: [String]
    -- ^ gene details as in 'GenomeDetails' but only for
    -- efficacious instructions
    }
    -- ^ summary of the agent, plus result of all inputs,
    -- analysis of efficacious genes, complete listing of the
    -- whole genome, stuff that can be computed using all inputs;
    -- requires significant computation.
  | Efficacious {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer,
    efficaciousInstructions :: IntMap [(Int, Instruction)],
    -- ^ map from cell index to indexed efficacious instructions
    allEfficaciousInstructions :: [(Int, Instruction)],
    -- ^ all efficacious instructions
    efficaciousReactions :: IntMap (Set Reaction),
    -- ^ map from cell index to efficacious reactions
    allEfficaciousReactions :: Set Reaction,
    -- ^ all efficacious reactions
    allEfficaciousGeneDescriptions :: [String]
    -- ^ gene details as in 'GenomeDetails' but only for
    -- efficacious instructions
    }
    -- ^ summary of the agent, plus analysis of efficacious
    -- genes, reduced from 'FullDetails'
  | RatingInputs {
    id :: Integer,
    rating :: rating,
    allInputs :: [(input, output, Int, Int, rating)]
    }
    -- ^ table of all inputs used for rating, and the resulting output
  | ReactionNetwork {
      id :: Integer,
      reactionNetwork :: StyledNetwork
      }
    deriving (Show, Data, Typeable, Generic)


$(deriveJSON myJSONOpts ''AgentResult)

-- | Type alias that's really what I had in mind for 'AgentResult', but...
type AgentResultX gs problem =
  AgentResult (Genome gs) (Rating problem) (Input problem) (Output problem)
