{-# LANGUAGE BangPatterns,
RecordWildCards, NamedFieldPuns, DisambiguateRecordFields,
TypeFamilies,
ParallelListComp,
FlexibleInstances, FlexibleContexts, UndecidableInstances, MultiParamTypeClasses,
DeriveDataTypeable,
TemplateHaskell, DeriveGeneric #-}

module RangeMap
       (Range(..),
        BoundedRange(..),
        Offsetable(..),
        RangeMap(..),
        guardNonNull,
        keyIntersection,
        isWellFormed)
       where

import qualified Prelude as P
import Prelude hiding (null, length, take, drop)

import Control.DeepSeq
import Control.Exception
import Control.Monad

import Data.Aeson hiding (Array)
import Data.Aeson.Types hiding (Array)
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.DeriveTH
import Data.Monoid
import Data.Typeable

import GHC.Generics

import AesonUtil
import ShortShow

rangeMap_vcid :: String
rangeMap_vcid = "$Id: RangeMap.hs,v 93fba847b3a8 2017-01-20 21:13:26Z garrett $"

-- | Abstraction for an interval-like discrete block type, more
-- or less indexed from 0

class Range a where

  -- | An empty interval, not necessarily unique
  empty :: a

  -- | Whether the range is empty; default implementation checks
  -- for zero length
  null :: a -> Bool

  null r = 0 == length r

  -- | How many items are in the range
  length :: a -> Int

  -- | Take this many items at the most from the beginning of the
  -- range
  take :: Int -> a -> a

  -- | Drop this many items at the most from the beginning of the
  -- range
  drop :: Int -> a -> a

  -- | Given @[begin, end)@ indices relative to the start of the
  -- range, reduce the range to just those indices.  Default
  -- implementation is in terms of 'drop' and 'take'.
  subrange :: Int -> Int -> a -> a

  subrange begin end r =
    take (end - begin) (drop begin r)


-- | Guard that the range is non-null
guardNonNull :: (Monad m, MonadPlus m, Range a) => a -> m ()
guardNonNull x =
  guard (not (null x))

-- | A pair of integers @[begin, end)@ is a 'Range'

instance Range (Int, Int) where
  empty = (0, 0)

  null (begin, end) = begin == end

  length (begin, end) = end - begin

  take n r@(begin, end) =
    if n <= 0
    then (begin, begin)
    else
      if n >= length r
      then r
      else
        (begin, begin + n)

  drop n r@(begin, end) =
    if n <= 0
    then r
    else
      if n >= length r
      then (end, end)
      else
        (begin + n, end)

-- A list is a range

instance Range [t] where

  empty = []

  null = P.null

  length = P.length

  take = P.take

  drop = P.drop

-- | A range-like object that can have bounds other than @[0, length)@
class Range t => BoundedRange t where

  -- | Return the @[begin, end)@ indices for this object
  bounds :: t -> (Int, Int)

  -- | Return the intersection of two objects.  The default
  -- implementation assumes @t@'s overlap if their bounds
  -- overlap.
  intersection :: t -> t -> t

  intersection = keyIntersection

instance BoundedRange (Int, Int) where
  bounds pair = pair

instance BoundedRange [t] where
  bounds xs = (0, P.length xs)

instance
  (Range (Maybe t),
   BoundedRange t)
  => BoundedRange (Maybe t) where
  bounds Nothing = (0,0)
  bounds (Just x) = bounds x

-- | Given a key interval in the same numbering system as the
-- keys in a 'RangeMap', return the part of the range spanned by
-- those keys.  This is different from 'subrange': The call
-- @subrange a b r@ gives the @a@th through @(b-1)@th items from
-- r.  For example, @subrange 0 1 r@ gives just the first item.
-- The call @keyIntersection (a, b) r@ gives the items in @r@ whose
-- indices are in @[a,b)@.  Return a null value if the
-- intersection is null.

keyIntersection ::
  (Range k, BoundedRange k,
   Range v, BoundedRange v) =>
  k -> v -> v

keyIntersection k v =
  let
    (xBegin, xEnd) = bounds k
    (yBegin, yEnd) = bounds v
    s = take (xEnd - xBegin) (drop (xBegin - yBegin) v)
  in s

-- | A 'RangeMap' maps consecutive integer indices to items in
-- other ranges.  The lengths of the key range and the value
-- range in each association item must agree.
-- To get anything done with this, you need instances
-- 'Range k', 'Range v', 'BoundedRange k'.

data RangeMap k v = RangeMap { rangeMapItems :: [(k, v)] }
  deriving (Eq, Ord, Data, Typeable)

instance (Show k, Show v) => Show (RangeMap k v) where
  show (RangeMap pairs) = show pairs

instance (Show k, Show v) => ShortShow (RangeMap k v) where
  shortShow rm = show rm

instance (Read k, Read v) => Read (RangeMap k v) where
  -- readsPrec :: Int -> String -> [(item, String)]
  readsPrec d str = do
    (pairs, rest) <- readsPrec d str
    return (RangeMap pairs, rest)

-- | Assert that the 'RangeMap' is well-formed, meaning it
-- should be of the form @[ ((x0, x1), y1), ((x1, x2), y2), ...]@
-- and @length y[n] == x[n] - x[n-1]@

isWellFormed (RangeMap []) = True

isWellFormed (RangeMap rm@((k, _):_)) =
  let
    (begin, end) = bounds k
  in
   loop begin rm
  where
    loop j ((k, v) : items) =
      let
        (begin, end) = bounds k
      in
       if (j == begin) && (begin < end)
       then
         loop end items
       else
         False

    loop _ [] = True


-- | A 'RangeMap' is a 'Range'

instance (BoundedRange k, Range k, Range v) => Range (RangeMap k v) where

  empty = RangeMap []

  null (RangeMap []) = True
  null (RangeMap _) = False

  length (RangeMap []) = 0

  length (RangeMap items) =
    -- assertWellFormed items $
    sum [length k | (k, v) <- items]

  take n whole@(RangeMap items) =
    -- assertWellFormed items $
    if n <= 0
    then empty
    else
      if n >= length whole
      then whole
      else RangeMap (loop n items)
    where
      loop _ [] = []
      loop m (item@(k, v) : items) =
        if m == 0
        then empty
        else
          if m <= length k
          then -- take part
            let
              newKey = take m k
              newValue = take m v
            in [(newKey, newValue)]
          else -- take whole and keep going
            item : loop (m - length k) items

  drop n whole@(RangeMap items) =
    -- assertWellFormed items $
    if n <= 0
    then whole
    else
      if n >= length whole
      then empty
      else RangeMap (loop n items)
    where
      loop _ [] = []
      loop m ((k, v) : items) =
        if m < length k
        then -- take part
          let
            newKey = drop m k
            newValue = drop m v
          in
           (newKey, newValue) : items
        else -- drop whole and keep going
          loop (m - length k) items


class Offsetable a where

  -- | Given some sort of position-like object, offset it by an
  -- integer amount
  offset :: Int -> a -> a


instance Offsetable (Int, Int) where
  offset n (begin, end) = (begin + n, end + n)


instance (Offsetable k, BoundedRange k, Range k, Range v)
         => Offsetable (RangeMap k v) where

  offset n (RangeMap pairs) =
    RangeMap [(offset n k, v) | (k, v) <- pairs]


instance (Offsetable k, BoundedRange k, Range k, Range v)
         => Monoid (RangeMap k v) where

  mempty = empty

  mappend xRM@(RangeMap xs) yRM@(RangeMap ys) =
    let
      (xBegin, xEnd) = bounds xRM
      (yBegin, yEnd) = bounds yRM
      RangeMap ysOffset = offset (xEnd - yBegin) yRM
    in
     RangeMap (xs ++ ysOffset)


instance
  (BoundedRange k, Range k, Range v) =>
  BoundedRange (RangeMap k v) where

    bounds (RangeMap []) = (0,0)

    bounds (RangeMap items) =
      let
        (kHead, _) = P.head items
        (begin, _) = bounds kHead
        (kTail, _) = P.last items
        (_, end) = bounds kTail
      in
       (begin, end)

$(derive makeBinary ''RangeMap)
$(derive makeNFData ''RangeMap)
$(deriveJSON myJSONOpts ''RangeMap)
