{-# LANGUAGE TypeFamilies, TypeSynonymInstances, FlexibleInstances,
  MultiParamTypeClasses #-}

-- | Tools for a 'Monad' for converting data objects to and from
-- strings of bits.

module BitSource where

import Control.Monad
import Control.Monad.Identity
import Control.Monad.State
import Data.Word
import Data.Int
import Data.Bits.Utils
--import Debug.Trace

import BitUtil
import BitString

bitSource_vcid :: String
bitSource_vcid = "$Id: BitSource.hs,v 1607c537d3ac 2017-01-19 18:10:07Z garrett $"

class Monad m => ReadBits m where

  -- | Read @n@ bits from the monad
  readBits ::
    Int -- ^ @n@
    -> m [Bool]
  readBitsAsWord :: Integral n => Int -> m n
  bitsLeft :: m Bool

class BitSpec spec where
  type FromBitsResult spec

  -- | Read an object from a stream of bits
  readFromBits :: ReadBits m => spec -> m (FromBitsResult spec)

class BitSpec spec => BitSpecSize spec where
  -- | How many bits does it take to represent 'x' using the
  -- given specification
  sizeInBits :: spec -- ^ encoding specification
                -> FromBitsResult spec -- ^ 'x'
                -> Int

-- | Attach bit source state to a 'Monad'
type BitSourceT m = StateT BitString m

instance Monad m => ReadBits (BitSourceT m) where
  readBits n = do
    bs <- get
    let hd = bsGetBits 0 n bs
    let tl = bsDrop n bs
    put tl
    return hd

  readBitsAsWord n = do
    bs <- get
    let hd = bsGetBitsAsWord 0 n bs
    let tl = bsDrop n bs
    put tl
    return (fromIntegral hd)


  bitsLeft = do
    xs <- get
    return (not (bsNull xs))

-- | A monad that's just a bit stream state
type BitSource = BitSourceT Identity

-- | Specify an 'Int' by the number of bits
data IntSpec = IntSpec Int
             deriving (Read, Show, Eq, Ord)

instance BitSpec IntSpec where
  type FromBitsResult IntSpec = Int

  readFromBits (IntSpec nBits) =
    readBitsAsWord nBits

instance BitSpecSize IntSpec where
  sizeInBits (IntSpec n) _ = n


-- | Specify a fixed-length vector of 'n' objects with the same bit spec

data FixedVectorSpec subSpec =
  FixedVectorSpec Int subSpec
  -- ^ 'n', and number of objects, specification for sub-objects
  deriving (Read, Show, Eq, Ord)

instance BitSpec subSpec => BitSpec (FixedVectorSpec subSpec) where
  type FromBitsResult (FixedVectorSpec subSpec) = [FromBitsResult subSpec]
  readFromBits (FixedVectorSpec nItems itemSpec) =
    replicateM nItems (readFromBits itemSpec)

instance BitSpecSize subSpec => BitSpecSize (FixedVectorSpec subSpec) where
  sizeInBits (FixedVectorSpec _ itemSpec) items =
    sum (map (sizeInBits itemSpec) items)


-- | Specify that the rest of all bits consists of items with the
-- same bit spec.

data ListRemainderSpec subSpec =
  ListRemainderSpec
  subSpec -- ^ specification for sub-objects

instance BitSpec subSpec => BitSpec (ListRemainderSpec subSpec) where
  type FromBitsResult (ListRemainderSpec subSpec) = [FromBitsResult subSpec]
  readFromBits spec@(ListRemainderSpec itemSpec) = do
    b <- bitsLeft
    if b
      then do
      item <- readFromBits itemSpec
      items <- readFromBits spec
      return (item : items)
      else
      return []

instance BitSpecSize subSpec => BitSpecSize (ListRemainderSpec subSpec) where
  sizeInBits (ListRemainderSpec itemSpec) items =
    sum (map (sizeInBits itemSpec) items)

-- | Specify a variable-length vector of items with the same bit
-- spec.

data VariableVectorSpec subSpec =
  VariableVectorSpec Int subSpec
  {- ^ number of bits to read for the count of objects
   specification for sub-object -}

instance BitSpec subSpec => BitSpec (VariableVectorSpec subSpec) where
  type FromBitsResult (VariableVectorSpec subSpec) = [FromBitsResult subSpec]
  readFromBits (VariableVectorSpec nBits itemSpec) = do
    nItems <- readFromBits (IntSpec nBits)
    replicateM nItems (readFromBits itemSpec)

instance BitSpecSize subSpec => BitSpecSize (VariableVectorSpec subSpec) where
  sizeInBits (VariableVectorSpec nBits itemSpec) items =
    nBits + sum (map (sizeInBits itemSpec) items)



instance (BitSpec t1, BitSpec t2) => BitSpec (t1, t2) where
  type FromBitsResult (t1, t2) = (FromBitsResult t1, FromBitsResult t2)
  readFromBits (s1, s2) = do
    x1 <- readFromBits s1
    x2 <- readFromBits s2
    return (x1, x2)

instance (BitSpecSize t1, BitSpecSize t2) => BitSpecSize (t1, t2) where
  sizeInBits (s1, s2) (x1, x2) =
    sizeInBits s1 x1 + sizeInBits s2 x2

instance (BitSpec t1, BitSpec t2, BitSpec t3) => BitSpec (t1, t2, t3) where
  type FromBitsResult (t1, t2, t3) = (FromBitsResult t1, FromBitsResult t2, FromBitsResult t3)
  readFromBits (s1, s2, s3) = do
    x1 <- readFromBits s1
    x2 <- readFromBits s2
    x3 <- readFromBits s3
    return (x1, x2, x3)

instance (BitSpecSize t1, BitSpecSize t2, BitSpecSize t3) =>
         BitSpecSize (t1, t2, t3) where
  sizeInBits (s1, s2, s3) (x1, x2, x3) =
    sizeInBits s1 x1 + sizeInBits s2 x2 + sizeInBits s3 x3

{- | Read an object from the front of a list of bits using the
given spec, and give the result and remaining bits -}

readFromBitList :: (BitSpec s) => s -> BitString -> (FromBitsResult s, BitString)
readFromBitList spec bits =
  runIdentity (runStateT (readFromBits spec) bits)


-- | Read an object from the front of a list of bits using the
-- given spec, and give the result discarding any remaining bits

fromBitList :: (BitSpec s) => s -> BitString -> FromBitsResult s
fromBitList spec bits =
  fst (readFromBitList spec bits)

-- | Encode a list of bits as bytes ('Word8'). The first four
-- bytes are the number of bits.

packIntoWords :: [Bool] -> [Word8]

packIntoWords bs =
  let bitCount :: Int32
      bitCount = fromIntegral $ length bs
  in
   map fromIntegral (getBytes bitCount) ++ bitsToBytes bs

-- a private looper:
bitsToBytes :: [Bool] -> [Word8]
bitsToBytes [] = []

bitsToBytes bs =
  let (front, back) = splitAt 8 bs
      --This isn't necessary:
      --front = front0 ++ (replicate (8 - length front0) False)
  in fromBits front : bitsToBytes back

-- | Unpack bits from a list of bytes ('Word8') from the same
-- format as produced by 'packIntoWords'.

unpackFromWords :: [Word8] -> [Bool]

unpackFromWords bs =
  let (bitCountBytes, payload) = splitAt 4 bs
      bitCountInt32s :: [Int32]
      bitCountInt32s = map fromIntegral bitCountBytes
      bitCount32 :: Int32
      bitCount32 = fromBytes bitCountInt32s
      bitCount :: Int
      bitCount = fromIntegral bitCount32
  in
   take bitCount (concatMap (toBits 8) payload)
