{-# LANGUAGE MultiParamTypeClasses, FlexibleInstances,
FlexibleContexts, BangPatterns, ScopedTypeVariables, RankNTypes,
UndecidableInstances #-}

{- |

Module: RandomST

Description : Monadic random object generation.

Copyright   : © 2015 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

This wraps a C implementation of the Mersenne twister algorithm,
specifically, a modification of the dSFMT package by Saito and
Matsumoto.

-}


module RandomST (
  RandomST(..),
  runRandomST,
  withSeed,
  withSeedFromInt,
  withSeedFromTime,
  withSeedFromDevRandom,
  createRandomSeed,
  randomUnitDouble,
  randomInt,
  Sample(..),
  pullIterate,
  pullIterateList,
  pullSampleList,
  pullSampleListDS,
  binomial,
  Likelihood(..),
  ExpectedValue(..),
  -- Distributions
  Surely(..),
  Alternative(..),
  UnitUniformDistribution(..),
  UniformDistribution(..),
  UniformFullRangeDistribution(..),
  UniformBits(..),
  BernoulliDistribution(..),
  BinomialDistribution(..),
  ExponentialDistribution(..),
  DiscretizedPoissonProcess(..)
  ) where

import Control.Monad
import Control.Monad.ST
import Data.Word
import Data.Array
import qualified Data.Sequence as Seq
import Data.Time
import Foreign
import Control.DeepSeq
import System.CPUTime
import System.IO
import System.Environment
import Control.Monad.Reader

--import Logging
import BitUtil
import BitString
--import Util

import qualified CSFMT as MT

randomST_vcid :: String
randomST_vcid = "$Id: RandomST.hs,v e3f8b7346295 2015-08-05 15:15:28Z garrett $"

type RandomST s a = ReaderT MT.Generator (ST s) a

runRandomST ::  forall a. (forall s. RandomST s a) -> MT.Generator -> a
runRandomST axn g = runST $ runReaderT axn g

-- | Takes a few 'Word32' to use as a seed for random number
-- generation during @axn@.  Also returns a few 'Word32' that
-- don't represent the whole generator state, but would be
-- appropriate for re-seeding the generator for the next
-- calculation.

withSeed :: [Word32] -> (forall s. RandomST s a) -> (a, [Word32])
withSeed words axn =
  withGenerator (newGeneratorFromWords words) axn

-- | Use a fixed integer to seed the random number generator
-- during @axn@.  Also returns a few 'Word32' that don't
-- represent the whole generator state, but would be appropriate
-- for re-seeding the generator for the next calculation.

withSeedFromInt :: Int -> (forall s. RandomST s a) -> (a, [Word32])
withSeedFromInt n axn =
  withGenerator (newGeneratorFromInt n) axn

-- | Use the current time to seed the random number generator
-- during @axn@.  Also returns a few 'Word32' that don't
-- represent the whole generator state, but would be appropriate
-- for re-seeding the generator for the next calculation.

withSeedFromTime :: (forall s. RandomST s a) -> IO (a, [Word32])
withSeedFromTime axn = do
  g <- newGeneratorFromTime
  return (withGenerator g axn)

-- | Use data from @\/dev\/urandom@ to seed the random number
-- generator during @axn@.  Also returns a few 'Word32' that
-- don't represent the whole generator state, but would be
-- appropriate for re-seeding the generator for the next
-- calculation.

withSeedFromDevRandom :: (forall s. RandomST s a) -> IO (a, [Word32])
withSeedFromDevRandom axn = do
  g <- newGeneratorFromDevRandom
  return (withGenerator g axn)

-- | Create a random seed.  First see if the system environment
-- variable RNGSEED is set.  Otherwise, pull a seed from
-- @\/dev\/urandom@

createRandomSeed :: MonadIO m => m [Word32]
createRandomSeed = do
    systemEnv <- liftIO getEnvironment
    mEnvSeed <- case lookup "RNGSEED" systemEnv of
      Just seedAsIntStr -> do
        --msgInfo ("Initializing random seed: $RNGSEED=" ++ seedAsIntStr)
        let readings = reads seedAsIntStr
        case readings of
          (n :: Int,_):_ -> return (Just (snd $ withSeedFromInt n $ return ()))
          [] -> do
            --msgInfo ("Couldn't parse int from " ++ seedAsIntStr)
            return Nothing
      Nothing ->
        return Nothing
    case mEnvSeed of
      Just seed ->
        return seed
      Nothing -> do
        --msgInfo "Initializing random seed from /dev/random"
        liftIO newSeedFromDevRandom


{- this is private -}
withGenerator ::  forall a. MT.Generator -> (forall s. RandomST s a)
                  -> (a, [Word32])
withGenerator g axn =
  runRandomST addReseed g
  where
    addReseed = do
      val <- axn
      newSeed <- replicateM 8 randomInt
      return (val, newSeed)

{- this is private -}
newGeneratorFromTime :: IO MT.Generator
newGeneratorFromTime = do
    ct <- getCPUTime
    now <- getCurrentTime
    let sec = round (1e6 * utctDayTime now)
    return (MT.newGeneratorFromInt $
            fromIntegral (sec + ct `div` cpuTimePrecision))

{- this is private -}
newSeedFromTime :: IO [Word32]
newSeedFromTime = do
  g <- newGeneratorFromTime
  return (fst $ withGenerator g $ replicateM 8 randomInt)

{- this is private -}

{- Note: @/dev/urandom@ doesn't block waiting for more entropy,
which @/dev/random@ does.  This pulls 32 bytes (256 bits, 8
'Word32's), which is supposedly plenty for most cryptographic
needs -}

newSeedFromDevRandom :: IO [Word32]
newSeedFromDevRandom =
  withBinaryFile "/dev/urandom" ReadMode $ \h -> do
    buf <- mallocArray 8
    _ <- hGetBuf h buf (8 * sizeOf (0 :: Word32))
    words <- peekArray 8 buf
    free buf
    return words

newGeneratorFromDevRandom :: IO MT.Generator
newGeneratorFromDevRandom = do
  words <- newSeedFromDevRandom
  return (MT.newGeneratorFromWords words)

{- this is private -}
newGeneratorFromInt :: Int -> MT.Generator
newGeneratorFromInt = MT.newGeneratorFromInt

{- this is private -}
newGeneratorFromWords :: [Word32] -> MT.Generator
newGeneratorFromWords = MT.newGeneratorFromWords

randomUnitDouble :: RandomST s Double
randomUnitDouble =
  do
    g <- ask
    lift (MT.randomUnitDouble g)

randomInt :: (Integral n, Bounded n) => RandomST s n
randomInt =
  do
    g <- ask
    lift (MT.randomInt g)

-- | Class Sample represents functions for pulling a sample of
-- type t out of some probability distribution.

class Sample distribution t where

  -- | Pull a random item of type t from the given distribution
  pullSample :: distribution -> RandomST s t


-- | Iterate n steps of a distribution formed by a Markov chain
-- and return the last step.

pullIterate :: (Sample distribution distribution) =>
               Int -> distribution -> RandomST s distribution

pullIterate 0 d = return d

pullIterate n d = do
    dNext <- pullSample d
    pullIterate (n - 1) dNext

-- | Iterate n steps of a distribution formed by a Markov chain
-- and return all the intermediate steps.

pullIterateList :: (Sample distribution distribution) =>
                   Int -> distribution -> RandomST s [distribution]

pullIterateList 0 d = return [d]

pullIterateList n d = do
  dNext <- pullSample d
  dRest <- pullIterateList (n - 1) dNext
  return (dNext : dRest)


-- | Pull a list of n random items of type t from the given distribution.

pullSampleList :: (Sample distribution t) =>
                  Int -> distribution -> RandomST s [t]

pullSampleList n dist =
  replicateM n (pullSample dist)


-- | Private; A very strict version of 'replicate'

replicateMDS ::
  (Monad m, NFData a) =>
  Int -> m a -> m [a]

replicateMDS 0 _ = return []
replicateMDS n axn = do
  next <- axn
  deepseq next (do
                   rest <- replicateMDS (n-1) axn
                   return (next:rest))

-- | Pull a list of n random items of type t from the given
-- distribution, but using deepseq on each one.

pullSampleListDS :: (Sample distribution t, NFData t) =>
              Int -> distribution -> RandomST s [t]
pullSampleListDS n dist = replicateMDS n (pullSample dist)

-- | Class Likelihood has one method that computes the probability of
-- getting the given data from a distribution.

class Likelihood distribution t where
    likelihood :: distribution -> t -> Double


-- | Class ExpectedValue has one method that computes the
-- expected value or mean of a distribution.

class ExpectedValue distribution t where
    expectedValue :: distribution -> t


instance (Sample d1 t1, Sample d2 t2) => Sample (d1, d2) (t1, t2) where
  pullSample (d1, d2) = do
    x1 <- pullSample d1
    x2 <- pullSample d2
    return (x1, x2)


-- | Return a fixed item.  Any time you sample from this
-- distribution it gives the same value, and doesn't even consume
-- a random number.

data Surely t = Surely t

instance Sample (Surely t) t where
    pullSample (Surely x) = return x

instance Eq t => Likelihood (Surely t) t where
    likelihood (Surely x) y = if x == y then 1.0 else 0.0

instance ExpectedValue (Surely t) t where
    expectedValue (Surely x) = x

-- | Uniformly distributed real numbers on the interval [0,1]

data UnitUniformDistribution =
  UnitUniformDistribution
  deriving (Read, Show, Eq, Ord)

unitUniformDistribution :: UnitUniformDistribution
unitUniformDistribution = UnitUniformDistribution

instance Sample UnitUniformDistribution Double where

  pullSample UnitUniformDistribution =
    randomUnitDouble

instance Likelihood UnitUniformDistribution Double where
    likelihood UnitUniformDistribution x =
        if x < 0 then 0.0
        else if x > 1 then 0.0
        else 1.0

-- | Pull an item distributed uniformly between min and max.

data UniformDistribution t
    = UniformDistribution (t, t) -- ^ The bounds (min, max) are given here.

instance Sample (UniformDistribution Double) Double where
  pullSample (UniformDistribution (a, b)) = do
    x <- randomUnitDouble
    return (a + x*(b - a))

instance Likelihood (UniformDistribution Double) Double where
    likelihood (UniformDistribution (a, b)) x =
        if x < a then 0.0
        else if x > b then 0.0
        else 1 / (b - a)

instance ExpectedValue (UniformDistribution Double) Double where
    expectedValue (UniformDistribution (a, b)) = (a + b) / 2


instance (Integral n, Bounded n) =>
         Sample (UniformDistribution n) n where
  pullSample (UniformDistribution (a, b)) = do
    x <- randomUnitDouble
    return (a + floor (x * fromIntegral (b - a + 1)))


data UniformFullRangeDistribution t =
  UniformFullRangeDistribution
  deriving (Read, Show, Eq, Ord)

instance (Integral n, Bounded n) =>
         Sample (UniformFullRangeDistribution n) n where
  pullSample UniformFullRangeDistribution = randomInt

-- | Pull 'True' (= 1) with probability @p@, or 'False' (= 0) with
-- probability 1 - @p@.

data BernoulliDistribution = BernoulliDistribution !Double -- ^ p
                           deriving (Read, Show, Eq)

instance Sample BernoulliDistribution Bool where
    pullSample (BernoulliDistribution p) = do
      x <- randomUnitDouble
      return (x < p)

instance Likelihood BernoulliDistribution Bool where
    likelihood (BernoulliDistribution p) x =
        case x of
          True -> p
          False -> 1 - p

instance Sample BernoulliDistribution Int where
    pullSample (BernoulliDistribution p) = do
      x <- randomUnitDouble
      return (if x < p then 1 else 0)

instance Likelihood BernoulliDistribution Int where
    likelihood (BernoulliDistribution p) x =
        case x of
          1 -> p
          0 -> 1 - p
          _ -> 0

instance Sample BernoulliDistribution Double where
    pullSample (BernoulliDistribution p) = do
      x <- randomUnitDouble
      return (if x < p then 1 else 0)

instance Likelihood BernoulliDistribution Double where
    likelihood (BernoulliDistribution p) x =
        case x of
          1 -> p
          0 -> 1 - p
          _ -> 0

instance ExpectedValue BernoulliDistribution Double where
    expectedValue (BernoulliDistribution p) = p

data BinomialDistribution = BinomialDistribution Double Int
                          deriving (Read, Show, Eq)

instance Sample BinomialDistribution Int where
    pullSample (BinomialDistribution p n) =
        let trial = BernoulliDistribution p
        in do trials <- pullSampleList n trial
              let result = sum trials
              result `seq` return result

instance Likelihood BinomialDistribution Int where
    likelihood (BinomialDistribution p n) m
        | (p == 1) =
            if m == n then 1 else 0
        | (p == 0) =
            if m == 0 then 1 else 0
        | (0 < p && p < 1) =
            if m < 0 then 0
            else if m > n then 0
                 else (fromIntegral (binomial (toInteger n)
                                              (toInteger m)))
                          * p ^ m * (1 - p) ^ (n - m)
        | otherwise = error ("likelihood of BinomialDistribution: " ++ show p)

instance ExpectedValue BinomialDistribution Double where
    expectedValue (BinomialDistribution p n) = p * fromIntegral n

-- | The bionomial coefficient (n choose k).

binomial :: Integer -- ^ n
         -> Integer -- ^ k
         -> Integer
binomial !n !k | (k == 0)  = 1 -- includes 0 choose 0
               | (n == 0)  = 0
               | (n > 0 && k > 0) =
                 (binomial (n-1) (k-1)) * n `div` k



-- | Pull from d1 with probability p and from d2 with probability 1 - p.

data Alternative d1 d2
    = Alternative Double d1 d2 -- ^ p, d1, d2

instance (Sample d1 t, Sample d2 t) =>
    Sample (Alternative d1 d2) t where

        pullSample (Alternative p d1 d2) = do
          u <- pullSample (BernoulliDistribution p)
          if u
            then pullSample d1
            else pullSample d2

-- | Pick an element from a list, uniformly

instance Sample [e] e where
  pullSample xs = do
    n <- pullSample (UniformDistribution (0, length xs - 1))
    return (xs !! n)

-- | Pick an element from an array, uniformly
instance Sample (Array Int e) e where
  pullSample a = do
    n <- pullSample (UniformDistribution (bounds a))
    return (a ! n)

-- | Pick an element from a sequence, uniformly
instance Sample (Seq.Seq e) e where
  pullSample s = do
    n <- pullSample (UniformDistribution (0, Seq.length s - 1))
    return (Seq.index s n)

-- | Produce a list of n 'True' and 'False' with equal
-- probability.  Efficiently implemented by pulling a 'Word32'
-- and emitting all of its bits.

data UniformBits = UniformBits !Int -- ^ n
                   deriving (Read, Show, Eq, Ord)

instance Sample UniformBits [Bool] where
  pullSample (UniformBits n) =
    let nWords = (div n 32) + 1
    in do
      (words :: [Word32]) <- replicateM nWords randomInt
      let bits = concatMap (toBits 32) words
      return (take n bits)

-- | Produce an exponentially distributed number with parameter
-- lambda.  The mean is 1/lambda.

data ExponentialDistribution =
  ExponentialDistribution !Double -- ^ lambda
  deriving (Read, Show, Eq, Ord)

instance Sample ExponentialDistribution Double where
  pullSample (ExponentialDistribution lambda) = do
    x <- randomUnitDouble
    return (-(log x) / lambda)

instance ExpectedValue ExponentialDistribution Double where
  expectedValue (ExponentialDistribution lambda) = 1/lambda

instance Likelihood ExponentialDistribution Double where
  likelihood (ExponentialDistribution lambda) x =
    lambda * exp (- lambda * x)

-- | Produce a list of n Boolean values, where each is true with
-- probability p, approximately.  Really what this does is set
-- some of them to 'True' such that the distance between 'True's
-- is exponential with paramter p.  If p is small, this is a
-- reasonable approximation, and generally faster than generating
-- a list from 'Bernoulli'. You can also pull a random list of
-- indices that would be marked 'True'.

data DiscretizedPoissonProcess =
  DiscretizedPoissonProcess !Double !Int -- ^ p and n
  deriving (Read, Show, Eq, Ord)

instance Sample DiscretizedPoissonProcess [Bool] where
  pullSample (DiscretizedPoissonProcess p n) = do
    x <- pullSample (ExponentialDistribution p)
    let timeToNext = round (x :: Double)
    if n < timeToNext
      then return (replicate n False)
      else if timeToNext == 0
         then do
            rest <- pullSample (DiscretizedPoissonProcess p
                                (n - timeToNext - 1))
            return (True:rest)
         else do
            rest <- pullSample (DiscretizedPoissonProcess p (n - timeToNext))
            return (prependReplicate (timeToNext - 1) False
                    (True:rest))

prependReplicate 0 _ rest = rest
prependReplicate n item rest = prependReplicate (n - 1) item (item:rest)

-- | Produce indices of random arrival events on [0,n]

instance Sample DiscretizedPoissonProcess [Int] where
  pullSample (DiscretizedPoissonProcess p n) =
    if p == 0 || n == 0 then return [] else loop 0 n
      where loop start remaining = do
              x <- pullSample (ExponentialDistribution p)
              let timeToNext = round (x :: Double)
              if timeToNext > remaining
                then return []
                else
                if timeToNext == 0 -- don't give two arrivals at the same time
                  then loop start remaining
                  else do
                    rest <- loop (start + timeToNext) (remaining - timeToNext)
                    return (start + timeToNext : rest)


-- | Pull @n@ bits, each 0 or 1 with even probability.
instance Sample UniformBits BitString where
  pullSample (UniformBits n) =
    let nWords = (div n 32) + 1
    in do
      words :: [Word32] <- replicateM nWords randomInt
      return (bsTake n (bsFromWord32s words))
