{-# LANGUAGE FlexibleInstances #-}

{-|

Module: MapUtil

Description: Tools for working with sets and maps

Copyright: &#xA9; 2014 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : portable


-}


module MapUtil where

import Data.Map (Map)
import qualified Data.Map as M
import Data.Set (Set)
import qualified Data.Set as S
import Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import Data.IntSet (IntSet)
import qualified Data.IntSet as IS
import qualified Data.Sequence as Seq
import Data.Foldable

mapUtil_vcid :: String
mapUtil_vcid = "$Id: MapUtil.hs,v d8e4380c1184 2017-01-19 18:11:01Z garrett $"

class AsMap c where
  asMap :: Ord k => c k -> Map k ()
  
instance AsMap [] where
  asMap sList = M.fromList [(k, ()) | k <- sList]

instance AsMap Set where
  asMap sSet = M.fromAscList [(k, ()) | k <- S.toAscList sSet ]
               
mapDifference :: (AsMap c, Ord k) => Map k a -> c k -> Map k a
mapDifference m s =
   M.difference m (asMap s)

mapIntersection :: (AsMap c, Ord k) => Map k a -> c k -> Map k a
mapIntersection m s =
  M.intersection m (asMap s)

               
class AsIntMap c where
  asIntMap :: c -> IntMap ()
  
instance AsIntMap [Int] where
  asIntMap sList = IM.fromList [ (k, ()) | k <- sList] 

instance AsIntMap IntSet where
  asIntMap sSet = IM.fromAscList [ (k, ()) | k <- IS.toAscList sSet]
  
instance AsIntMap (Seq.Seq Int) where
  asIntMap sSeq = IM.fromList [ (k, ()) | k <- toList sSeq ] 

intMapDifference :: AsIntMap c => IntMap a -> c -> IntMap a
intMapDifference m s =
   IM.difference m (asIntMap s)

intMapIntersection :: AsIntMap c => IntMap a -> c -> IntMap a
intMapIntersection m s =
   IM.intersection m (asIntMap s)

-- | Map a function over a 'Set' of keys and return a 'Map' of
-- keys to values
setMap ::
  Ord key =>
  (key -> value) -> Set key -> Map key value
  
setMap f keySet =
  M.fromDistinctAscList [(key, f key) | key <- S.toAscList keySet]


-- | Map a function over an 'IntSet' of keys and return a 'IntMap' of
-- keys to values
intSetMap ::
  (Int -> value) -> IntSet -> IntMap value
  
intSetMap f iSet =
  IM.fromDistinctAscList [(key, f key) | key <- IS.toAscList iSet]
